#define VOLUMETRIC_SCATTERING
#define SUNLIGHT_DIRECT_LIGHTING

#include "Shared/MultisampleTexture.inc"

uniform SAMPLER_2D_AUTO_MULTISAMPLE sSourceTex; // Full resolution
uniform SAMPLER_2D_AUTO_MULTISAMPLE sDepthTex; // Full resolution
uniform sampler2D sHalfResSourceTex; // Half resolution
uniform sampler2D sHalfResDepthTex; // Half resolution
//x = near * far, y = near - far, z = far, w = 1 / far
uniform vec4 uZBufferParams;

in vec2 vTexCoord;

#include "Shared/DepthUtils.inc"
#include "Shared/VolumetricScattering/ApplyVolumetrics.inc"

vec4 GetNearestDepthSample(vec2 uv)
{
    const float pointSampleThreshold = 0.002; // Based on manual tweaking.  If any of the surrounding low-res pixels are too far in depth to the target high-res pixel, then we can't use bilinear filtering.

    float highResDepthValue;
#if defined(MSAA)
    highResDepthValue = float(texelFetch(sDepthTex, ivec2(uv * vec2(textureSize(sDepthTex))), gl_SampleID));
#else
    highResDepthValue = float(texture2DLodCompat(sDepthTex, uv, 0.0));
#endif
    vec4 highResDepth = vec4(highResDepthValue); // Fill all 4 components with copies of the depth value

    vec2 halfResTexSize = vec2(textureSize(sHalfResSourceTex, 0));
    vec2 halfResTexSizeRcp = vec2(1.0) / halfResTexSize;

    // Sample 4 depths from the nearby low-res depth texture
    // Round down full-res uv to nearest bottom-left half-res texel
    vec2 uv00TexelNumber = floor(((uv - halfResTexSizeRcp * 0.25) * halfResTexSize) - vec2(0.001)); // Remove an epsilon to aid rounding
    vec2 uv00 = (uv00TexelNumber + 0.5) * halfResTexSizeRcp; // 0.5 is to reach the centre of the bottom left pixel in the quadrant of the downsampled image (needed for the gather function)

    vec4 lowResDepth;
    //lowResDepth = textureGather(sHalfResDepthTex, uv00).wzxy; // Need to swizzle since glsl spec is retarded (0,1)(1,1)(1,0)(0,0)
    lowResDepth = textureGatherCompat(sHalfResDepthTex, uv00, halfResTexSizeRcp); // Need to swizzle since glsl spec is retarded (0,1)(1,1)(1,0)(0,0)

    vec4 depthDiff = abs(lowResDepth - highResDepth);
    bvec4 aboveThreshold = greaterThan(depthDiff, vec4(pointSampleThreshold));
    bool anyAbove = any(aboveThreshold);

    if (!anyAbove) // If we're not at a depth discontinuity (an edge), then use a bilinear filter
    {
        return texture2DLodCompat(sHalfResSourceTex, uv, 0.0);
    }

    // Let's avoid branching - we need to find the nearest depth low-res sample
    float minDepthDiff = min(depthDiff[0], min(depthDiff[1], min(depthDiff[2], depthDiff[3])));
    vec3 lowest = vec3(equal(vec3(minDepthDiff), depthDiff.xyz));
    float id = 0.0;
    float stillLooking = 1.0;
    stillLooking *= 1.0 - lowest.x;
    id += stillLooking;
    stillLooking *= 1.0 - lowest.y;
    id += stillLooking;
    stillLooking *= 1.0 - lowest.z;
    id += stillLooking;
    int nearestID = int(id);

    // UV of the pixel with the closest depth value using the id
    vec2 nearestUV = uv00 + vec2(float(mod(float(nearestID), 2.0)), float(nearestID / 2)) * halfResTexSizeRcp;

    vec4 closestSample = texture2DLodCompat(sHalfResSourceTex, nearestUV, 0.0);

    const float acceptableDepthDiff = pointSampleThreshold;
    if (minDepthDiff > acceptableDepthDiff) // ...if we can't find a close enough depth, just extrapolate the half-res value by the extra depth
    {
        float extrapolation = GetViewSpaceDepth(highResDepthValue, uZBufferParams.xyz) / GetViewSpaceDepth(lowResDepth[nearestID], uZBufferParams.xyz);
        vec4 extrapolatedData = closestSample * extrapolation;
        vec4 change = abs(extrapolatedData - closestSample);
        change.xz = change.xz * exp(-change.y); // Reduce the extra inscattering by the extra outscattering
        return closestSample + change;
    }
    return closestSample;
}

// Add a fast pseudo-random dithering (based on interleaved gradient noise https://bartwronski.com/2016/10/30/dithering-part-three-real-world-2d-quantization-dithering/)
// It seems to perform decently under motion in this dithering shadertoy https://www.shadertoy.com/view/MslGR8
vec4 Dither(vec4 its)
{
    const vec3 magic = vec3(0.06711056, 0.00583715, 52.9829189);
    const float strength = 0.0625; // or (1/16)... How much the dithering pattern will modulate the source value - manually adjusted to look good for the volumetrics information
    return vec4(vec3(its.rgb * (1.0 + fract(magic.z * fract(dot(gl_FragCoord.xy, magic.xy))) * strength)), its.a);
}

void main()
{
    vec4 halfResSourceSample = GetNearestDepthSample(vTexCoord);
    vec4 fullResDestinationSample;

#if defined(MSAA)
    fullResDestinationSample = texelFetch(sSourceTex, ivec2(vTexCoord * vec2(textureSize(sSourceTex))), gl_SampleID);
#else
    fullResDestinationSample = texture2DLodCompat(sSourceTex, vTexCoord, 0.0);
#endif

    halfResSourceSample = Dither(halfResSourceSample);
    gl_FragColor = ApplyVolumetricScattering(fullResDestinationSample, halfResSourceSample);
}
