#ifndef CONVERSION_UTILS_INC
#define CONVERSION_UTILS_INC


/** Converts a colour from SRGB space to linear space */
//[ //The obfuscator doesn't like functions with multiple returns
vec3 SRGBToLinear(vec3 srgbColour)
{
    // Fast approximation taken from http://chilliant.blogspot.co.uk/2012/08/srgb-approximations-for-hlsl.html
    // @TODO: Still needs more testing and profiling before switching over
    //return srgbColour * (srgbColour * (srgbColour * 0.305306011 + 0.682171111) + 0.012522878);
#if defined(GAMMA_CORRECT_INPUTS)
    return srgbColour * srgbColour;
#else
    return pow(srgbColour, vec3(2.2, 2.2, 2.2));
#endif
}
//]

/** Converts a linear space colour to SRGB space */
vec3 LinearToSRGB(vec3 linearColour)
{
    // Approximation taken from http://chilliant.blogspot.co.uk/2012/08/srgb-approximations-for-hlsl.html
    // NOTE: This is necessary for platforms that support sRGB atlas textures, as the error margin in the previous
    // method pow(linearColour, 1.0 / 2.2) was much too great. There is a further optimisation on this page involving
    // three sqrt()s, but that can give negative values with small input numbers so should be avoided
    return max(vec3(1.055) * pow(linearColour, vec3(0.416666667)) - vec3(0.055), vec3(0.0));
}

float LinearToSRGB(float linearColour)
{
    const float exp = 1.0 / 2.2;
    return pow(linearColour, exp); // monitor will apply an additional pow(x, 2.2) curve
}

/** Runescape specific linear to sRGB curve that stops dark areas from being too bright */
vec3 LinearToSRGBRunescape(vec3 linearColour)
{
//   const float exp = 1.0 / 2.0;
//   return pow(linearColour, vec3(exp, exp, exp)); // monitor will apply an additional pow(x, 2.2) curve
// Optimised version of the above code
   return sqrt(linearColour);
}

float LinearToSRGBRunescape(float linearColour)
{
//   const float exp = 1.0 / 2.0;
//   return pow(linearColour, exp); // monitor will apply an additional pow(x, 2.2) curve
// Optimised version of the above code
   return sqrt(linearColour);
}

// These are adapted into GLSL from the HLSL functions found at : http://stackoverflow.com/questions/4728581/hsl-image-adjustements-on-gpu, which were
//  in turn adapted from those found at http://paulbourke.net/texture_colour/colourspace/ (Credits to Paul Bourke for this one.)
// HSL --> RBG --> HSL functions with little GPU overlay, even so these should only be used when we absolutely need to,
// currently this is restricted to animating vertex colours by shifting in HSL colour space... (For both Models and Billboards)
// Don't add anything new using this unless you absolutely have to!
// RBG to HSL
vec4 convertRGBtoHSL(vec4 col)
{
    const float onesixth = 1.0 / 6.0;

    float red = col.r;
    float green = col.g;
    float blue = col.b;

    float minc = min(min(col.r, col.g), col.b);
    float maxc = max(max(col.r, col.g), col.b);
    float delta = maxc - minc;

    float lum = (minc + maxc) * 0.5;
    float sat = 0.0;
    float hue = 0.0;

    if (lum > 0.0 && lum < 1.0)
    {
        float mul = (lum < 0.5) ? (lum) : (1.0 - lum);
        sat = delta / (mul * 2.0);
    }

    //Avoid NaNs caused by div0s below
    if (delta > 0.0)
    {
        vec3 masks = vec3(
            (maxc == red   && maxc != green) ? 1.0 : 0.0,
            (maxc == green && maxc != blue) ? 1.0 : 0.0,
            (maxc == blue  && maxc != red) ? 1.0 : 0.0
        );

        vec3 adds = vec3(
            ((green - blue) / delta),
            2.0 + ((blue - red) / delta),
            4.0 + ((red - green) / delta)
        );

        hue += dot(adds, masks);
        hue *= onesixth;

        if (hue < 0.0)
        {
            hue += 1.0;
        }
    }

    return vec4(hue, sat, lum, col.a);
}

/** HSL to RBG */
vec4 convertHSLtoRGB(vec4 col)
{
    const float onethird = 1.0 / 3.0;
    const float twothird = 2.0 / 3.0;
    const float rcpsixth = 6.0;

    float hue = col.x;
    float sat = col.y;
    float lum = col.z;

    vec3 xt = vec3(
        rcpsixth * (hue - twothird),
        0.0,
        rcpsixth * (1.0 - hue)
    );

    if (hue < twothird)
    {
        xt.r = 0.0;
        xt.g = rcpsixth * (twothird - hue);
        xt.b = rcpsixth * (hue - onethird);
    }

    if (hue < onethird)
    {
        xt.r = rcpsixth * (onethird - hue);
        xt.g = rcpsixth * hue;
        xt.b = 0.0;
    }

    xt = min(xt, 1.0);

    float sat2 = 2.0 * sat;
    float satinv = 1.0 - sat;
    float luminv = 1.0 - lum;
    float lum2m1 = (2.0 * lum) - 1.0;
    vec3  ct = (sat2 * xt) + satinv;

    vec3 rgb;
    if (lum >= 0.5)
    {
        rgb = (luminv * ct) + lum2m1;
    }
    else
    {
        rgb = lum    * ct;
    }

    return vec4(rgb, col.a);
}

vec3 oldscapeConvertHueToRGB(float h)
{
    vec3 rgb = vec3(1.0, -1.0, -1.0) * abs(vec3(6.0 * h) - vec3(3.0, 2.0, 4.0)) + vec3(-1.0, 2.0, 2.0);
    return clamp(rgb, vec3(0.0), vec3(1.0));
}

/**
 * Converts a normalized HSL value to an RGB value in the same manner as OldScape.
 * Note that this function behaves slightly differently than convertHSLtoRGB:
 * It handles edge cases in a robust manner (like luminance values of 1.0)
 * for which convertHSLtoRGB just returns black.
 */
vec3 oldscapeConvertHSLtoRGB(vec3 hsl)
{
    vec3 rgb = oldscapeConvertHueToRGB(hsl.x);
    float c = (1.0 - abs(2.0 * hsl.z - 1.0)) * hsl.y;
    return (rgb - 0.5) * c + hsl.z;
}

/**
 * Converts a normalized RGB value to HSL in the same manner as OldScape.
 * This conversion function handles edge cases slightly differently than
 * convertRGBtoHSL.
 */
vec3 oldscapeConvertRGBtoHSL(vec3 c)
{
    float cMin = min(min(c.r, c.g), c.b);
    float cMax = max(max(c.r, c.g), c.b);
    float sum = cMax + cMin;
    float delta = cMax - cMin;

    vec3 hsl = vec3(0.0, 0.0, sum / 2.0);

    if(delta != 0.0)
    {
        hsl.y = delta / mix(2.0 - sum, sum, hsl.z < 0.5);

        vec3 dRGB = (vec3(cMax) - c) / (6.0 * delta) + 0.5;

        bvec2 maxi = equal(c.rg, vec2(cMax));
        vec3 hueCoeffs = mix(mix(vec3(2.0 / 3.0, dRGB.g, -dRGB.r), vec3(1.0 / 3.0, dRGB.r, -dRGB.b), bvec3(maxi.y)), vec3(0.0, dRGB.b, -dRGB.g), bvec3(maxi.x));
        hsl.x = fract(dot(hueCoeffs, vec3(1.0)));
    }

    return hsl;
}

#endif
