uniform vec3 uSamplePixelSizeAndIteration;

in vec2 vTexCoord;

uniform sampler2D sSourceTex;
uniform sampler2D sFocusTex;
uniform sampler2D sSourceDepthTex;
uniform vec4 uBokehParams;
uniform vec2 uFocusSize;

#include "DepthOfFieldShared.inc"

void main()
{
    vec2 samplePixelSize = uSamplePixelSizeAndIteration.xy;
    float iteration = uSamplePixelSizeAndIteration.z;

    vec4 focus = texture2DLodCompat(sFocusTex, vTexCoord, 0.0);
    float blur;
#if defined(SCENE_BLUR_NEAR)
    blur = focus.x;
#else
    blur = focus.y;
#endif
    // Early out on pixels unaffected by this blur plane
    if (blur == 0.0)
    {
    #if defined(SCENE_BLUR_NEAR)
        // Check if early out is going to cause artifacts (affects near plane only) and if so output original pixel instead
        vec4 nearFocus = KawaseFilter1(sFocusTex, vTexCoord, uFocusSize, iteration, 0.0);
        if (nearFocus.x + nearFocus.y + nearFocus.z + nearFocus.w > 0.0)
        {
            gl_FragColor.rgb = texture2DLodCompat(sSourceTex, vTexCoord, 0.0).rgb;
            return;
        }
    #endif
        discard;
        return;
    }

    vec3 result1 = vec3(0.0), result2 = vec3(0.0), result3 = vec3(0.0), result4 = vec3(0.0);
    vec4 mask = vec4(1.0);
    float numBlurPixels = 4.0;
    KawaseFilter3(sSourceTex, vTexCoord, samplePixelSize, iteration, result1, result2, result3, result4, 0.0);

    // For far plane blur, cancel out any filter texels that hit sharp objects
    // This prevents any ghosting or blurring of sharp objects
#if defined(SCENE_BLUR_FAR)
    vec4 depths = KawaseFilter1(sSourceDepthTex, vTexCoord, samplePixelSize, iteration, 0.0);
    depths = GetViewSpaceDepth(depths, uZBufferParams.xyz) - uDofFocalPoint;
    vec4 nearMask = step(uDofParams.x + uDofParams.y * 0.1, -depths);
    vec4 farMask = step(uDofParams.z + uDofParams.w * 0.1, depths);
    mask = max(nearMask, farMask);
#endif

    // We also cancel out any filter texels that hit NaN pixels, as these would otherwise create a patch of black pixels
    mask = max(mask - vec4(isNaN(result1.r), isNaN(result2.r), isNaN(result3.r), isNaN(result4.r)), 0.0);

    bvec4 boolMask = bvec4(mask);
    result1 = boolMask.x ? result1 : vec3(0.0);
    result2 = boolMask.y ? result2 : vec3(0.0);
    result3 = boolMask.z ? result3 : vec3(0.0);
    result4 = boolMask.w ? result4 : vec3(0.0);
    numBlurPixels = mask.x + mask.y + mask.z + mask.w;

    // Get max and averaged blur filter results, taking into account that some texels may be null
    vec3 maxFilter = max(result1, max(result2, max(result3, result4)));
    vec3 blurFilter = (result1 + result2 + result3 + result4) / max(1.0, numBlurPixels);
    vec3 maxFilterBokeh = maxFilter;

#if defined(SCENE_BLUR_FAR)
    vec3 currentTexel = texture2DLodCompat(sSourceTex, vTexCoord, 0.0).rgb;
    // If we have failed to receive any filter texels, then output unfiltered texel instead
    // This replaces black artifacts in a few edge cases with incorrectly sharp texels, which are less jarring
    maxFilter = mix(vec3(0.0), maxFilter, step(0.5, numBlurPixels));
    blurFilter = mix(currentTexel, blurFilter, step(0.5, numBlurPixels));
    // Apply bokeh multiplier
    maxFilterBokeh *= uBokehParams.w;
#else
    maxFilterBokeh *= uBokehParams.z;
#endif

    // Create bokeh effect by interpolating with max filter based on max filter values
    // Bokeh produced is square due to Kawase kernel, but this does exist in real lenses
    gl_FragColor.rgb = mix(blurFilter, maxFilterBokeh, smoothstep(uBokehParams.x, uBokehParams.y, maxFilter));
}
