// Dual filtering blur
// See "Bandwidth-Efficient Rendering" by Marius Bjorge (Siggraph 2015)
// This works by basically doing N passes of a progressively downscaled filtered blur, followed by N upscaling passes
// with a slightly different filter in order to reconstruct blurred data for a pixel based on the lower res version of the image.
uniform float uSourceTexSlice;
uniform sampler2D sSourceTex;
uniform vec4 uSamplePixelAndHalfPixelSizes;

in vec2 vTexCoord;

vec3 DualFilteringDownscale(sampler2D tex, vec2 uv, vec2 halfPixelSize)
{
    vec3 sum = texture2DLodCompat(tex, uv, 0.0).rgb * 4.0;

    // Get 4 samples at the corners of the current pixel (could use textureGather here but it wouldn't result in a perf gain since we need data from the 3 channels)
    sum += texture2DLodCompat(tex, uv + halfPixelSize, 0.0).rgb;
    sum += texture2DLodCompat(tex, uv - halfPixelSize, 0.0).rgb;
    sum += texture2DLodCompat(tex, uv + vec2(-halfPixelSize.x, halfPixelSize.y), 0.0).rgb;
    sum += texture2DLodCompat(tex, uv - vec2(-halfPixelSize.x, halfPixelSize.y), 0.0).rgb;

    return sum / 8.0;
}

vec3 DualFilteringUpscale(sampler2D tex, vec2 uv, vec2 pixelSize, vec2 halfPixelSize)
{
    // Get 4 samples at the corners of the current pixel (could use textureGather here but it wouldn't result in a perf gain since we need data from the 3 channels)
    vec3 sum = texture2DLodCompat(tex, uv + halfPixelSize, 0.0).rgb * 2.0;
    sum += texture2DLodCompat(tex, uv - halfPixelSize, 0.0).rgb * 2.0;
    sum += texture2DLodCompat(tex, uv + vec2(-halfPixelSize.x, halfPixelSize.y), 0.0).rgb * 2.0;
    sum += texture2DLodCompat(tex, uv - vec2(-halfPixelSize.x, halfPixelSize.y), 0.0).rgb * 2.0;

    // Get 4 samples at the ends of a + shaped pixel region
    sum += texture2DLodCompat(tex, uv + vec2(0.0, pixelSize.y), 0.0).rgb;
    sum += texture2DLodCompat(tex, uv - vec2(0.0, pixelSize.y), 0.0).rgb;
    sum += texture2DLodCompat(tex, uv + vec2(pixelSize.x, 0.0), 0.0).rgb;
    sum += texture2DLodCompat(tex, uv - vec2(pixelSize.x, 0.0), 0.0).rgb;

    return sum / 12.0;
}

void main()
{
    vec3 rgb;

#ifdef DOWNSAMPLE_PASS
    rgb = DualFilteringDownscale(sSourceTex, vTexCoord, uSamplePixelAndHalfPixelSizes.zw);
#else
    rgb = DualFilteringUpscale(sSourceTex, vTexCoord, uSamplePixelAndHalfPixelSizes.xy, uSamplePixelAndHalfPixelSizes.zw);
#endif

    gl_FragColor = vec4(rgb, 1.0);
}
