#include "Shared/DepthUtils.inc"

#define GENERATE_NORMALS 0

uniform vec4 uParams;
uniform mat4 uSceneProjectionMatrix;
uniform mat4 uSceneInvProjMatrix;
uniform SAMPLER_2D_AUTO_MULTISAMPLE uNormalTexture;
uniform SAMPLER_2D_AUTO_MULTISAMPLE uDepthTexture;
uniform sampler2D uNoiseTexture;

in vec2 vTextureUV;

const float NUM_STEPS = 4.0;
const float NUM_DIRECTIONS = 4.0;


float FallOff(float distanceSquare)
{
    return distanceSquare * uParams.y + 1.0;
}

vec2 RotateDirection(vec2 direction, vec2 CosSin)
{
    return vec2(direction.x * CosSin.x - direction.y * CosSin.y,
                direction.x * CosSin.y + direction.y * CosSin.x);
}

// p = viewspace position at the kernel centre
// n = viewspace normal at the kernel centre
// s = viewspace position of the current sample
float ComputeAO(vec3 p, vec3 n, vec3 s)
{
    vec3 v = s - p;
    float vv = dot(v, v);
    float nv = dot(n, v) / sqrt(vv);
    return clamp(nv - uParams.z, 0.0, 1.0) * clamp(FallOff(vv), 0.0, 1.0);
}

float ComputeCoarseAO(vec2 baseUV, vec2 invViewportUVRes, vec4 uvOffsetScale, vec2 texSize, float radiusPixels, vec4 jitter, vec3 viewPosition, vec3 viewNormal)
{
    float stepSizePixels = radiusPixels / (NUM_STEPS + 1.0);

    const float alpha = 2.0 * PI / NUM_DIRECTIONS;
    float AO = 0.0;

    for (float directionIndex = 0.0; directionIndex < NUM_DIRECTIONS; ++directionIndex)
    {
        float angle = alpha * directionIndex;

        // Compute normalized 2D direction
        vec2 direction = RotateDirection(jitter.xy, vec2(cos(angle), sin(angle)));

        // Jitter starting sample within the first step
        float rayPixels = (jitter.z * stepSizePixels + 1.0);

        for (int stepIndex = 0; stepIndex < int(NUM_STEPS); ++stepIndex)
        {
            vec2 snappedUV = floor(rayPixels * direction + 0.5) * invViewportUVRes + baseUV;
            vec3 S = GetViewSpacePos(snappedUV, uvOffsetScale, texSize, uDepthTexture, uSceneInvProjMatrix);
            rayPixels += stepSizePixels;
            AO += ComputeAO(viewPosition, viewNormal, S);
        }
    }
    AO *= (1.0 / ((1.0 - uParams.z) * NUM_DIRECTIONS * NUM_STEPS));
    return clamp(1.0 - AO * 2.0, 0.0, 1.0);
}

vec3 MinDiff(vec3 p, vec3 pr, vec3 pl)
{
    vec3 v1 = pr - p;
    vec3 v2 = p - pl;
    return (dot(v1,v1) < dot(v2,v2)) ? v1 : v2;
}

vec3 ReconstructNormal(vec2 baseUV, vec2 invViewportUVRes, vec4 uvOffsetScale, vec2 texSize, vec3 p)
{
    vec3 pr = GetViewSpacePos(baseUV + vec2(invViewportUVRes.x, 0.0), uvOffsetScale, texSize, uDepthTexture, uSceneInvProjMatrix);
    vec3 pl = GetViewSpacePos(baseUV + vec2(-invViewportUVRes.x, 0.0), uvOffsetScale, texSize, uDepthTexture, uSceneInvProjMatrix);
    vec3 pt = GetViewSpacePos(baseUV + vec2(0.0, invViewportUVRes.y), uvOffsetScale, texSize, uDepthTexture, uSceneInvProjMatrix);
    vec3 pb = GetViewSpacePos(baseUV + vec2(0.0, -invViewportUVRes.y), uvOffsetScale, texSize, uDepthTexture, uSceneInvProjMatrix);
    return normalize(cross(MinDiff(p, pr, pl), MinDiff(p, pt, pb)));
}

void main()
{
    vec3 viewPosition = GetViewSpacePos(vTextureUV, uViewportOffsetScale, uFullScreenLookupScale.zw, uDepthTexture, uSceneInvProjMatrix);
    vec3 viewNormal;

    #if GENERATE_NORMALS
        viewNormal = -ReconstructNormal(vTextureUV, uViewportLookupScale.xy, uViewportOffsetScale, uFullScreenLookupScale.zw, viewPosition);
    #else
        vec2 screenTexCoords = vTextureUV * uViewportOffsetScale.zw + uViewportOffsetScale.xy;
        viewNormal = texture2DMultisample(uNormalTexture, screenTexCoords, uFullScreenLookupScale.zw).xyz;
        // Skip sky pixels
        if (dot(viewNormal, vec3(1.0)) == 0.0)
        {
            gl_FragColor = vec4(1.0);
            return;
        }
        viewNormal = viewNormal * 2.0 - 1.0;
    #endif

    vec4 jitter = texture2DLodCompat(uNoiseTexture, gl_FragCoord.xy / 4.0, 0.0);

    // Calculate AO
    float radiusToScreen = (uViewportLookupScale.w * uParams.x) / viewPosition.z;
    float AO = ComputeCoarseAO(vTextureUV, uViewportLookupScale.xy, uViewportOffsetScale, uFullScreenLookupScale.zw, radiusToScreen, jitter, viewPosition, viewNormal);
    gl_FragColor = vec4(vec3(pow(AO, uParams.w)), 1.0);
}


