#ifndef HL2BASIS_INC
#define HL2BASIS_INC

// Hl2-basis works by evaluating lighting contributions over a hemisphere facing towards the camera, WITHOUT involving
// normals at that stage of calculation. This means we can cheaply calculate lighting in a vertex shader, and then
// use normals in the pixel shader to approximate per-pixel lighting.
//
// While it's not perfect, it's a very good and cheap approximation for particle lighting, and can often yield more
// natural-looking results than per-pixel lighting since we end up with less harsh shadow cutoffs, specially when
// particles move in front or behind of eachother.

vec3 CalculateHL2BasisContribution(vec3 direction)
{
    const float SQRT_2 = 1.414213562373095;
    const float SQRT_3 = 1.732050807568877;
    const float SQRT_6 = 2.449489742783178;
    const float SQRT_2_DIV_3 = 0.816496580927726;

    const mat3 hl2Basis = mat3(
        vec3(-1.0 / SQRT_2, 1.0 / SQRT_3, 1.0 / SQRT_6),
        vec3(1.0 / SQRT_2, 1.0 / SQRT_3, 1.0 / SQRT_6),
        vec3(0.0, 1.0 / SQRT_3, -SQRT_2_DIV_3)
    );

    return vec3(
        clamp(dot(direction, hl2Basis[0]), 0.0, 1.0),
        clamp(dot(direction, hl2Basis[1]), 0.0, 1.0),
        clamp(dot(direction, hl2Basis[2]), 0.0, 1.0)
    );
}

void AccumulateHL2BasisLighting(inout mat3 lightingData, vec3 lightDir, vec3 lightColour, float lightAttenuation)
{
    vec3 contribution = CalculateHL2BasisContribution(lightDir);

    lightColour *= lightAttenuation;

    lightingData += mat3(
        lightColour * contribution.x,
        lightColour * contribution.y,
        lightColour * contribution.z
    );
}

vec3 CalculateHL2BasisLighting(mat3 lightingData, vec3 normalVS)
{
    vec3 contribution = CalculateHL2BasisContribution(normalVS);

    vec3 lighting =
        lightingData[0] * contribution.x +
        lightingData[1] * contribution.y +
        lightingData[2] * contribution.z;

    return lighting;
}

#endif
