#ifndef LUMINANCE_UTILS_INC
#define LUMINANCE_UTILS_INC

#if __VERSION__ >= 320
#define PRECISE_VARIABLE precise
#define FUSED_MULTIPLY_ADD(a, b, c) fma(a, b, c)
#else
#define PRECISE_VARIABLE
#define FUSED_MULTIPLY_ADD(a, b, c) ((a * b) + c)
#endif

const vec3 LUM_CONVERT_SCALE = vec3(0.2125, 0.7154, 0.0721);    // RGB->Greyscale for modern CRT HDTV phosphor tv/monitors
const vec3 BLUE_SHIFT_SCALE = vec3(1.05, 0.97, 1.27);           // The per-color weighting to be used for blue shift under low light;
const float MIN_VALID_LUMINANCE = 0.00001;                        // The minimum valid luminance allowed

float AutoExposureKeyGrzegorz(float colourLum)
{
    return max(0.1, 1.6 - (1.5 / (colourLum * 0.1 + 1.0)));
}

/**
    Function to automatically calculate the luminance exposure key to be
    used by various tone mapping operators.
 */

float CalculateLuminance(vec3 colour)
{
   // 1e-5 necessary to stop the singularity at lum == 0
   return max(dot(colour, LUM_CONVERT_SCALE) + MIN_VALID_LUMINANCE, MIN_VALID_LUMINANCE);
}

float CalculateLuminanceHSP(vec3 colour)
{
   // This method returns a brightness that is perceptually closer to the input colour than the standard luminance approach
   // This allows for bloom pass brightpass cut-offs to apply more evenly to RGB channels, instead of heavily favouring greens
   // See http://alienryderflex.com/hsp.html
   // 1e-5 necessary to stop the singularity at lum == 0
    return max(sqrt(dot((colour * colour), vec3(0.299, 0.587, 0.114))) + MIN_VALID_LUMINANCE, MIN_VALID_LUMINANCE);
}

/**
 * Calculate the linear attenuation between a and b based on attenuated speed t
 * See https://www.nde-ed.org/Physics/X-Ray/attenuationCoef.xhtml
 */
float LinearAttenuate(float a, float b, float t)
{
    float interpT = 1.0 - exp(-t);
    return mix(a, b, clamp(interpT, 0.0, 1.0));
}

/**
 * Calculate the linear attenuation between a and b based on attenuated speed t
 * using an approximation of the calculation. Note that this approximation only
 * is valid if t is [0.0, 1.0], with especially close results if t is [0.0, 0.5]
 */
float LinearAttenuateApprox(float a, float b, float t)
{
    // 0.38 was chosen for having the closest approximation to `1.0 - exp(-t)`
    // See https://graphtoy.com/?f1(x,t)=1.0-exp(-x)&v1=true&f2(x,t)=x-((x*x)*0.5)&v2=true&f3(x,t)=&v3=true&f4(x,t)=&v4=true&f5(x,t)=&v5=false&f6(x,t)=&v6=true&grid=true&coords=0.4636034846918319,0.3235597688000768,0.8321201939329607
    float interpT = t - ((t * t) * 0.38);
    return mix(a, b, clamp(interpT, 0.0, 1.0));
}

float CalculateRodsConesTau(float luminance)
{
    // tau interpolates between adaptation rates of cones and rods 0.4 for rods and 0.1 for cones
    // The current average lum is used as the interpolation coef
    float rho = 0.04 / (0.04 + luminance); // Rod sensitivity model
    // Adaptation constants for rods and cones
    const float rods = 0.2;
    const float cones = 0.4;
    return rho * rods + (1.0 - rho) * cones;
}

/**
    An exponential decay function for calculating the adaptive luminance to
    simulate the gradual adaptation of the human eye to different lighting conditions.
    Akin to the the auto-exposure feature of a camera. The adaptive luminance chases
    the average luminance.

    Based on Pattanaik's technique

    @param colourLum The luminance of the current pixel
    @param timeDelta Elapsed time in seconds since last rendered frame
 */
float CalcAdaptedLumRodsCones(float currAvgLum, float prevAdaptedLum, float timeDelta, float exposureSpeed)
{
    float tau = CalculateRodsConesTau(currAvgLum);
    float adaptedLum = LinearAttenuate(prevAdaptedLum, currAvgLum, timeDelta / (tau * exposureSpeed));
    return max(adaptedLum, MIN_VALID_LUMINANCE);
}

/**
    An exponential decay function for calculating the adaptive luminance to
    simulate the gradual adaptation of the human eye to different lighting conditions.
    Akin to the the auto-exposure feature of a camera. The adaptive luminance chases
    the average luminance.

    Based on Pattanaik's technique

    Uses the approximated linear attenuation function

    @param colourLum The luminance of the current pixel
    @param timeDelta Elapsed time in seconds since last rendered frame
 */
float CalcAdaptedLumRodsConesApprox(float currAvgLum, float prevAdaptedLum, float timeDelta, float exposureSpeed)
{
    float tau = CalculateRodsConesTau(currAvgLum);
    float adaptedLum = LinearAttenuateApprox(prevAdaptedLum, currAvgLum, timeDelta / (tau * exposureSpeed));
    return max(adaptedLum, MIN_VALID_LUMINANCE);
}

/**
    An exponential decay function for calculating the adaptive luminance to
    simulate the gradual adaptation of the human eye to different lighting conditions.
    Akin to the the auto-exposure feature of a camera. The adaptive luminance chases
    the average luminance.

    Based on Pattanaik's technique

    Uses the approximated linear attenuation function and precise keyword

    @param colourLum The luminance of the current pixel
    @param timeDelta Elapsed time in seconds since last rendered frame
 */
//[
float CalcAdaptedLumRodsConesApproxPrecise(float currAvgLum, float prevAdaptedLum, float timeDelta, float exposureSpeed)
{
    // NOTE: We are very liberally using the `precise` keyword here
    // This is because these calculations _appear_ to be going wrong
    // on PowerVR devices and cause precision loss, leading to very
    // dark scenes due to miscalculated luminance values

    const float rods = 0.2;
    const float cones = 0.4;

    PRECISE_VARIABLE float rho = 0.04 / (0.04 + currAvgLum);
    PRECISE_VARIABLE float coneEval = (1.0 - rho) * cones;
    PRECISE_VARIABLE float tau = FUSED_MULTIPLY_ADD(rho, rods, coneEval);

    PRECISE_VARIABLE float t = -timeDelta / (tau * exposureSpeed);
    PRECISE_VARIABLE float interpT = t - ((t * t) * 0.5);
    PRECISE_VARIABLE float adaptedLum = mix(currAvgLum, prevAdaptedLum, clamp(interpT, 0.0, 1.0));

    return max(adaptedLum, MIN_VALID_LUMINANCE);
}
//]

/**
    Simplified adaptive luminance calculation taken from the Microsoft HDR Lighting sample
 */

float CalcAdaptedLumMS(float currAvgLum, float prevAdaptedLum, float timeDelta, float exposureSpeed)
{
    // The user's adapted luminance level is simulated by closing the gap between
    // adapted luminance and current luminance by 2% every frame, based on a
    // 30 fps rate. This is not an accurate model of human adaptation, which can
    // take longer than half an hour.

    float tau = 50.0 * exposureSpeed;
    float adaptedLum = (prevAdaptedLum + (currAvgLum - prevAdaptedLum) * (1.0 - pow(0.98, tau * timeDelta)));
    return max(adaptedLum, MIN_VALID_LUMINANCE);
}

#endif
