#include "ConversionUtils.inc"
#include "Shared/Caustics.inc"
#include "Shared/DistanceFog/Uniforms.inc"
#include "Shared/DistanceFog/Functions.inc"
#include "Shared/ModelGeometryUtils.inc"
#include "Shared/ModelGeometryFragmentUtils.inc"
#include "Shared/NormalMap.inc"
#include "Shared/TextureCompression.inc"
#include "Shared/ViewportMap/ViewportMap.inc"
#include "Shared/VolumetricScattering/VolumetricFunctions.inc"
#include "Shared/VolumetricScattering/ApplyVolumetrics.inc"
#include "Shared/LightingUtils.inc"
#include "Shared/StippleTransparencyUtils.inc"
#include "Shared/StippleCutoutUtils.inc"
#include "lighting/brdf/Fresnel.inc"
#include "lighting/Lighting.inc"

void main()
{
#if defined(DEBUG_VERTEX_BONE_COLOUR) // We want to output the vertex colour only when debugging verticies
    gl_FragColor = vec4(vVertexAlbedo);

    if (uDebugReturn != 0.0)
    {
        return;
    }
#endif

#if defined(VERTEX_COLOUR) && defined(ALPHA_ENABLED)
    // Vertex alpha test (as early as possible to avoid unnecessary shader instructions being executed)
    if (vVertexAlbedo.a == 0.0)
    {
        discard;
        return;
    }
#endif

    highp vec4 worldPos = vec4(vWorldPosition.xyz, 1.0);

#if defined(CLIP_PLANE) && !defined(PUSH_TO_FARPLANE)
    if (dot(worldPos, uClipPlane) < 0.0)
    {
        //[
        discard;
        return;
        //]
    }
#endif

#if defined(TEXTURE_ATLAS)
    TextureSettings textureSettings;
    getTextureSettings(vMaterialSettingsSlotXY_BatchFlags.xy, textureSettings);
#endif

    highp vec2 uvDX = vec2(0.0);
    highp vec2 uvDY = vec2(0.0);
    highp vec3 positionDX = vec3(0.0);
    highp vec3 positionDY = vec3(0.0);
    vec4 texelAlbedo = vec4(1.0); //NOTE: The alpha of texelAlbedo has multiple purposes, see TEXTURE_ALPHA_USAGE below
    vec4 texelHDRNormal;
#if defined(COMPRESSED_NORMALS)
    texelHDRNormal = vec4(0.0, 0.5, 0.0, 0.5); // 0 HDR, no-op normal (Y, Z (Unused), X)
#else
    texelHDRNormal = vec4(0.0, 0.5, 1.0, 0.5); // 0 HDR, no-op normal (Y, Z, X)
#endif
    //NOTE: texelHDRNormal's normal value will be replaced if we read from a texture, and must be decoded with UnpackNormal further down

    highp vec3 viewVector = worldPos.xyz - uCameraPosition;
#if defined(TEXTURE_ATLAS)
    // Copy to local highp to workaround precision bug
    highp float textureAnimationTime = uTextureAnimationTime;
    vec2 textureUV = vTextureUV + fract(textureSettings.uvAnim * textureAnimationTime);

    CalculateDerivatives(uvDX, uvDY, positionDX, positionDY, textureUV, viewVector);
    SampleTextures(texelAlbedo, texelHDRNormal, textureUV, textureSettings, vMaterialSettingsSlotXY_BatchFlags.z, uvDX, uvDY);
#endif

    float specularGlossMask = 1.0;
    float emissiveLevel = 0.0;
    float envMappingLevel = 0.0;

#if defined(TEXTURE_ALPHA_USAGE)
    //What should the texture alpha channel be used for
    //x = Specular, y = Emissive, z = Env Mapping
    //These flag values are either 0.0 or 1.0
    vec3 alphaUsage = step(0.5, fract(vMaterialSettingsSlotXY_BatchFlags.z * vec3(64.0, 32.0, 16.0)));
    // Check forced texture flag if textures are off globally, and if we don't have textures then mask out the alphaUsage that relies on alpha channel
    #if !defined(TEXTURE_ALBEDO_GLOBAL)
        alphaUsage.xz *= step(0.5, fract(vMaterialSettingsSlotXY_BatchFlags.z * 8.0));
    #endif
    // NOTE: Scale the specular strength up to compensate for multiplying the albedo lighting into the specular term below
    specularGlossMask += alphaUsage.x * texelAlbedo.a * 4.0;
    // NOTE: Emissive doesn't actually use the alpha channel, that's
    // still used as normal for alpha blending
    emissiveLevel = alphaUsage.y;
    envMappingLevel = alphaUsage.z * texelAlbedo.a;
    //Replace the alpha if we need to (only for specular and env map)
    texelAlbedo.a = min(texelAlbedo.a + alphaUsage.x + alphaUsage.z, 1.0);
#endif

    vec4 albedo = vec4(0.0, 0.0, 0.0, 1.0);
#if defined(VERTEX_COLOUR)
    albedo = texelAlbedo * vVertexAlbedo;
#else
    albedo = texelAlbedo;
#endif

#if defined(DEBUG_GEOMETRY_INSTANCE_COLOUR)
    gl_FragColor = vec4(uDebugInstanceColour.xyz, albedo.a);

    if (uDebugReturn != 0.0)
    {
        return;
    }
#endif

#if defined(TEXTURE_ALPHA_USAGE) && defined(ALPHA_ENABLED)
    // Alpha enabled is also checked for here so that this discard should only test for non-opaque alpha usage
    // Alpha test (as early as possible to avoid unnecessary shader instructions being executed)
    // Throw away any fragments which are below the set threshold
    if (albedo.a <= uAlphaTestThreshold)
    {
        discard;
        return;
    }
#endif

#if defined(STIPPLE_TRANSPARENCY_CUTOUT) && defined(VIEWPORTLOOKUPSCALE)

#if defined(ALPHA_ENABLED)
    // Handle alpha-tested separately since for those we have to fall back to stippling instead of taking advantage of alpha blending
    // Note: The threshold cannot be 0.0 as even alpha-blended have a very small alpha threshold instead of having a 0 threshold.
    if (uAlphaTestThreshold > 0.01)
    {
        if (IsStippleCutoutVisible(viewVector, gl_FragCoord.xy * uViewportLookupScale.xy, gl_FragCoord.xy))
        {
            discard;
            return;
        }
    }
    else
    {
        albedo.a *= 1.f - GetStippleCutoutAdjustedAlpha(viewVector, gl_FragCoord.xy * uViewportLookupScale.xy, 1.0);
    }
#else
    if (IsStippleCutoutVisible(viewVector, gl_FragCoord.xy * uViewportLookupScale.xy, gl_FragCoord.xy))
    {
        discard;
        return;
    }
#endif

#endif

#if defined(STIPPLE_TRANSPARENCY_ENABLED)
    if (!IsStipplePixelVisible(viewVector, gl_FragCoord.xy, albedo.a))
    {
        discard;
        return;
    }
#endif

#if defined(GOURAUD_SHADING) || defined(UNLIT_SHADING)
    gl_FragColor = albedo;

#if defined(DISCRETIZED_LUMINANCE)
    if (vMaterialSettingsSlotXY_BatchFlags.xy == vec2(0.0))
    {
        gl_FragColor.rgb = DiscretizeLuminance(gl_FragColor.rgb);
    }
#endif

#if defined(LIGHT_SCATTERING) || defined(FOG_DISTANCE)
    gl_FragColor.rgb = ApplyInOutScattering(gl_FragColor.rgb, vOutScattering, vInScattering);
#endif

    //[
    return;
    //]
#endif

    highp vec3 worldNormal = normalize(vNormal);
    highp vec3 tangent = vec3(0.0, 1.0, 0.0);
    highp vec3 bitangent = vec3(0.0, 1.0, 0.0);
    // The bitangent and tangent are needed for some BRDFs, therefore we always need to compute them
#if defined(PER_FRAGMENT_TANGENTS)
    ComputeTangentBitangentFromDerivatives(tangent, bitangent, worldNormal, positionDX, positionDY, uvDX, uvDY);
#endif

    // Normal map
#if defined(USE_NORMAL_MAP)

#if !defined(PER_FRAGMENT_TANGENTS)
    tangent = normalize(vTangent.xyz);
    bitangent = ComputeBitangent(worldNormal, vTangent);
#endif

    vec3 textureNormal = UnpackNormal(texelHDRNormal.gba, textureSettings.normalScale);
    worldNormal = ApplyNormalMap(textureNormal, worldNormal, tangent, bitangent);
#endif

// Return if using these debug types
#if defined(DEBUG_TANGENTS) || defined(DEBUG_BITANGENTS)

#if defined(DEBUG_TANGENTS)
    gl_FragColor.rgb = (normalize(tangent) * 0.5 + 0.5);
    gl_FragColor.a = 1.0;
#endif
#if defined(DEBUG_BITANGENTS)
    gl_FragColor.rgb = (normalize(bitangent) * 0.5 + 0.5);
    gl_FragColor.a = 1.0;
#endif

    if (uDebugReturn != 0.0)
    {
        return;
    }
#endif

#if defined(VIEWPORTMAP)
    albedo.rgb = SampleViewportMapColour(gl_FragCoord.xy, textureSettings.viewportMapUVScaleAndAnim);
#endif

#if defined(DEBUG_TEXEL_DENSITY)
    #if defined(TEXTURE_ATLAS)
        gl_FragColor = vec4(GetTexelDensityDebugColour(vTextureUV, textureSettings.textureMeta1.z, viewVector), 1.0);
    #else
        gl_FragColor = vec4(1.0);
    #endif

    if (uDebugReturn != 0.0)
    {
        return;
    }
#endif

#if defined(VIEWPORTLOOKUPSCALE)
    vec2 screenTexCoords = gl_FragCoord.xy * uFullScreenLookupScale.xy;
#endif

    // Ambient occlusion
    float ambientOcclusion = 1.0;
#if defined(SSAO) && !defined(REFRACTION)
    ambientOcclusion = texture2D(uSSAOMap, screenTexCoords).r;
#endif

    LightingTerms indirectLighting;
    ClearLightingTerms(indirectLighting);

    // Diffuse irradiance
    indirectLighting.Diffuse = uAmbientColour;

    vec3 irradiance = vec3(1.0, 1.0, 1.0);
#if defined(IRRADIANCE_LIGHTING)
    irradiance = EvaluateSHLighting2ndOrder(worldNormal, uIrradianceSHCoefs);
    indirectLighting.Diffuse *= irradiance;
#endif
    // Apply ambient occlusion to indirect diffuse lighting
#if defined(SSAO)
    indirectLighting.Diffuse *= ambientOcclusion;
#endif

    highp float cameraDistance = length(viewVector);
    // Normalize view direction reusing length value
    highp vec3 viewDirection = viewVector / cameraDistance;

    LightingTerms directLighting;
    ClearLightingTerms(directLighting);

    // Direct sunlight lighting
#if defined(SUNLIGHT_DIRECT_LIGHTING)
    int iCascade = -1;
    // Sunlight shadow attenuation
    float amountInSunlight = 0.0;

    #if defined(DEFERRED_SHADOWS)
        directLighting = EvaluateSunlightRT5(amountInSunlight, worldNormal, bitangent, -viewDirection, screenTexCoords, textureSettings.specular, emissiveLevel);
    #else
        // Compute depth in camera space
        // Camera plane distance more accurate and must be used for cascade interval select to work without artifacts
        highp vec3 camLook = vec3(uViewMatrix[0][2], uViewMatrix[1][2], uViewMatrix[2][2]);
        highp float viewSpaceDepth = abs(dot(viewVector, camLook));
        directLighting = EvaluateSunlightRT5(iCascade, amountInSunlight, worldPos, worldNormal, bitangent, viewSpaceDepth, -viewDirection, textureSettings.specular, emissiveLevel);
    #endif
#else
    directLighting.Diffuse = vec3(1.0);
#endif

#if defined(TEXTURE_ATLAS) && defined(DEBUG_MATERIAL_HIGHLIGHT)
    if (uDebugMaterialHighlight != -1.0)
    {
        float lightStrength = mix(0.1, 0.5, length(directLighting.Diffuse));
        float base = 0.1;
        vec3 col = mix(vec3(base) * lightStrength, vec3(0,1,0), textureSettings.materialID == uDebugMaterialHighlight ? 1 : 0);
        gl_FragColor = vec4(col, 1);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    }
#endif

    // @todo Put env mapping code into a standalone function
    // Note GLOBAL_ENVIRONMENTMAPPING is unconditional (hardware capabilities only)
    float envFresnel = FresnelSchlick(0.8, max(0.0, dot(-viewDirection, worldNormal)), 5.0);
#if defined(TEXTURE_ATLAS) && defined(TEXTURE_ALPHA_USAGE) && defined(GLOBAL_ENVIRONMENTMAPPING)
    if (envMappingLevel > 0.0)
    {
        // Previous version would end up with a backwards environment mapping, wrong side and not reflected.
        // So now we get the inverse reflection, and spin around the x and y to put the texture the right way up
        vec3 envMapDirection = reflect(-viewDirection, worldNormal);
        envMapDirection.x = -envMapDirection.x;
        envMapDirection.y = -envMapDirection.y;
        indirectLighting.Specular = textureCubeSRGB(uGlobalEnvironmentMap, envMapDirection).rgb * ambientOcclusion;

    #if defined(NORMALIZED_ENVIRONMENTMAPPING)
        // Scale/modulate the environment specular by the incoming irradiance to try and normalise the environment map more to the local environment lighting
        // This would be more accurate if we had local light probes with normalised environment maps.
        indirectLighting.Specular *= (irradiance * uGlobalEnvironmentMappingParams.w);
    #endif

        albedo.rgb = mix(albedo.rgb, indirectLighting.Specular, envMappingLevel * envFresnel);
    }
#endif

    vec4 fragColour = vec4(0.0, 0.0, 0.0, albedo.a);

    // Point lights
#if defined(POINT_LIGHTING)
    vec3 pointLightsTerm = vec3(0.0, 0.0, 0.0);
    vec3 pointLightsSpecular = vec3(0.0, 0.0, 0.0);
    const vec3 specReflDefault = vec3(0.65, 0.65, 0.65);
    const float perceptualRoughnessDefault = 1.0;
    const float fresnelExpDefault = 5.0;
    #if defined(POINT_LIGHTING_VERTEX_DATA)
        EvaluatePointLights(pointLightsTerm, pointLightsSpecular, specReflDefault, textureSettings.specular, perceptualRoughnessDefault, fresnelExpDefault, -viewDirection, vWorldPosition.xyz, worldNormal, viewSpaceDepth, vTilePosition);
    #else
        EvaluatePointLights(pointLightsTerm, pointLightsSpecular, specReflDefault, textureSettings.specular, perceptualRoughnessDefault, fresnelExpDefault, -viewDirection, vWorldPosition.xyz, worldNormal, viewSpaceDepth);
    #endif

    #if defined(DEBUG_POINTLIGHTS)
        gl_FragColor = vec4(pointLightsTerm, albedo.a);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DEBUG_POINTLIGHTS_SPECULAR)
        gl_FragColor = vec4(pointLightsSpecular, albedo.a);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DIFFUSE_LIGHTING)
        directLighting.Diffuse += pointLightsTerm;
    #else
        directLighting.Diffuse = pointLightsTerm;
    #endif

    #if defined(POINT_LIGHTING_SPECULAR)
        directLighting.Specular += pointLightsSpecular;
    #endif
#else
    #if defined(DEBUG_POINTLIGHTS)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DEBUG_POINTLIGHTS_SPECULAR)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if !defined(DIFFUSE_LIGHTING)
        directLighting.Diffuse = vec3(0.0);
    #endif
#endif

    float causticsTerm = 0.0;
#if defined(CAUSTICS)
    causticsTerm = CalculateCausticsTerm(vWorldPosition.xyz, amountInSunlight, worldNormal);
#endif

    //Final compositing of lighting terms...

    // Ambient
#if defined(AMBIENT_LIGHTING)
    fragColour.rgb += indirectLighting.Diffuse;
    //fragColour.rgb += indirectLighting.Specular * specularGlossMask;
#endif
    // Diffuse
    fragColour.rgb += directLighting.Diffuse;
    // Specular
#if defined(SPECULAR_LIGHTING)
    fragColour.rgb += directLighting.Specular * specularGlossMask;
#endif
    // Caustics
#if defined(CAUSTICS)
    fragColour.rgb += uSunColour * causticsTerm;
#endif
    // Albedo
#if defined(ALBEDO_LIGHTING)
    fragColour.rgb *= albedo.rgb;
#endif
    //Emissive
#if defined(TEXTURE_ALPHA_USAGE)
    #if defined(DEBUG_EMISSIVE_MAP)
        fragColour.rgb = vec3(emissiveLevel);
    #else
        if(emissiveLevel > 0.0) // Compensate for tone map brightening by making emissive objects fainter
        {
            fragColour.a *= fragColour.a;
        }
        fragColour.rgb = mix(fragColour.rgb, albedo.rgb, emissiveLevel);
    #endif
#else
    #if defined(DEBUG_EMISSIVE_MAP)
        fragColour.rgb = vec3(0.0);
    #endif
#endif

#if defined(REFRACTION)
    if (textureSettings.refraction.x > 0.0 || textureSettings.refraction.z > 0.0)
    {
        fragColour.rgb = CalculateRefractionColour(vWorldPosition.xyz, worldNormal, -viewDirection, textureSettings.specular, textureSettings.refraction, worldNormal.xz, fragColour);
    }
#endif

    //Light scattering (now includes fog!)
#if defined(LIGHT_SCATTERING) || defined(FOG_DISTANCE)
    fragColour.rgb = ApplyInOutScattering(fragColour.rgb, vOutScattering, vInScattering);
#endif

#if defined(VOLUMETRIC_SCATTERING_SUPPORTED)
    vec4 scattering = GetScatteredInRay2(8, viewDirection, cameraDistance, CalculateScatteringOffset(gl_FragCoord.xy));
    fragColour = ApplyVolumetricScattering(fragColour, scattering);
#endif

#if defined(TINT)
    #if defined (TINT_SHIFT_HSL)
        int weightFlags = floatBitsToInt(uTint.w);
        bvec3 hslMask = bvec3((ivec3(weightFlags) >> ivec3(8, 9, 10)) & ivec3(1));
        vec3 baseHsl = oldscapeConvertRGBtoHSL(fragColour.rgb);
        fragColour.rgb = oldscapeConvertHSLtoRGB(mix(baseHsl, mix(baseHsl, uTint.xyz, float(weightFlags & 0xff) / 128.0f), hslMask));
    #elif defined(PUSH_TO_FARPLANE) //Overbright for the sky model (used in irradiance)
        fragColour.rgb += uTint.rgb;
    #endif
#endif

    //Debug lighting
#if defined(DEBUG_ALBEDO)
    fragColour = albedo;
#endif
#if defined(DEBUG_NORMALS)
    fragColour.rgb = (worldNormal * 0.5 + 0.5);
    fragColour.a = 1.0;
#endif
#if defined(DEBUG_FRESNEL)
    fragColour.rgb = vec3(envFresnel, envFresnel, envFresnel);
    fragColour.a = 1.0;
#endif
    //Debug specular map
#if defined(DEBUG_SPECULAR_MAP)
    //Remove the default specular amount (0.5) and divide by the specular scale we applied earlier (4.0)
    //This outputs the texture-supplied specular mask
    fragColour.rgb = vec3(max(0.0, (specularGlossMask - 0.5) / 4.0));
#endif
    // Debug shadows
#if defined(SUNLIGHT_SHADOWS) && defined(DEBUG_SUNLIGHT_SHADOW_CASCADE) && !defined(DEFERRED_SHADOWS)
    fragColour.rgb = ShadowMapCascadeColour(iCascade, int(uMappingParams.w)).rgb;
#endif
    // RT7 debug, not in use so display magenta
#if defined(DEBUG_RT7_EMISSIVE) || defined(DEBUG_RT7_METALNESS) || defined(DEBUG_RT7_ROUGHNESS)
    fragColour = vec4(1.0, 0.0, 1.0, 1.0);
#endif

#if defined(FORCE_OPAQUE)
    fragColour.a = 1.0;
#endif

#if defined(DISCRETIZED_LUMINANCE)
    if (vMaterialSettingsSlotXY_BatchFlags.xy == vec2(0.0))
    {
        fragColour.rgb =  DiscretizeLuminance(fragColour.rgb);
    }
#endif

#if defined(PREMULTIPLY_ALPHA)
    fragColour.rgb *= fragColour.a;
#endif

    gl_FragColor = fragColour;
}
