#include "ConversionUtils.inc"
#include "Shared/ModelGeometryUtils.inc"
#include "Shared/StippleTransparencyUtils.inc"
#include "Shared/StippleCutoutUtils.inc"

#if defined(TEXTURE_ATLAS) && (defined(ALPHA_ENABLED) || defined(NORMAL_MAP))
    #include "Shared/ModelGeometryFragmentUtils.inc"
    #include "Shared/NormalMap.inc"
    #include "Shared/TextureCompression.inc"
#endif

void main()
{
    vec3 normal = normalize(vNormal);

#if defined(VERTEX_COLOUR) && defined(ALPHA_ENABLED)
    // Vertex alpha test (as early as possible to avoid unnecessary shader instructions being executed)
    if (vVertexAlbedo.a == 0.0)
    {
        discard;
        return;
    }
#endif

    highp vec3 viewVector = vWorldPosition.xyz - uCameraPosition;

    highp vec2 uvDX = vec2(0.0);
    highp vec2 uvDY = vec2(0.0);
    highp vec3 positionDX = vec3(0.0);
    highp vec3 positionDY = vec3(0.0);
    vec4 texelAlbedo = vec4(1.0); //NOTE: The alpha of texelAlbedo has multiple purposes, see TEXTURE_ALPHA_USAGE below
    vec4 texelHDRNormal;
#if defined(COMPRESSED_NORMALS)
    texelHDRNormal = vec4(0.0, 0.5, 0.0, 0.5); // 0 HDR (R), no-op normal (GBA = Y, Z (Unused), X)
#else
    texelHDRNormal = vec4(0.0, 0.5, 1.0, 0.5); // 0 HDR (R), no-op normal (GBA = Y, Z, X)
#endif
    //NOTE: texelHDRNormal's normal value will be replaced if we read from a texture, and must be decoded with UnpackNormal further down

#if defined(TEXTURE_ATLAS) && (defined(ALPHA_ENABLED) || defined(NORMAL_MAP))
    TextureSettings textureSettings;
    getTextureSettings(vMaterialSettingsSlotXY_BatchFlags.xy, textureSettings);

    // Copy to local highp to workaround precision bug
    highp float textureAnimationTime = uTextureAnimationTime;
    vec2 textureUV = vTextureUV + fract(textureSettings.uvAnim * textureAnimationTime);

    CalculateDerivatives(uvDX, uvDY, positionDX, positionDY, vTextureUV, viewVector);
    SampleTextures(texelAlbedo, texelHDRNormal, textureUV, textureSettings, vMaterialSettingsSlotXY_BatchFlags.z, uvDX, uvDY);
#endif

#if defined(ALPHA_ENABLED)
    // x = Specular, y = Env map
    vec2 alphaUsage = step(0.5, fract(vMaterialSettingsSlotXY_BatchFlags.z * vec2(64.0, 16.0)));
    // Replace alpha if this is a specular or env map texture or if alpha is opaque
    float alpha = min(texelAlbedo.a + alphaUsage.x + alphaUsage.y, 1.0);

    #if defined(VERTEX_COLOUR)
        alpha *= vVertexAlbedo.a;
    #endif

    //Throw away any fragments which are below the set threshold
    if (alpha <= uAlphaTestThreshold)
    {
        discard;
        return;
    }
#endif

#if defined(STIPPLE_TRANSPARENCY_CUTOUT) && defined(VIEWPORTLOOKUPSCALE)
    if (IsStippleCutoutVisible(viewVector, gl_FragCoord.xy * uViewportLookupScale.xy, gl_FragCoord.xy))
    {
        discard;
        return;
    }
#endif

#if defined(STIPPLE_TRANSPARENCY_ENABLED)
    {
        float srcAlpha;

    #if defined(ALPHA_ENABLED)
        srcAlpha = alpha;
    #else
        srcAlpha = texelAlbedo.a;
    #endif

        if (!IsStipplePixelVisible(viewVector, gl_FragCoord.xy, srcAlpha))
        {
            discard;
            return;
        }
    }
#endif

#if defined(USE_NORMAL_MAP)
    vec3 normalMap = UnpackNormal(texelHDRNormal, textureSettings.normalScale);
    #if defined(PER_FRAGMENT_TANGENTS)
        normal = ApplyNormalMap(normalMap, normal, positionDX, positionDY, uvDX, uvDY);
    #else
        normal = ApplyNormalMap(normalMap, normal, vTangent);
    #endif
    normal = normalize(Mat4ToMat3(uViewMatrix) * normal);
#endif

    gl_FragColor = vec4(normal * 0.5 + 0.5, 1.0);

#if defined(FACE_NORMAL)
    vec3 faceNormal = normalize(Mat4ToMat3(uViewMatrix) * cross(dFdy(vWorldPosition), dFdx(vWorldPosition)));
    fFaceNormal = vec4(faceNormal * 0.5 + 0.5, 1.0);
#endif
}
