#include "ConversionUtils.inc"
#include "Shared/Caustics.inc"
#include "Shared/ModelGeometryUtils.inc"
#include "Shared/ModelGeometryFragmentUtils.inc"
#include "Shared/DistanceFog/Uniforms.inc"
#include "Shared/DistanceFog/Functions.inc"
#include "Shared/EnvMapping.inc"
#include "Shared/NormalMap.inc"
#include "Shared/TextureCompression.inc"
#include "Shared/ViewportMap/ViewportMap.inc"
#include "Shared/VolumetricScattering/VolumetricFunctions.inc"
#include "Shared/VolumetricScattering/ApplyVolumetrics.inc"
#include "Shared/LightingUtils.inc"
#include "Shared/StippleTransparencyUtils.inc"
#include "Shared/StippleCutoutUtils.inc"
#include "lighting/brdf/Fresnel.inc"
#include "lighting/Lighting.inc"

void main()
{
#if defined(DEBUG_VERTEX_BONE_COLOUR) // We want to output the vertex colour only when debugging verticies
    gl_FragColor = vec4(vVertexAlbedo);

    if (uDebugReturn != 0.0)
    {
        return;
    }
#endif

    highp vec4 worldPos = vec4(vWorldPosition.xyz, 1.0);

#if defined(CLIP_PLANE) && !defined(PUSH_TO_FARPLANE)
    if (dot(worldPos, uClipPlane) < 0.0)
    {
    //[
        discard;
        return;
    //]
    }
#endif

#if defined(TEXTURE_ATLAS)
    TextureSettings textureSettings;
    getTextureSettings(vMaterialSettingsSlotXY_BatchFlags.xy, textureSettings);
#endif

    highp vec2 uvDX = vec2(0.0);
    highp vec2 uvDY = vec2(0.0);
    highp vec3 positionDX = vec3(0.0);
    highp vec3 positionDY = vec3(0.0);
    vec4 albedo = vec4(1.0);
    vec4 texelEmissiveNormal;
    // Default normal value for texelEmissiveNormal needs to be swizzled as it will be unpacked later
#if defined(COMPRESSED_NORMALS)
    texelEmissiveNormal = vec4(0.0, 0.5, 0.0, 0.5); // 0 emissive (R), no-op normal (GBA = Y, Z (Unused), X)
#else
    texelEmissiveNormal = vec4(0.0, 0.5, 1.0, 0.5); // 0 emissive (R), no-op normal (GBA = Y, Z, X)
#endif
    vec4 texelMetalnessRoughness = vec4(0.0, 0.9, 0.0, 0.0); //0 metalness (R) and 0.9 roughness (G)

    highp vec3 viewVector = worldPos.xyz - uCameraPosition;

#if defined(TEXTURE_ATLAS)
    // Copy to local highp to workaround precision bug
    highp float textureAnimationTime = uTextureAnimationTime;
    vec2 textureUV = vTextureUV + fract(textureSettings.uvAnim * textureAnimationTime);

    CalculateDerivatives(uvDX, uvDY, positionDX, positionDY, textureUV, viewVector);
    SampleTexturesRT7(albedo, texelEmissiveNormal, texelMetalnessRoughness, textureUV, textureSettings, vMaterialSettingsSlotXY_BatchFlags.z, uvDX, uvDY);
#endif

    // Base colour albedo, used for both diffuse and specular colour for our metalness PBR wordflow
#if defined(VERTEX_COLOUR)
    albedo = albedo * vVertexAlbedo;
#endif

#if defined(DEBUG_GEOMETRY_INSTANCE_COLOUR)
    gl_FragColor = vec4(uDebugInstanceColour.xyz, albedo.a);

    if (uDebugReturn != 0.0)
    {
        return;
    }
#endif

#if defined(DEBUG_EMISSIVE_SCALE) && defined(TEXTURE_ATLAS)
    textureSettings.emissiveScale = uDebugEmissiveScale;
    texelEmissiveNormal.r = clamp(texelEmissiveNormal.r * uDebugEmissiveMaskMultiplier, 0.0, 1.0);
#endif

#if defined(TEXTURE_ALPHA_USAGE) || defined(VERTEX_COLOUR)
    // Alpha test (as early as possible to avoid unnecessary shader instructions being executed)
    // Throw away any fragments which are below the set threshold
    if (albedo.a <= uAlphaTestThreshold)
    {
        discard;
        return;
    }
#endif

#if defined(STIPPLE_TRANSPARENCY_CUTOUT) && defined(VIEWPORTLOOKUPSCALE)

#if defined(ALPHA_ENABLED)
    // Handle alpha-tested separately since for those we have to fall back to stippling instead of taking advantage of alpha blending
    // Note: The threshold cannot be 0.0 as even alpha-blended have a very small alpha threshold instead of having a 0 threshold.
    if (uAlphaTestThreshold > 0.01)
    {
        if (IsStippleCutoutVisible(viewVector, gl_FragCoord.xy * uViewportLookupScale.xy, gl_FragCoord.xy))
        {
            discard;
            return;
        }
    }
    else
    {
        albedo.a *= 1.f - GetStippleCutoutAdjustedAlpha(viewVector, gl_FragCoord.xy * uViewportLookupScale.xy, 1.0);
    }
#else
    if (IsStippleCutoutVisible(viewVector, gl_FragCoord.xy * uViewportLookupScale.xy, gl_FragCoord.xy))
    {
        discard;
        return;
    }
#endif

#endif

#if defined(STIPPLE_TRANSPARENCY_ENABLED)
    if (!IsStipplePixelVisible(viewVector, gl_FragCoord.xy, albedo.a))
    {
        discard;
        return;
    }
#endif

#if defined(GOURAUD_SHADING)
    gl_FragColor = albedo;

#if defined(DISCRETIZED_LUMINANCE)
    gl_FragColor.rgb = DiscretizeLuminance(gl_FragColor.rgb);
#endif

    //[
    return;
    //]
#endif

#if defined(DEBUG_TEXEL_DENSITY)
    #if defined(TEXTURE_ATLAS)
        gl_FragColor = vec4(GetTexelDensityDebugColour(vTextureUV, textureSettings.textureMeta1.z, viewVector), 1.0);
    #else
        gl_FragColor = vec4(1.0);
    #endif
    if (uDebugReturn != 0.0)
    {
        return;
    }
#endif

    vec3 worldNormal = normalize(vNormal);
    vec3 tangent;
    vec3 bitangent;
    // The bitangent and tangent are needed for some BRDFs, therefore we always need to compute them
#if defined(PER_FRAGMENT_TANGENTS)
    ComputeTangentBitangentFromDerivatives(tangent, bitangent, worldNormal, positionDX, positionDY, uvDX, uvDY);
#endif

    // Normal map
#if defined(USE_NORMAL_MAP)

#if !defined(PER_FRAGMENT_TANGENTS)
    tangent = normalize(vTangent.xyz);
    bitangent = ComputeBitangent(worldNormal, vTangent);
#endif

    vec3 textureNormal = UnpackNormal(texelEmissiveNormal.gba, textureSettings.normalScale);
    worldNormal = ApplyNormalMap(textureNormal, worldNormal, tangent, bitangent);
    // Return if using these debug types - note gl_FragColor is set in ApplyNormalMap
    #if defined(DEBUG_TANGENTS) || defined(DEBUG_BITANGENTS)
        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif
#endif

#if defined(VIEWPORTLOOKUPSCALE)
    vec2 screenTexCoords = gl_FragCoord.xy * uFullScreenLookupScale.xy;
#endif

#if defined(VIEWPORTMAP)
    albedo.rgb = SampleViewportMapColour(gl_FragCoord.xy, textureSettings.viewportMapUVScaleAndAnim);
#endif

    LightingTerms indirectLighting;
    ClearLightingTerms(indirectLighting);

    // Diffuse irradiance
    indirectLighting.Diffuse = uAmbientColour;
    vec3 irradiance = vec3(1.0, 1.0, 1.0);
#if defined(IRRADIANCE_LIGHTING)
    irradiance = EvaluateSHLighting2ndOrder(worldNormal, uIrradianceSHCoefs);
    indirectLighting.Diffuse *= irradiance;
#endif

    highp float cameraDistance = length(viewVector);
    // Normalize view direction reusing length value
    highp vec3 viewDirection = viewVector / cameraDistance;

    float invRoughness = 1.0 - texelMetalnessRoughness.g;

    // When using a metalness map, insulative surfaces � pixels set to 0.0 (black) in the metalness map � are assigned a fixed reflectance value (linear: 0.04 sRGB: 0.22) and use the albedo map for the diffuse value.
    // For metallic surfaces � pixels set to 1.0 (white) in the metalness map � the specular color and intensity is taken from the albedo map, and the diffuse value is set to 0 (black) in the shader.
    // Gray values in the metalness map will be treated as partially metallic and will pull the reflectivity from the albedo and darken the diffuse proportionally to that value (partially metallic materials are uncommon).
    // Taken from https://www.marmoset.co/posts/physically-based-rendering-and-you-can-too/
    vec3 specularAlbedo = mix(vec3(textureSettings.nonMetalSpecular), albedo.rgb, texelMetalnessRoughness.r);
    albedo.rgb = mix(albedo.rgb, vec3(0.0), texelMetalnessRoughness.r);
    // Calculate specular power into the correct range 0-8192 using the inverse of roughness
    float specPow = pow(uGlobalEnvironmentMappingParams.y, invRoughness);

    LightingTerms directLighting;
    ClearLightingTerms(directLighting);

    // Direct sunlight lighting
#if defined(SUNLIGHT_DIRECT_LIGHTING)
    int iCascade = -1;
    // Sunlight shadow attenuation
    float amountInSunlight = 0.0;

    #if defined(DEFERRED_SHADOWS)
        directLighting = EvaluateSunlightRT7(amountInSunlight, worldNormal, bitangent, -viewDirection, screenTexCoords, specularAlbedo, specPow, texelMetalnessRoughness.g, textureSettings.fresnelExponent);
    #else
        // Compute depth in camera space
        // Camera plane distance more accurate and must be used for cascade interval select to work without artifacts
        highp vec3 camLook = vec3(uViewMatrix[0][2], uViewMatrix[1][2], uViewMatrix[2][2]);
        highp float viewSpaceDepth = abs(dot(viewVector, camLook));
        directLighting = EvaluateSunlightRT7(iCascade, amountInSunlight, worldPos, worldNormal, bitangent, viewSpaceDepth, -viewDirection, specularAlbedo, specPow, texelMetalnessRoughness.g, textureSettings.fresnelExponent);
    #endif
#else
    directLighting.Diffuse = vec3(1.0);
#endif

#if defined(TEXTURE_ATLAS) && defined(DEBUG_MATERIAL_HIGHLIGHT)
    if (uDebugMaterialHighlight != -1.0)
    {
        float lightStrength = mix(0.1, 0.5, length(directLighting.Diffuse));
        float base = 0.1;
        vec3 col = mix(vec3(base) * lightStrength, vec3(0,1,0), textureSettings.materialID == uDebugMaterialHighlight ? 1 : 0);
        gl_FragColor = vec4(col, 1);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    }
#endif

    // Bias the IBL fresnel according to roughness (i.e. fresnel is reduced as roughness increases) and distance from the camera
    vec3 envSpecular = FresnelSchlickRoughness(specularAlbedo, max(0.0, dot(-viewDirection, worldNormal)), textureSettings.fresnelExponent, invRoughness);

    // Note GLOBAL_ENVIRONMENTMAPPING is unconditional (hardware capabilities only)
#if defined(TEXTURE_ATLAS) && defined(GLOBAL_ENVIRONMENTMAPPING)
    SampleCubeMapSpecular(indirectLighting.Specular, uGlobalEnvironmentMappingParams.x, uGlobalEnvironmentMap, -viewDirection, worldNormal, envSpecular, invRoughness);
#endif
#if defined(NORMALIZED_ENVIRONMENTMAPPING)
    // Scale/modulate the environment specular by the incoming irradiance to try and normalise the environment map more to the local environment lighting
    // This would be more accurate if we had local light probes with normalised environment maps.
    indirectLighting.Specular *= (irradiance * uGlobalEnvironmentMappingParams.w);
#endif
#if defined(SPECULAR_HORIZON_OCCLUSION)
    indirectLighting.Specular *= SpecularHorizonOcclusion(uGlobalEnvironmentMappingParams.z, viewDirection, worldNormal);
#endif

    vec4 fragColour = vec4(0.0, 0.0, 0.0, albedo.a);

    // Point lights
#if defined(POINT_LIGHTING)
    vec3 pointLightsTerm = vec3(0.0, 0.0, 0.0);
    vec3 pointLightsSpecular = vec3(0.0, 0.0, 0.0);
    #if defined(POINT_LIGHTING_VERTEX_DATA)
        EvaluatePointLights(pointLightsTerm, pointLightsSpecular, specularAlbedo, specPow, texelMetalnessRoughness.g, textureSettings.fresnelExponent, -viewDirection, vWorldPosition.xyz, worldNormal, viewSpaceDepth, vTilePosition);
    #else
        EvaluatePointLights(pointLightsTerm, pointLightsSpecular, specularAlbedo, specPow, texelMetalnessRoughness.g, textureSettings.fresnelExponent, -viewDirection, vWorldPosition.xyz, worldNormal, viewSpaceDepth);
    #endif

    #if defined(DEBUG_POINTLIGHTS)
        gl_FragColor = vec4(pointLightsTerm, albedo.a);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DEBUG_POINTLIGHTS_SPECULAR)
        gl_FragColor = vec4(pointLightsSpecular, albedo.a);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DIFFUSE_LIGHTING)
        directLighting.Diffuse += pointLightsTerm;
    #else
        directLighting.Diffuse = pointLightsTerm;
    #endif

    #if defined(POINT_LIGHTING_SPECULAR)
        directLighting.Specular += pointLightsSpecular;
    #endif
#else
    #if defined(DEBUG_POINTLIGHTS)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DEBUG_POINTLIGHTS_SPECULAR)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if !defined(DIFFUSE_LIGHTING)
        directLighting.Diffuse = vec3(0.0);
    #endif
#endif

    float causticsTerm = 0.0;
#if defined(CAUSTICS)
    causticsTerm = CalculateCausticsTerm(vWorldPosition.xyz, amountInSunlight, worldNormal);
#endif

    // Combined lighting terms
    vec3 totalDiffuse = vec3(0.0, 0.0, 0.0);
    vec3 totalSpecular = vec3(0.0, 0.0, 0.0);

    // Ambient
#if defined(AMBIENT_LIGHTING)
    // Apply ambient occlusion to indirect lighting
    #if defined(SSAO) && !defined(REFRACTION)
        float ambientOcclusion = texture2D(uSSAOMap, screenTexCoords).r;
        indirectLighting.Diffuse *= ambientOcclusion;
        indirectLighting.Specular *= ambientOcclusion;
    #endif

    totalDiffuse += indirectLighting.Diffuse;
    totalSpecular += indirectLighting.Specular;
#endif
    // Diffuse
    totalDiffuse += directLighting.Diffuse;
    // Specular
#if defined(SPECULAR_LIGHTING)
    totalSpecular += directLighting.Specular;
#endif
    // Caustics
#if defined(CAUSTICS)
    totalDiffuse += uSunColour * causticsTerm;
#endif

    // Albedo
#if defined(ALBEDO_LIGHTING)
    totalDiffuse *= albedo.rgb;
#endif

    // Although combining diffuse and specular here doesn't follow the laws of conservation of energy, for pure non-metals (i.e. metalness of 0.0),
    // diffuse albedo will actually be set to black, which has almost the same effect.
    fragColour.rgb = totalDiffuse + totalSpecular;

#if defined(REFRACTION)
    if (fragColour.a < 0.99)
    {
        fragColour.rgb = CalculateRefractionColour(
            vWorldPosition.xyz,
            worldNormal,
            viewDirection,
            invRoughness,
            uRefractionParams,
            worldNormal.xy, // Using the x and y component for the offset prevents the refracted image from shifting
            fragColour,
            indirectLighting.Specular,
            envSpecular);
    }
#endif

    // Emissive - RT7 emissive method, lerps between lit colour and emissive colour using the emissive mask
#if defined(TEXTURE_ATLAS)
    // The colour can be either the texture albedo or a specified colour in the material settings
    vec3 emissiveColour = mix(albedo.rgb, textureSettings.emissiveColourAndEmissiveSource.rgb, textureSettings.emissiveColourAndEmissiveSource.a);
    // Scale the colour and add it to itself - this allows for 0 value emissiveScale which equates to being unlit
    emissiveColour += emissiveColour * textureSettings.emissiveScale;
    // Lerp between the lit colour and the emissive colour based on the mask
    fragColour.rgb = mix(fragColour.rgb, emissiveColour, texelEmissiveNormal.r);
#endif

    // Light scattering (now includes fog!)
#if defined(LIGHT_SCATTERING) || defined(FOG_DISTANCE)
    fragColour.rgb = ApplyInOutScattering(fragColour.rgb, vOutScattering, vInScattering);
#endif
    // Volumetric scattering (should only be required here for alpha blended surfaces)
#if defined(VOLUMETRIC_SCATTERING_SUPPORTED)
    vec4 scattering = GetScatteredInRay2(8, viewDirection, cameraDistance, CalculateScatteringOffset(gl_FragCoord.xy));
    fragColour = ApplyVolumetricScattering(fragColour, scattering);
#endif

#if defined(TINT) && defined(PUSH_TO_FARPLANE) // Overbright for the sky model (used in irradiance)
    fragColour.rgb += uTint.rgb;
#endif

    // Debug lighting
#if defined(DEBUG_ALBEDO)
    fragColour = albedo;
#endif
#if defined(DEBUG_NORMALS)
    fragColour.rgb = (worldNormal * 0.5 + 0.5);
    fragColour.a = 1.0;
#endif
#if defined(DEBUG_FRESNEL)
    fragColour.rgb = vec3(envSpecular);
    fragColour.a = 1.0;
#endif
    // Debug shadows
#if defined(SUNLIGHT_SHADOWS) && defined(DEBUG_SUNLIGHT_SHADOW_CASCADE) && !defined(DEFERRED_SHADOWS)
    fragColour.rgb *= ShadowMapCascadeColour(iCascade, int(uMappingParams.w)).rgb;
#endif
    // Debug RT7 maps
#if defined(DEBUG_RT7_EMISSIVE)
    fragColour = vec4(texelEmissiveNormal.rrr, 1.0);
#endif
#if defined(DEBUG_RT7_METALNESS)
    fragColour = vec4(texelMetalnessRoughness.rrr, 1.0);
#endif
#if defined(DEBUG_RT7_ROUGHNESS)
    fragColour = vec4(texelMetalnessRoughness.ggg, 1.0);
#endif

#if defined(FORCE_OPAQUE)
    fragColour.a = 1.0;
#endif

#if defined(DISCRETIZED_LUMINANCE)
    fragColour.rgb = DiscretizeLuminance(fragColour.rgb);
#endif

#if defined(PREMULTIPLY_ALPHA)
    fragColour.rgb *= fragColour.a;
#endif

    gl_FragColor = fragColour;
}
