#include "ParticleCommon.inc"
#include "Shared/DepthUtils.inc"
#include "Shared/ViewportMap/ViewportMap.inc"
#include "ParticleLighting.inc"
#include "Shared/TextureCompression.inc"
#include "Shared/VolumetricScattering/VolumetricFunctions.inc"
#include "Shared/VolumetricScattering/ApplyVolumetrics.inc"

void SampleParticleTextures(out vec4 albedo, out vec3 normal, TextureSettings textureSettings, vec2 uv)
{
    vec4 texelAlbedo = vec4(1.0);

    // Initialising the normal here so Angle's happy
    normal = vec3(0.0, 1.0, 0.0);

#if defined(HDR_SCALE) || defined(NORMAL_MAP)
    vec4 texelHDRNormal = vec4(0.0);
    getTexel(uv, textureSettings.textureMeta1, textureSettings.textureMeta2, uAtlasMeta, 0.0, uTextureAtlas, texelAlbedo, texelHDRNormal);

    #if defined(ETC_CHANNEL_SWIZZLE)
        // Swizzle ETC channels back into the expected order
        texelHDRNormal = texelHDRNormal.agbr; //NX,NY,NZ,HD -> HD,NY,NZ,NX
    #endif

#if defined(NORMAL_MAP)
    normal = UnpackNormal(texelHDRNormal.yzw);
#endif

#if defined(HDR_SCALE)
    texelAlbedo = HDRScale(texelAlbedo, texelHDRNormal.x);
#endif

#else
    getTexel(uv, textureSettings.textureMeta1, uAtlasMeta, 0.0, uTextureAtlas, texelAlbedo);
#endif

    vec4 vtxAlbedo = vVertexAlbedo;

#if defined(VIEWPORTMAP)
    texelAlbedo.rgb = SampleViewportMapColour(gl_FragCoord.xy, textureSettings.viewportMapUVScaleAndAnim);
    vtxAlbedo.rgb = vec3(1.0);
#endif

    albedo = vtxAlbedo * texelAlbedo;
}

//Main function
void main()
{
    //[
    #if defined(CLIP_PLANE)
        if (dot(vWorldPosition, uClipPlane) < 0.0)
        {
            discard;
            return;
        }
    #endif
    //]

    TextureSettings textureSettings;
    getTextureSettings(vMaterialSettingsSlotXY, textureSettings);

    vec3 normal;
    vec4 fragColour;

    SampleParticleTextures(fragColour, normal, textureSettings, vTextureUV.xy);

    // Sample the next frame if we're using flipbook animation
    if (vUVInterpolationAmount > 0.0)
    {
        vec3 nextFrameNormal = normal;
        vec4 nextFrameSample;

        SampleParticleTextures(nextFrameSample, nextFrameNormal, textureSettings, vTextureUV.zw);

        // Blend between the two frames
        fragColour = mix(fragColour, nextFrameSample, vUVInterpolationAmount);
    #if defined(NORMAL_MAP)
        normal = normalize(mix(normal, nextFrameNormal, vUVInterpolationAmount));
    #endif
    }

    vec3 viewVector = vWorldPosition.xyz - uCameraPosition;
    float cameraDistance = length(viewVector);
    vec3 viewDirection = viewVector / cameraDistance;

#if defined(SOFT_PARTICLES)
    // The following is equivalent to `if (vSoftFadeEnabled) { fragColour.a *= GetSoftParticleFade() }` but avoids conditionals
    fragColour.a *= 1.0 - vSoftFadeEnabled * (1.0 - GetSoftParticleFade());
#endif

    // Disable rendering of the many particle fragments that are almost if not completely invisible
    if (fragColour.a <= 0.001)
    {
        discard;
        return;
    }

#if defined(LIT_PARTICLES) && defined(SUNLIGHT_DIRECT_LIGHTING) && defined(SUNLIGHT_SHADOWS) && defined(NORMAL_MAP)
    if (vLightingContribution != 0.0)
    {
        // After sampling from the normal map we end up with a viewpace normal but we need it in WS for lighting calcs
        normal = -(uInvViewMatrix * vec4(normal, 0.0)).xyz;

        fragColour.rgb = ApplyParticleLighting(fragColour.rgb, normal, vWorldPosition, vViewSpaceDepth);
    }
#endif

    // Apply light scattering after particle lighting to avoid lit particles not looking as foggy as they should
#if defined(LIGHT_SCATTERING) || defined(FOG_DISTANCE)
    vec3 outScattering = vOutScattering;
    vec3 inScattering = vInScattering;

    #if defined(VOLUMETRIC_SCATTERING_SUPPORTED)
        vec4 scattering = GetScatteredInRay2(8, viewDirection, cameraDistance, CalculateScatteringOffset(gl_FragCoord.xy));
        fragColour = ApplyVolumetricScattering(fragColour, scattering);
    #endif

    // Disable inscattering if we're using additive particles (where vAlphaOutput == 0)
    inScattering *= vAlphaOutput;

    fragColour.rgb = ApplyInOutScattering(fragColour.rgb, outScattering, inScattering);
#endif

    // We're using ONE as the srcBlendFactor so we can support both alpha and additive blending without
    // changing render states in between so we can batch more particles at once, so we need to basically
    // use pre-multiplied alpha here.
    fragColour.rgb *= fragColour.a;
    fragColour.a *= vAlphaOutput;

    gl_FragColor = fragColour;
}
