#ifndef PARTICLE_COMMON_INC
#define PARTICLE_COMMON_INC

// Note for shaders that only need the transform functionality from this for vertex shaders, you can define
// PARTICLE_COMMON_TRANSFORMS_ONLY to omit all the varying/uniform-using code.

#include "UndefineInOut.inc"

// Must be kept in sync with Particle.cpp's Particle::Flags
#define PARTICLE_FLAG_ADDITIVE_RENDER_MODE_BIT 0
#define PARTICLE_FLAG_HEADING_ALIGNED_BIT 1
#define PARTICLE_FLAG_FRAME_BLEND_BIT 2
#define PARTICLE_FLAG_LIGHTING_PER_PARTICLE 3
#define PARTICLE_FLAG_LIGHTING_PER_VERTEX 4
#define PARTICLE_FLAG_LIGHTING_PER_PIXEL 5
#define PARTICLE_FLAG_NO_SOFT_FADE 6

bool TestParticleFlag(int flags, int flagBit)
{
    bool result;

#if __VERSION__ <= 120
    // Compat mode doesn't support bitwise ops so we have to use this ugly and slow method instead.
    const float epsilon = 0.000001;
    float curBitValue = pow(2.0, float(flagBit));
    float nextBitValue = pow(2.0, float(flagBit + 1));
    float modResult = mod(float(flags), nextBitValue);
    result = modResult >= (curBitValue - epsilon);
#else
    int flagMask = 1 << flagBit;
    result = (flags & flagMask) == flagMask;
#endif

    return result;
}

/** Unpacks two floats in a 0..16 range from a single 0..255 float value */
void UnpackFloat8(out float v0, out float v1, float packedValue)
{
#if __VERSION__ <= 120
    // Compat mode doesn't support bitwise ops so we have to use this ugly and slow method instead.
    float scaledV = packedValue / 16.0;
    float fractScaledV =  fract(scaledV);
    v0 = fractScaledV * 16.0;
    v1 = (scaledV - fractScaledV);
#else
    v0 = float(int(packedValue) & 0xF);
    v1 = float((int(packedValue) & 0xF0) >> 4);
#endif
}

#if !defined(SHADER_TYPE_PIXEL)
/**
    Contains code common to all particle shaders
 */

/** Rotates the particle world space position, keeping it aligned to its heading */
void ApplyParticleOffset3DWS(inout vec4 worldPosOut, vec3 velocity, vec2 offset, vec3 wsOriginDir)
{
    vec3 fwd;

    if (velocity == vec3(0.0, 0.0, 0.0))
    {
        fwd = vec3(0.0, 0.0, 1.0);
    }
    else
    {
        fwd = normalize(velocity);
    }

    // Lock the X axis so it faces to the side of the camera, so particles don't get infinitely thin as
    // we rotate the camera around from their base X axis.
    vec3 right = cross(fwd, wsOriginDir);

    if (right == vec3(0.0, 0.0, 0.0))
    {
        vec3 right = cross(vec3(0.0, 1.0, 0.0), fwd);

        if (right == vec3(0.0, 0.0, 0.0))
        {
            right = cross(vec3(0.0, 0.0, 1.0), fwd);
        }
    }

    right = normalize(right);

    worldPosOut.xyz += fwd.xyz * offset.y + right.xyz * offset.x;
}

/** Rotates the particle camera space position */
void ApplyParticleOffset2DVS(inout vec4 cameraSpacePosOut, vec2 posOffset, vec2 packedRotationAngle)
{
    // Unpack and convert from 0..65535 back to 0..2pi
    float rotationAngle = (packedRotationAngle.y * 256.0 + packedRotationAngle.x) / 65535.0;
    rotationAngle = rotationAngle * 2.0 * PI;

    float cosRotation = cos(rotationAngle);
    float sinRotation = sin(rotationAngle);

    cameraSpacePosOut.xy += vec2(
        posOffset.x * cosRotation - posOffset.y * sinRotation,
        posOffset.x * sinRotation + posOffset.y * cosRotation
    );
}

#if !defined(PARTICLE_COMMON_TRANSFORMS_ONLY)
/** Calculates the particle UVs for the two active animation frames */
void CalculateParticleUV(out vec4 uv, out float interpolationAmount, vec2 posOffset, float particleCreationTime, vec4 particleUVAnimationData)
{
    // Can just use step here since posOffset represents the current vertex's offset from the particle's origin
    vec2 baseUV = step(vec2(0.0, 0.0), -posOffset);
    float numAnimationFrames;
    float numTilesPerDimension;

    UnpackFloat8(numAnimationFrames, numTilesPerDimension, particleUVAnimationData.x);

    if (numAnimationFrames > 1.0)
    {
        if (TestParticleFlag(int(particleUVAnimationData.z), PARTICLE_FLAG_FRAME_BLEND_BIT))
        {
            // Using flipbook animation and blending between frames
            float framerate = particleUVAnimationData.y;

            float uvPerFrame = 1.0 / float(numTilesPerDimension);
            baseUV *= uvPerFrame;

            // Determine how far along we are in the animation
            float timeSinceParticleCreation = uTextureAnimationTime - particleCreationTime;
            float animationT = timeSinceParticleCreation * framerate;

            // Get the current and next frames
            // Note: Because of the atlas packing and the requirements it instills on textures,
            // we can't assume that (numAnimationFrames = numTilesPerDimension^2) so we need the additional
            // modulus here.
            float currentFrameT = floor(animationT);
            float currentFrame = mod(currentFrameT, numAnimationFrames);
            float nextFrame = mod(currentFrame + 1.0, numAnimationFrames);

            float currentFrameX = mod(currentFrame, numTilesPerDimension);
            float currentFrameY = (currentFrame - currentFrameX) / numTilesPerDimension;

            float nextFrameX = mod(nextFrame, numTilesPerDimension);
            float nextFrameY = (nextFrame - nextFrameX) / numTilesPerDimension;

            vec2 currentFrameUV = vec2(currentFrameX, currentFrameY) * uvPerFrame;
            vec2 nextFrameUV = vec2(nextFrameX, nextFrameY) * uvPerFrame;

            interpolationAmount = animationT - currentFrameT;
            uv = vec4(baseUV + currentFrameUV, baseUV + nextFrameUV);
        }
        else
        {
            // UV tiles
            float uvPerTile = 1.0 / float(numTilesPerDimension);

            float tileIndex = particleUVAnimationData.y;
            float tileX = mod(tileIndex, numTilesPerDimension);
            float tileY = (tileIndex - tileX) / numTilesPerDimension;

            vec2 tileUV;

            // Special case when we're laying out the frames horizontally
            if (numAnimationFrames != numTilesPerDimension)
            {
                tileUV = (baseUV + vec2(tileX, tileY)) * uvPerTile;
            }
            else
            {
                baseUV.x *= uvPerTile;
                tileUV = baseUV + vec2(tileX * uvPerTile, 0.0);
            }

            uv = vec4(tileUV, tileUV);
            interpolationAmount = 0.0;
        }
    }
    else
    {
        uv = vec4(baseUV, baseUV);
        interpolationAmount = 0.0;
    }
}

/** Helper for copying particle vertex data into varyings */
void AssignParticleVaryings(vec4 vertexColour, vec4 particleUV, float uvInterpolationAmount, vec2 particleMaterialSettingsSlotXY)
{
    vVertexAlbedo = vertexColour;
    vTextureUV = particleUV;
    vMaterialSettingsSlotXY = particleMaterialSettingsSlotXY;
    vUVInterpolationAmount = uvInterpolationAmount;

// Convert to linear space for lighting calculations
#if defined(GAMMA_CORRECT_INPUTS)
    vVertexAlbedo.rgb = SRGBToLinear(vVertexAlbedo.rgb);
#endif
}

#endif

#else

#include "Shared/DepthUtils.inc"

float GetSoftParticleFade()
{
    float softFade = 1.0;

    // See http://developer.download.nvidia.com/SDK/10/direct3d/Source/SoftParticles/doc/SoftParticles_hi.pdf for more info on soft particle rendering
#if defined(VIEWPORTLOOKUPSCALE) && defined(REFRACTION) && defined(SOFT_PARTICLES)
    vec2 sceneSamplePos = gl_FragCoord.xy * uViewportLookupScale.xy;
    float sceneViewDepth = texture2DLodCompat(uRefractionDepth, sceneSamplePos, 0.0).r;
    sceneViewDepth = GetViewSpaceDepth(sceneViewDepth, vec3(uZBufferParams.xyz));

    // Compute softness of particle fragment
    softFade = clamp((sceneViewDepth - vViewSpaceDepth) * uParticleInvSoftScale, 0.0, 1.0);
#endif

    return softFade;
}

#endif

#endif
