#include "ParticleCommon.inc"
#include "Shared/ViewportMap/ViewportMap.inc"
#include "Shared/TextureCompression.inc"

#define PARTICLE_REFRACTION_DEPTH_CHECK_ENABLED 1

void SampleParticleTextures(inout vec4 albedo, inout vec3 normal, inout float hdrScale, TextureSettings textureSettings, vec2 uv)
{
    vec4 texelAlbedo = vec4(1.0);
    vec4 texelHDRNormal = vec4(0.0);

    getTexel(uv, textureSettings.textureMeta1, textureSettings.textureMeta2, uAtlasMeta, 0.0, uTextureAtlas, texelAlbedo, texelHDRNormal);

    #if defined(ETC_CHANNEL_SWIZZLE)
        // Swizzle ETC channels back into the expected order
        texelHDRNormal = texelHDRNormal.agbr; //NX,NY,NZ,HD -> HD,NY,NZ,NX
    #endif

    normal = UnpackNormal(texelHDRNormal.yzw);
    albedo = vVertexAlbedo * texelAlbedo;
    hdrScale = texelHDRNormal.x;
}

vec3 SampleDistortedMap(vec3 viewNormal, float distortionStrength)
{
    vec2 refractionViewportScaleRcp = vec2(1.0) / uRefractionViewportLookupScale.zw;

    vec2 viewportOffset = uViewportOffsetScale.xy * uFullScreenLookupScale.zw;
    vec2 baseSamplingCoords = (gl_FragCoord.xy - viewportOffset) * uViewportLookupScale.xy;
    vec2 distortedMapCoords = baseSamplingCoords + (viewNormal.xz * distortionStrength) * refractionViewportScaleRcp;

#if PARTICLE_REFRACTION_DEPTH_CHECK_ENABLED
    // If the depth in the refraction map is below our surface take an unperturbed refraction sample
    // Because we may use a lower resolution for the refraction textures, we can end up sampling an incorrect depth (compared to the fullscreen buffer),
    // which is mainly noticeable on object silhouettes. To address this, we bias the depth sampling offset slightly which mostly takes care of the issue.
    const float refractionNormalBias = 3.0;

    float refractionDepth = texture2DLodCompat(uRefractionDepth, distortedMapCoords + (viewNormal.xz * refractionNormalBias) * refractionViewportScaleRcp, 0.0).r;

    if (refractionDepth < gl_FragCoord.z)
    {
        distortedMapCoords = baseSamplingCoords;
    }
#endif

    return texture2DLodCompat(uRefractionMap, vec3(distortedMapCoords, 0.0), 0.0).rgb;
}

//Main function
void main()
{
    //[
    #if defined(CLIP_PLANE)
        if (dot(vWorldPosition, uClipPlane) < 0.0)
        {
            discard;
            return;
        }
    #endif
    //]

    TextureSettings textureSettings;
    getTextureSettings(vMaterialSettingsSlotXY, textureSettings);

    vec3 normal = vec3(0.0, 1.0, 0.0);
    vec4 fragColour;
    float distortionStrength = 0.0;

    SampleParticleTextures(fragColour, normal, distortionStrength, textureSettings, vTextureUV.xy);

    // Sample the next frame if we're using flipbook animation
    if (vUVInterpolationAmount > 0.0)
    {
        vec3 nextFrameNormal;
        vec4 nextFrameSample;
        float nextFrameDistortionStrength;

        SampleParticleTextures(nextFrameSample, nextFrameNormal, nextFrameDistortionStrength, textureSettings, vTextureUV.zw);

        // Blend between the two frames
        fragColour = mix(fragColour, nextFrameSample, vUVInterpolationAmount);
        normal = normalize(mix(normal, nextFrameNormal, vUVInterpolationAmount));
        distortionStrength = mix(distortionStrength, nextFrameDistortionStrength, vUVInterpolationAmount);
    }

    // Discard fragments with minimal distortion strength
    if (distortionStrength <= 0.01)
    {
        discard;
        return;
    }

    fragColour.a *= GetSoftParticleFade();

    // Disable rendering of the many particle fragments that are almost if not completely invisible
    if (fragColour.a <= 0.01)
    {
        discard;
        return;
    }

    // Distortion strength's sampled in the 0..1 range so we need to convert it back to a usable range
    distortionStrength *= 255.0;

    // Modulate the direction of the distortion normals and tint based on the alpha map
    normal = mix(vec3(0.0, 1.0, 0.0), normal, fragColour.a);
    vec3 tint = mix(vec3(1.0), fragColour.rgb, fragColour.a);

    fragColour.rgb = tint * SampleDistortedMap(normal, distortionStrength);

    gl_FragColor = fragColour;
}
