/*
 * SHADER: RGBLogLuminance.cs
 * AUTHOR: Gert Dhaenens
 *
 * The idea for it is to generate the average log luminance value of said texture.
 * Classical approaches is to downsample the RGB scene texture and calculate the single channel
 * log luminance. We then take the higher resolution log luminance and downsample to take averages
 * and take advantage of filtering. After that, we take the average of a region and output a single value.
 * This is considered the inverse exposure of the scene (i.e. brightness).
 *
 * The old logic is as follows
 * 1. Draw call 0: Downsample the original texture to a 64x64 using a 3x3 box filter and write out the log luminance
 * 2. Draw call 1: Downsample 64x64 to 16x16
 * 3. Draw call 2: Downsample 16x16 to 4x4
 * 4. Draw call 3: Calculate the average of the 4x4 and write out exp(logLum) to a 1x1 texture
 * 5. Draw call 4: Adapt the current loglum from the previous frame to the new value as
 *
 * Historically, this downsample chain is expensive and causes a lot of round trips of memory with
 * each stage being an explicit dependency of the previous state, meaning that no pipelining occurs
 * and caches have to be invalidated between every single exeuction, even the CB/DB blocks.
 * This is less bad on tiled renderer as it stays within tile memory, but still forces the explicit syncs.
 *
 * The new logic uses a double dispatch method to save round trips to memory
 * 1. Dispatch 0: Downsample the original texture to a 64x64 using a 4x4 box filter (bilinear sample) and calculate log luminance
 * 2. Dispatch 0: Downsample 64x64 to 32x32 using groupshared memory
 * 3. Dispatch 0: Downsample 32x32 to 16x16 using groupshared memory
 * 4. Dispatch 0: Downsample 16x16 to 8x8 using groupshared memory and write to 8x8 texture
 * 5. Dispatch 1: Downsample 8x8 to 2x2 using 4x4 box filter (bilinear sample)
 * 6. Dispatch 1: Calculate average of 2x2 and write out exp(logLum) to 1x1 texture
 * 7. Dispatch 1: Calculate adapted log lum based on previous value and write to 1x1 texture
 *
 * The group shared memory size is dependent on the size of our thread group,
 * which differs between the first and second pass.
 * However, the layout and logic between all of the various passes and steps is the same
 * Consider a thread group size of 2x2x1, the layout will be as follows
 *
 * |----|----|----|----|
 * | 00 | 01 | 02 | 03 |
 * |----|----|----|----|
 * | 04 | 05 | 06 | 07 |
 * |----|----|----|----|
 * | 08 | 09 | 10 | 11 |
 * |----|----|----|----|
 * | 12 | 13 | 14 | 15 |
 * |----|----|----|----|
 *
 * Each thread that is active reads a 2x2 sample from a region in memory
 * and then writes back the result to the original target group memory address.
 *
 * For example:
 *
 * Step 1 - 4 active threads (2x2x1)
 * - Thread 0,0 reads 00, 01, 04 and 05 and writes it's result to 00
 * - Thread 0,1 reads 02, 03, 06 and 07 and writes it's result to 01
 * - Thread 1,0 reads 08, 09, 12 and 13 and writes it's result to 04
 * - Thread 1,1 reads 10, 11, 14 and 15 and writes it's result to 05
 *
 * Step 2 - 1 active thread (1x1x1)
 * - Thread 0,0 reads 00, 01, 04 and 05 and writes it's result to 00
 */

/**
 * Offsets to sample 4x4 region using 2x2 texture samples
 * by taking advantage of bilinear sampling
 * Encoded as 2 vec4's for 16-byte alignment
 */
uniform vec4 uSampleOffsets4x4[2u];

/** Time delta to apply to our adaptation */
uniform float uTimeDelta;

/** Reciprocal of the exposure speed multiplier to apply to our adaptation */
uniform float uExposureSpeedMul;

/**
 * Multiple OpenGLES-based devices (e.g. ARM) flush-to-zero when denormalised
 * numbers are used as input for operations. Since OpenGLES doesn't allow us
 * to disable this behaviour and we can have luminance values that are very
 * low, we want to multiply our colours with a multiplier to avoid our luminance
 * values denormalising
 */
uniform float uColourScale;
uniform float uInverseColourScale;

/**
 * Calculate the flattened index into groupshared memory from a 2D index
 */
uint GroupMemoryIndex(uvec2 localIndex)
{
    return ( localIndex.y * cGroupSharedDim ) + localIndex.x;
}

/**
 * Calculate the flattened index into groupshared memory from a 2D index
 * @param offset The offset to apply to the local index
 */
uint GroupMemoryIndex(uvec2 localIndex, uvec2 offset)
{
    return ( ( localIndex.y + offset.y ) * cGroupSharedDim ) + ( localIndex.x + offset.x );
}

/**
 * Encode our luminance value
 * Encoding can be disabled from the CPU side with defines due to
 * certain devices (e.g. Mali-T880) having issues with log precision
 */
float EncodeLuminance(float luminance)
{
    // The shader minifier doesn't like returns within preprocessor if/else blocks and omits the else if we do that
    float res = 0.0;
#if AVG_LUMINANCE_FILTER_COMPUTE_DOWNSAMPLE_LOG_ENCODE
    res = log(luminance);
#else
    res = max(luminance * uColourScale, MIN_VALID_LUMINANCE);
#endif
    return res;
}

/**
 * Decode our luminance value
 * Encoding can be disabled from the CPU side with defines due to
 * certain devices (e.g. Mali-T880) having issues with log precision
 */
float DecodeLuminance(float luminance)
{
    // The shader minifier doesn't like returns within preprocessor if/else blocks and omits the else if we do that
    float res = 0.0;
#if AVG_LUMINANCE_FILTER_COMPUTE_DOWNSAMPLE_LOG_ENCODE
    res = exp(luminance);
#else
    res = max(luminance * uInverseColourScale, MIN_VALID_LUMINANCE);
#endif
    return res;
}

/**
 * Downsample a 2x2 region from groupshared
 * @param i_baseGridIndex The index of the top left within the groupshared memory grid to sample from
 * @returns The average of a 2x2 region in groupshared memory
 */
float DownsampleFromGroupshared(uvec2 i_baseGridIndex)
{
    // Calculate our indices
    uint idx00 = GroupMemoryIndex(i_baseGridIndex, uvec2(0u, 0u));
    uint idx10 = GroupMemoryIndex(i_baseGridIndex, uvec2(1u, 0u));
    uint idx01 = GroupMemoryIndex(i_baseGridIndex, uvec2(0u, 1u));
    uint idx11 = GroupMemoryIndex(i_baseGridIndex, uvec2(1u, 1u));

    // Read first - then calculate in order to hopefully latency hide
    // On some platforms - this entire access will be coalesced into a single memory
    // fetch since we are reading from a tiled pattern within our groupshared memory
    float l00 = gsLogLuminanceSamples[idx00];
    float l01 = gsLogLuminanceSamples[idx10];
    float l10 = gsLogLuminanceSamples[idx01];
    float l11 = gsLogLuminanceSamples[idx11];

    // Calculate the average
    return dot(vec4(l00, l01, l10, l11), vec4(0.25));
}
/**
 * Calculate the average luminance of 4 colours at the same time
 */
float CalculateAvgLuminanceFast(vec3 colour0, vec3 colour1, vec3 colour2, vec3 colour3)
{
    const vec3 LUM_CONVERT_SCALE_QUARTER = LUM_CONVERT_SCALE * 0.25;

    float avgLum0 = dot(colour0, LUM_CONVERT_SCALE_QUARTER);
    float avgLum1 = dot(colour1, LUM_CONVERT_SCALE_QUARTER);
    float avgLum2 = dot(colour2, LUM_CONVERT_SCALE_QUARTER);
    float avgLum3 = dot(colour3, LUM_CONVERT_SCALE_QUARTER);

    return max(avgLum0 + avgLum1 + avgLum2 + avgLum3, MIN_VALID_LUMINANCE);
}

/**
 * Do a 4x4 box filter from our source texture and result the average log luminance
 * This function expects the source texture to be in RGB format
 */
float DownsampleFromSourceRGB()
{
    // Pre-fetch our offsets from our uniforms
    vec2 o00 = uSampleOffsets4x4[0].xy;
    vec2 o01 = uSampleOffsets4x4[0].zw;
    vec2 o10 = uSampleOffsets4x4[1].xy;
    vec2 o11 = uSampleOffsets4x4[1].zw;

    // Size of a single texel of a 64x64 texture
    const vec2 texelSize64x64 = 1.0 / vec2(64.0, 64.0);

    // Calculate our texture coordinate based on our dispatch size
    vec2 texCoord = vec2(gl_GlobalInvocationID.xy) * texelSize64x64 + (texelSize64x64 * 0.5);

    // Sample our textures
    vec3 s00 = textureLod(sSourceTex, texCoord + o00, 0.0).rgb;
    vec3 s01 = textureLod(sSourceTex, texCoord + o01, 0.0).rgb;
    vec3 s10 = textureLod(sSourceTex, texCoord + o10, 0.0).rgb;
    vec3 s11 = textureLod(sSourceTex, texCoord + o11, 0.0).rgb;

    // Calculate the average luminance
    float avgLum = CalculateAvgLuminanceFast(s00, s01, s10, s11);

    // Encode before returning
    return EncodeLuminance(avgLum);
}

/**
 * Do a 4x4 box filter from our source texture and result the average log luminance
 * This function expects the source texture to already be in log-luminance format
 */
float DownsampleFromSourceL()
{
    // Pre-fetch our offsets from our uniforms
    vec2 o00 = uSampleOffsets4x4[0].xy;
    vec2 o01 = uSampleOffsets4x4[0].zw;
    vec2 o10 = uSampleOffsets4x4[1].xy;
    vec2 o11 = uSampleOffsets4x4[1].zw;

    // Size of a single texel of a 2x2 texture
    const vec2 texelSize2x2 = 1.0 / vec2(2.0, 2.0);

    // Calculate our texture coordinate based on our dispatch size
    vec2 texCoord = vec2(gl_GlobalInvocationID.xy + uvec2(0u,0u)) * texelSize2x2 + (texelSize2x2 * 0.5);

    // Sample our textures
    float s00 = textureLod(sSourceTex, texCoord + o00, 0.0).r;
    float s01 = textureLod(sSourceTex, texCoord + o01, 0.0).r;
    float s10 = textureLod(sSourceTex, texCoord + o10, 0.0).r;
    float s11 = textureLod(sSourceTex, texCoord + o11, 0.0).r;

    // Return the average luminance - this will incur a stall since we have
    // to wait for the texture samples to complete before we can calculate this
    return dot(vec4(s00, s01, s10, s11), vec4(0.25));
}

/**
 * Our first pass does the initial set of downsamples
 * 1. Downsample from any resolution to 64x64 by using 4x4 box filter sample (linear filter) (from tezxture to groupshared)
 * 2. Downsample from 64x64 to 32x32 (groupshared to groupshared)
 * 3. Downsample from 32x32 to 16x16 (groupshared to groupshared)
 * 3. Downsample from 16x16 to 8x8 and write out result (groupshared to texture)
 */
void FirstPassEntryPoint()
{
    uvec2 baseWorkGroupIndex = gl_WorkGroupID.xy * gl_WorkGroupSize.xy;

    // Any resolution to 64x64 using box filter - all lanes active
    // First pass downsamples from RGB and calculates luminance
    {
        uint targetGroupMemoryIndex = GroupMemoryIndex(gl_LocalInvocationID.xy);

        float avgLumEnc = DownsampleFromSourceRGB();
        gsLogLuminanceSamples[targetGroupMemoryIndex] = avgLumEnc;

    #if AVG_LUMINANCE_FILTER_COMPUTE_DOWNSAMPLE_DEBUG
        imageStore(uDebugTargetAvgLumTex64x64, ivec2(gl_GlobalInvocationID.xy), vec4(DecodeLuminance(avgLumEnc), 0.0, 0.0, 0.0));
    #endif
    }
    barrier();

    // 64x64 to 32x32 - 1/4th of lanes active
    // Write result to groupshared
    if (gl_LocalInvocationID.x < 4u && gl_LocalInvocationID.y < 4u)
    {
        uint targetGroupMemoryIndex = GroupMemoryIndex(gl_LocalInvocationID.xy);
        uvec2 sourceGroupMemoryPosition = gl_LocalInvocationID.xy << 1u;
        float avgLumEnc = DownsampleFromGroupshared(sourceGroupMemoryPosition);
        gsLogLuminanceSamples[targetGroupMemoryIndex] = avgLumEnc;

    #if AVG_LUMINANCE_FILTER_COMPUTE_DOWNSAMPLE_DEBUG
        imageStore(uDebugTargetAvgLumTex32x32, ivec2((baseWorkGroupIndex >> 1u) + gl_LocalInvocationID.xy), vec4(DecodeLuminance(avgLumEnc), 0.0, 0.0, 0.0));
    #endif
    }
    barrier();

    // 32x32 to 16x16 - 1/16th of lanes active
    // Write result to groupshared
    if (gl_LocalInvocationID.x < 2u && gl_LocalInvocationID.y < 2u)
    {
        uint targetGroupMemoryIndex = GroupMemoryIndex(gl_LocalInvocationID.xy);
        uvec2 sourceGroupMemoryPosition = gl_LocalInvocationID.xy << 2u;
        float avgLumEnc = DownsampleFromGroupshared(sourceGroupMemoryPosition);
        gsLogLuminanceSamples[targetGroupMemoryIndex] = avgLumEnc;

    #if AVG_LUMINANCE_FILTER_COMPUTE_DOWNSAMPLE_DEBUG
        imageStore(uDebugTargetAvgLumTex16x16, ivec2((baseWorkGroupIndex >> 2u) + gl_LocalInvocationID.xy), vec4(DecodeLuminance(avgLumEnc), 0.0, 0.0, 0.0));
    #endif
    }
    barrier();

    // 16x16 to 8x8 - only lane (0,0,0) active
    // Write result to texture
    if (gl_LocalInvocationID.x < 1u && gl_LocalInvocationID.y < 1u)
    {
        uvec2 sourceGroupMemoryPosition = gl_LocalInvocationID.xy << 3u;
        float avgLumEnc = DownsampleFromGroupshared(sourceGroupMemoryPosition);

    #if AVG_LUMINANCE_FILTER_COMPUTE_DOWNSAMPLE_DEBUG
        imageStore(uDebugTargetAvgLumTex8x8, ivec2((baseWorkGroupIndex >> 3u) + gl_LocalInvocationID.xy), vec4(DecodeLuminance(avgLumEnc), 0.0, 0.0, 0.0));
    #endif

        // Write out to our image
        // We divide our coord by 8 since we execute as 64x64x1 total threads
        // count but run with 8x8x1 work groups.
        ivec2 destCoord = ivec2(gl_GlobalInvocationID.xy >> 3u);
        imageStore(uTargetAvgLumTex, destCoord, vec4(avgLumEnc, 0.0, 0.0, 0.0));
    }
}

/**
 * Our second pass does the second set of downsamples
 * Work group size: 2x2x1
 * 1. Downsample from 8x8 to 2x2 by using 4x4 box filter sample (linear filter) (from texture to groupshared)
 * 2. Downsample from 2x2 to 1x1 and write out result (groupshared to texture)
 * 3. Calculate the adapted luminace based on previous frame and write to texture
 */
void SecondPassEntryPoint()
{
    // 8x8 to 2x2 using box filter - all lanes active
    // Second pass downsamples from luminance
    {
        uint targetGroupMemoryIndex = GroupMemoryIndex(gl_LocalInvocationID.xy);
        float avgLumEnc = DownsampleFromSourceL();
        gsLogLuminanceSamples[targetGroupMemoryIndex] = avgLumEnc;

    #if AVG_LUMINANCE_FILTER_COMPUTE_DOWNSAMPLE_DEBUG
        imageStore(uDebugTargetAvgLumTex2x2, ivec2(gl_GlobalInvocationID.xy), vec4(DecodeLuminance(avgLumEnc), 0.0, 0.0, 0.0));
    #endif
    }
    barrier();

    // 2x2 to 1x1 - only lane (0,0,0) active
    // Write result to texture
    if (gl_LocalInvocationID.x < 1u && gl_LocalInvocationID.y < 1u)
    {
        uvec2 sourceGroupMemoryPosition = gl_LocalInvocationID.xy << 1u;
        float avgLumEnc = DownsampleFromGroupshared(sourceGroupMemoryPosition);
        float avgLum = DecodeLuminance(avgLumEnc);

        // Kick off our texture sample for our adapted AFTER memcnt barriers
        float prevAdaptedLum = imageLoad(uPrevAdaptedLumTex, ivec2(0,0)).r;

        // Write out to our average luminance image
        imageStore(uTargetAvgLumTex, ivec2(0,0), vec4(avgLum, 0.0, 0.0, 0.0));

        // Calculate our new adaptive luminance
        // adaptiveLum needs to be declared outside of #ifdef due to minifier issues
        float adaptiveLum;
    #if AVG_LUMINANCE_FILTER_LINEAR_ATTEN_APPROX
        adaptiveLum = max(CalcAdaptedLumRodsConesApprox(avgLum, prevAdaptedLum, uTimeDelta, uExposureSpeedMul), MIN_VALID_LUMINANCE);
    #else
        adaptiveLum = max(CalcAdaptedLumRodsCones(avgLum, prevAdaptedLum, uTimeDelta, uExposureSpeedMul), MIN_VALID_LUMINANCE);
    #endif

        // Write our new value to our adapted texture
        imageStore(uTargetAdaptedLumTex, ivec2(0,0), vec4(adaptiveLum, 0.0, 0.0, 0.0));
    }
}
