const highp float CAUSTICS_FIXED_POINT_SCALE = 10000.0;

#if defined(CAUSTICS) && !defined(CAUSTICS_COMPUTE) && !defined(CAUSTICS_STENCIL)

float CalculateCausticsTerm(highp vec3 worldPos, float amountInSunlight, vec3 normal)
{
    float causticsTerm = 0.0;
    float fade = 0.0;

    if (worldPos.y <= uCausticsPlaneHeight)
    {
        // Fade fully out in shadow (incl sunlight front facing optimisation)
        fade = step(1.0, amountInSunlight); // Ensures we never see caustics when amountInSunlight < 1.0
    }
    else
    {
    #if defined(CAUSTICS_OVERWATER)
        // Fade such that only horizontal downward facing surfaces receive overwater caustics
        fade = clamp(normal.y * -1.0, 0.0, 1.0);
        float heightFade = smoothstep(uCausticsOverWaterFade.x, uCausticsOverWaterFade.y, worldPos.y - uCausticsPlaneHeight);
        fade *= 1.0 - heightFade;
    #else
        //[
        return 0.0;
        //]
    #endif
    }

    if (fade > 0.0)
    {
        highp vec4 causticsPos = uCausticsViewProjMatrix * vec4(worldPos, 1.0);
        causticsPos.xy /= (2.0 * causticsPos.w);
        // Fade near edge of caustics map
        vec2 edge = abs(causticsPos.xy);
        causticsPos.xy += 0.5;
        edge = smoothstep(0.4, 0.5, edge);
        fade *= max(0.0, 1.0 - (edge.x + edge.y));
        // 3x3 box filter
        if (fade > 0.0)
        {
            causticsTerm += textureOffset(uCausticsMap, causticsPos.xy, ivec2(-1, -1)).r;
            causticsTerm += textureOffset(uCausticsMap, causticsPos.xy, ivec2(-1, 0)).r;
            causticsTerm += textureOffset(uCausticsMap, causticsPos.xy, ivec2(-1, 1)).r;
            causticsTerm += textureOffset(uCausticsMap, causticsPos.xy, ivec2(0, -1)).r;
            causticsTerm += texture2D(uCausticsMap, causticsPos.xy).r * 5.0; // Looks better with some extra on the centre sample, otherwise loses definition
            causticsTerm += textureOffset(uCausticsMap, causticsPos.xy, ivec2(0, 1)).r;
            causticsTerm += textureOffset(uCausticsMap, causticsPos.xy, ivec2(1, -1)).r;
            causticsTerm += textureOffset(uCausticsMap, causticsPos.xy, ivec2(1, 0)).r;
            causticsTerm += textureOffset(uCausticsMap, causticsPos.xy, ivec2(1, 1)).r;
            causticsTerm *= fade / 12.0;
        }
    }
    return causticsTerm;
}

#endif

#if defined(CAUSTICS_COMPUTE)
    void WriteCausticsRay(vec3 normal, float waterDepth)
    {
        vec2 dudvCaustics = normal.xz * waterDepth * uCausticsRefractionScale;
        // Removed the scaling dudv by alpha here as a simplification.
        // The refracted light has a fade applied anyway so reducing the refraction at the coord level is subtle.

        vec2 refractedCoord = (gl_FragCoord.xy + dudvCaustics * 2.0) / uCausticsComputeResolution * uCausticsMapSize;
        // Two curves here: fading in at water edge, and fading out in deep water
        highp float depthFade = min(uCausticsFade.x / waterDepth * uCausticsFade.y, 7.0 * uCausticsFade.y);
        highp float edgeFade = smoothstep(uCausticsFade.z, uCausticsFade.w, waterDepth);
        highp float refractedLight = edgeFade * depthFade * CAUSTICS_FIXED_POINT_SCALE;
        if (refractedLight >= 1.f)
        {
            imageAtomicAdd(uCausticsIntegerMap, ivec2(refractedCoord.xy), uint(refractedLight));
        }
    }
#endif

