#ifndef DEPTH_UTILS_INC
#define DEPTH_UTILS_INC
/**
    Utility functions for converting non-linear depth buffer values to view space depth and position
 */


/**
    This converts a (non-linear) post-homogenous-divided NDC depth to a linear view-space depth value

    @param nonLinearClipDepth (non-linear) post-homogenous-divided NDC depth to a linear view-space depth
    @param ZBufferParams x = near * far, y = near - far, z = far
    @return View space linear depth value
 */
highp float GetViewSpaceDepth(highp float nonLinearClipDepth, highp vec3 ZBufferParams)
{
    return ZBufferParams.x / (ZBufferParams.y * nonLinearClipDepth + ZBufferParams.z);
}
highp vec4 GetViewSpaceDepth(highp vec4 nonLinearClipDepth, highp vec3 ZBufferParams) // An override for vector input
{
    return ZBufferParams.x / (ZBufferParams.y * nonLinearClipDepth + ZBufferParams.z);
}

/**
    This converts a (non-linear) post-homogenous-divided NDC depth to a linear view-space position

    @param xy Screen space NDC coordinates in the range 0-1
    @param nonLinearClipDepth (non-linear) post-homogenous-divided NDC depth to a linear view-space depth
    @param sceneInvProjMatrix The inverse of the projection matrix used to create nonLinearClipDepth
 */
highp vec3 GetViewSpacePos(vec2 xy, highp float nonLinearClipDepth, highp mat4 sceneInvProjMatrix)
{
    highp vec4 clipPos = vec4(2.0 * xy - 1.0, 2.0 * nonLinearClipDepth - 1.0, 1.0);
    highp vec4 eyePos = sceneInvProjMatrix * clipPos;
    eyePos.xyz /= eyePos.w;
    return eyePos.xyz;
}

highp vec4 GetWorldSpacePos(vec2 xy, highp float nonLinearClipDepth, highp mat4 sceneInvViewProjMatrix)
{
    highp vec4 clipPos = vec4(2.0 * xy - 1.0, 2.0 * nonLinearClipDepth - 1.0, 1.0);
    highp vec4 worldPos = sceneInvViewProjMatrix * clipPos;
    worldPos = worldPos / worldPos.w;
    return worldPos;
}

/**
    This obtains the view space depth from the provided view space vector and the view matrix.

    @param viewVector View space coordinates
    @param viewMatrix View matrix used originally as part of the calculation process for viewVector
    @return View space linear depth value
 */
highp float GetViewSpaceDepthFromPos(vec3 viewVector, mat4 viewMatrix)
{
    vec3 camLook = vec3(viewMatrix[0][2], viewMatrix[1][2], viewMatrix[2][2]);
    return dot(viewVector, camLook);
}

// Check allows this to be used outside of RT7 shaders
#if defined(SAMPLER_2D_AUTO_MULTISAMPLE)
    /**
        This converts a (non-linear) post-homogenous-divided NDC depth to a linear view-space position

        @param xy Screen space NDC coordinates in the range 0-1
        @param depthTexture The depth texture that contains the non-linear depth value to be converted
        @param sceneInvProjMatrix The inverse of the projection matrix used to create depthTexture
     */
    highp vec3 GetViewSpacePos(vec2 xy, SAMPLER_2D_AUTO_MULTISAMPLE depthTexture, highp mat4 sceneInvProjMatrix)
    {
        highp float nonLinearClipDepth;

    #if defined(VIEWPORTLOOKUPSCALE)
        nonLinearClipDepth = texture2DMultisample(depthTexture, xy, uViewportLookupScale.zw).r;
    #else
        nonLinearClipDepth = texture2DMultisample(depthTexture, xy, textureSize(depthTexture)).r;
    #endif

        return GetViewSpacePos(xy, nonLinearClipDepth, sceneInvProjMatrix);
    }

    highp vec3 GetViewSpacePos(vec2 xy, vec4 uvOffsetScale, vec2 texSize, SAMPLER_2D_AUTO_MULTISAMPLE depthTexture, highp mat4 sceneInvProjMatrix)
    {
        highp float nonLinearClipDepth = texture2DMultisample(depthTexture, (xy * uvOffsetScale.zw) + uvOffsetScale.xy, texSize).r;
        return GetViewSpacePos(xy, nonLinearClipDepth, sceneInvProjMatrix);
    }
#endif

#endif // DEPTH_UTILS_INC
