#ifndef LIGHTING_UTILS_H
#define LIGHTING_UTILS_H

#include "lighting/Lighting.inc"

/**
 * Rounds the luminance of the provided colour to 1.0 / 128.0,
 * mirroring the look of SD Oldscape, which stores colour as HSL
 * values with 7 luminance bits.
 */
vec3 DiscretizeLuminance(vec3 fragColour)
{
    float lightness = 0.5 * (max(fragColour.r, max(fragColour.g, fragColour.b)) + min(fragColour.r, min(fragColour.g, fragColour.b)));
    float newLightness = floor(128.0 * lightness) / 128.0;
    return fragColour * mix(newLightness / lightness, 1.0, lightness == 0.0);
}

#if !defined(DEFERRED_SHADOWS)

/**
    Helper that evalulates sunlight using shadow attenuation and the RT5 BRDFs
 */
LightingTerms EvaluateSunlightRT5(
    inout int iCascadeInOut,
    inout float amountInSunlightOut,
    highp vec4 worldPos,
    vec3 worldNormal,
    vec3 worldTangent,
    float viewDepth,
    vec3 viewDirection,
    float specPow,
    float emissiveLevel)
{
    float NDotL = max(0.0, dot(worldNormal, uInvSunDirection));
    float sunLightProportion = NDotL;

    amountInSunlightOut = 1.0;
#if defined(SUNLIGHT_SHADOWS)
    if (emissiveLevel == 0.0 && uMappingParams.z != 0.0)
    {
        // We only need to perform shadow attenuation test for pixels that are front facing to the sunlight
        if (sunLightProportion > 0.0)
        {
            highp vec4 sunlightViewPos = uSunlightViewMatrix * worldPos;

#if defined(USE_NORMAL_MAP)
            highp vec4 dudv = vec4(worldNormal.xy, 0.0, 0.0) * 32.0; // Most shadows projected onto normal mapped surfaces will be closer to the x/z plane
            amountInSunlightOut = DirLightShadowAtten(iCascadeInOut, worldPos + dudv, sunlightViewPos + dudv, viewDepth, uSunlightShadowMap, uSunlightShadowTranslucencyMap, 1.0);
#else
            amountInSunlightOut = DirLightShadowAtten(iCascadeInOut, worldPos, sunlightViewPos, viewDepth, uSunlightShadowMap, uSunlightShadowTranslucencyMap, 1.0);
#endif
        }
    }
#endif
    // Attenuate sunlight by shadow term
    sunLightProportion *= amountInSunlightOut;

    // Default RT5 model material base specular reflectance colour
    float specRefl = 0.65;
    LightingTerms directLighting;
    ClearLightingTerms(directLighting);
    EvaluateDirLightRT5(directLighting, worldNormal, worldTangent, viewDirection, uInvSunDirection, specRefl, specPow, 5.0, sunLightProportion, NDotL, uSunColour);
    return directLighting;
}

#else

/**
    Helper that evalulates sunlight using shadow attenuation from a buffer and the RT5 BRDFs
 */
LightingTerms EvaluateSunlightRT5(
    inout float amountInSunlightOut,
    vec3 worldNormal,
    vec3 worldTangent,
    vec3 viewDirection,
    vec2 fullscreenTexCoords,
    float specPow,
    float emissiveLevel)
{
    float NDotL = max(0.0, dot(worldNormal, uInvSunDirection));
    float sunLightProportion = NDotL;

    amountInSunlightOut = 1.0;
#if defined(SUNLIGHT_SHADOWS) && defined(DEFERRED_SHADOWS)
    if (emissiveLevel == 0.0 && uMappingParams.z != 0.0)
    {
        amountInSunlightOut = texture2DLod(uShadowBuffer, fullscreenTexCoords, 0.0).r;
    }
#endif
    // Attenuate sunlight by shadow term
    sunLightProportion *= amountInSunlightOut;

    // Default RT5 model material base specular reflectance colour
    float specRefl = 0.65;
    LightingTerms directLighting;
    ClearLightingTerms(directLighting);
    EvaluateDirLightRT5(directLighting, worldNormal, worldTangent, viewDirection, uInvSunDirection, specRefl, specPow, 5.0, sunLightProportion, uSunColour);
    return directLighting;
}

#endif

#if !defined(DEFERRED_SHADOWS)

/**
    Helper that evalulates sunlight using shadow attenuation and the RT7 BRDFs

    @param specPow Blinn-Phong specular power/shininess, range 0-8192
    @param perceptualRoughness Roughness from 0-1, needed for NDFs
 */
LightingTerms EvaluateSunlightRT7(
    inout int iCascadeInOut,
    inout float amountInSunlightOut,
    highp vec4 worldPos,
    vec3 worldNormal,
    vec3 worldTangent,
    float viewDepth,
    vec3 viewDirection,
    vec3 specRefl,
    float specPow,
    float perceptualRoughness,
    float fresnelExp)
{
    float NDotL = max(0.0, dot(worldNormal, uInvSunDirection));
    float sunLightProportion = NDotL;

    amountInSunlightOut = 1.0;
#if defined(SUNLIGHT_SHADOWS)
    if (uMappingParams.z != 0.0)
    {
        // We only need to perform shadow attenuation test for pixels that are front facing to the sunlight
        if (NDotL > 0.0)
        {
            highp vec4 sunlightViewPos = uSunlightViewMatrix * worldPos;
            highp vec4 dudv = vec4(worldNormal.xy, 0.0, 0.0) * 32.0; // 32 or 64 are in world units
            amountInSunlightOut = DirLightShadowAtten(iCascadeInOut, worldPos + dudv, sunlightViewPos + dudv, viewDepth, uSunlightShadowMap, uSunlightShadowTranslucencyMap, 1.0);
        }
    }
#endif
    // Attenuate sunlight by shadow term
    sunLightProportion *= amountInSunlightOut;

    LightingTerms directLighting;
    ClearLightingTerms(directLighting);

    EvaluateDirLightRT7(directLighting, worldNormal, worldTangent, viewDirection, uInvSunDirection, specRefl, specPow, perceptualRoughness, fresnelExp, sunLightProportion, NDotL, uSunColour);

    return directLighting;
}

#else

/**
    Helper that evalulates sunlight using shadow attenuation from a buffer and the RT7 BRDFs

    @param specPow Blinn-Phong specular power/shininess, range 0-8192
    @param perceptualRoughness Roughness from 0-1, needed for NDFs
 */
LightingTerms EvaluateSunlightRT7(
    inout float amountInSunlightOut,
    vec3 worldNormal,
    vec3 worldTangent,
    vec3 viewDirection,
    vec2 fullscreenTexCoords,
    vec3 specRefl,
    float specPow,
    float perceptualRoughness,
    float fresnelExp)
{
    float NDotL = max(0.0, dot(worldNormal, uInvSunDirection));
    float sunLightProportion = NDotL;

    amountInSunlightOut = 1.0;
#if defined(SUNLIGHT_SHADOWS) && defined(DEFERRED_SHADOWS)
    if (uMappingParams.z != 0.0)
    {
        amountInSunlightOut = texture2DLod(uShadowBuffer, fullscreenTexCoords, 0.0).r;
    }
#endif

    LightingTerms directLighting;
    ClearLightingTerms(directLighting);

    EvaluateDirLightRT7(directLighting, worldNormal, worldTangent, viewDirection, uInvSunDirection, specRefl, specPow, perceptualRoughness, fresnelExp, NDotL, sunLightProportion, uSunColour);

    return directLighting;
}

#endif


#endif
