#include "UndefineInOut.inc"

#if defined(TEXTURE_ATLAS)
    // viewVector can also be the world position, but using the camera position as the origin improves precision:
    // https://www.enkisoftware.com/devlogpost-20150131-1-Normal_generation_in_the_pixel_shader
    void CalculateDerivatives(inout highp vec2 uvDX, inout highp vec2 uvDY, inout highp vec3 positionDX, inout highp vec3 positionDY, highp vec2 uv, highp vec3 viewVector)
    {
        #if defined(STANDARD_DERIVATIVES)
            uvDX = dFdx(uv);
            uvDY = dFdy(uv);
            #if defined(PER_FRAGMENT_TANGENTS)
                // Note: these derivatives are noisy on certain models when viewed close to the camera - 'Fine' is used to help reduce this noise
                positionDX = dFdxFine(viewVector);
                positionDY = dFdyFine(viewVector);
            #endif
        #endif
    }

#if defined(RT7_MATERIAL)
    /**
        @param texel1 rgb: albedo
        @param texel2 r: emissive, gba: no-op normal (GBA = Y, Z, X)
        @param texel3 r: metalness, g: roughness
     */
    void SampleTexturesRT7(inout vec4 texel1, inout vec4 texel2, inout vec4 texel3, vec2 textureUV, TextureSettings textureSettings, float batchFlags, vec2 uvDX, vec2 uvDY)
    {
        #if defined(TEXTURE_ATLAS)
            #if !defined(TEXTURE_ALBEDO_GLOBAL)
                // Check forced texture flag if textures are off globally
                if (fract(batchFlags * 8.0) > 0.5)
                {
            #endif
                    getTexel(textureUV, textureSettings.textureMeta1, textureSettings.textureMeta2, textureSettings.textureMeta3, uAtlasMeta, uvDX, uvDY, textureSettings.wrapping, uTextureAtlas, texel1, texel2, texel3);
                    #if defined(ETC_CHANNEL_SWIZZLE)
                        // Swizzle ETC channels back into the expected order
                        texel2 = texel2.agbr; //NX,NY,NZ,HD -> HD,NY,NZ,NX
                        texel3 = texel3.rabg; //M,X,X,R -> M,R,X,X
                    #endif
            #if !defined(TEXTURE_ALBEDO_GLOBAL)
                }
            #endif
        #endif
    }
#else
    /**
        @param texelAlbedo rgb: albedo, a: alpha/specular/envmapping
        @param texelHDRNormal r: HDR, gba: no-op normal (GBA = Y, Z, X)
     */
    void SampleTextures(inout vec4 texelAlbedo, inout vec4 texelHDRNormal, vec2 textureUV, TextureSettings textureSettings, float batchFlags, vec2 uvDX, vec2 uvDY)
    {
        #if defined(TEXTURE_ATLAS)
            #if !defined(TEXTURE_ALBEDO_GLOBAL)
                // Check forced texture flag if textures are off globally
                if (fract(batchFlags * 8.0) > 0.5)
                {
            #endif
                //Only sample the second texture if we need the data
                #if defined(HDR_SCALE) || defined(USE_NORMAL_MAP)
                    getTexel(textureUV, textureSettings.textureMeta1, textureSettings.textureMeta2, uAtlasMeta, uvDX, uvDY, textureSettings.wrapping, uTextureAtlas, texelAlbedo, texelHDRNormal);
                    #if defined(ETC_CHANNEL_SWIZZLE)
                        // Swizzle ETC channels back into the expected order
                        texelHDRNormal = texelHDRNormal.agbr; //NX,NY,NZ,HD -> HD,NY,NZ,NX
                    #endif
                #else
                    getTexel(textureUV, textureSettings.textureMeta1, uAtlasMeta, uvDX, uvDY, textureSettings.wrapping, uTextureAtlas, texelAlbedo);
                #endif
            #if !defined(TEXTURE_ALBEDO_GLOBAL)
                }
            #endif

            #if !defined(TEXTURE_ALPHA_USAGE) && !defined(ALPHA_ENABLED) && !defined(GL_ES)
                // Scrap the alpha channel - compiler will optimise out throughout
                // The check for ES is due to the possibility of GEOMETRY_BATCH_ALPHA_BLEND_DISCARD
                texelAlbedo.a = 1.0;
            #endif

            #if defined(HDR_SCALE)
                texelAlbedo = HDRScale(texelAlbedo, texelHDRNormal.r); //We want to apply HDR scaling to the texture colour ONLY, not the vertex colour
            #endif
        #endif
    }
#endif

#endif

#include "DefineInOut.inc"
