#include "UndefineInOut.inc"
#include "Shared/ModelUtils.inc"

void AssignTextureAtlasVaryings(out vec2 textureUVOut, out vec2 materialSettingsSlotXYOut, vec2 textureUV, vec2 materialSettingsSlotXY)
{
    textureUVOut = textureUV.xy;
    materialSettingsSlotXYOut = materialSettingsSlotXY;
}

mat3 ResolveNormalTransformMatrix(mat4 transformMatrix)
{
    mat3 normalMatrix = Mat4ToMat3(transformMatrix);

#if defined(ANIMATION_VERTEX)
    //Bone transformations can sometimes result in 0.0 scales that result in NaN normals
    //Catch any 0.0 scales and replace with a default scale value of 1.0
    //We use step() because some platforms don't support branching in vertex shaders
    //This is equivalent to:
    //if (normalMatrix[0].x == 0.0 && normalMatrix[1].y == 0.0)
    //{
    //    normalMatrix[0].x = normalMatrix[1].y = 1.0;
    //}

    //This is the smallest possible epsilon for a highp float
    const highp float ZERO_EDGE = 2.0e-16;
    float normalFixRatio = step(ZERO_EDGE, abs(normalMatrix[0].x) + abs(normalMatrix[1].y));
    normalMatrix[0].x = mix(1.0, normalMatrix[0].x, normalFixRatio);
    normalMatrix[1].y = mix(1.0, normalMatrix[1].y, normalFixRatio);
#endif

    return normalMatrix;
}

// Converts a (-1,1) clipspace position to a UV-space (0..1) position
vec2 ClipSpacePosToUVSpacePos(vec2 clipSpacePos)
{
    return clipSpacePos.xy * 0.5 + vec2(0.5, 0.5);
}

// Offsets a world space position along a world space normal, with an enforced distance in pixels.
// The final offset is calculated by obtaining the final screenspace XY distance for the base and offset world space positions,
// normalizing it and then scaling it by the desired pixel distance. This target delta is then added onto the base clipspace
// position.
// Notes:
// - This function handles the clipspace position .w divide internally.
// - This is not a general use extrusion function as it preserves the original position's depth instead of using the offset
// position's depth. It's done this way in order to minimise Z-fighting artefacts.
// - An ideal solution to Z-fighting issues would be to flatten the model to a plane facing the camera at a specific model position
// (the model's centre would be ideal) after extrusion. That does add an additional expense however which is unnecessary in this case.
vec4 OffsetPositionAlongNormal(vec4 worldSpacePos, vec3 worldSpaceNormal, vec4 clipSpacePos, vec2 screenSize, float pixelDistance)
{
    // Branching here should have minimal impact since we're branching on a shader uniform and it will yield
    // the same result for all vertices
    if(pixelDistance <= 0.0)
    {
        return clipSpacePos;
    }
    else
    {
        vec2 uvSpacePos = clipSpacePos.xy / clipSpacePos.w;

        // Get the original pixel position
        vec2 pixelPos = ClipSpacePosToUVSpacePos(uvSpacePos) * screenSize;

        // Offset the world position along the inverse of the world normal and get the new pixel position
        // NOTE: We use the inverse to minimise depth issues from using the same depth for a different screen
        // position post offsetting.
        vec4 offsetWorldSpacePos = worldSpacePos;
        offsetWorldSpacePos.xyz -= worldSpaceNormal;

        vec4 offsetClipSpacePos = uViewProjMatrix * offsetWorldSpacePos;
        vec2 offsetPixelPos = ClipSpacePosToUVSpacePos(offsetClipSpacePos.xy / offsetClipSpacePos.w) * screenSize;

        // Reset back to the non-offset clipspace position
        offsetClipSpacePos.xy = uvSpacePos;

        // Get the normalised offset direction
        vec2 offsetDirection = offsetPixelPos - pixelPos;

        // Convert back the offset so it's a positive extrude op rather than a negative
        offsetDirection = -offsetDirection;

        // Avoid normalising a 0 length vector
        float offsetDirectionLenSq = dot(offsetDirection, offsetDirection);

        if (offsetDirectionLenSq > 0.0)
        {
            offsetDirection /= sqrt(offsetDirectionLenSq);

            // Apply the pixel offset
            offsetClipSpacePos.xy += (offsetDirection * pixelDistance) / screenSize;
        }

        // Ensure the screenspace position remains correct after perspective correction
        offsetClipSpacePos.xy *= offsetClipSpacePos.w;

        return offsetClipSpacePos;
    }
}

#if defined(ANIMATION_SKINNED_VERTEX)
    mat4 GetBoneMatrix(out float labelOut)
    {
        labelOut = floor(aVertexPosition_BoneLabel.w / 256.0);

        float boneIndex = aVertexPosition_BoneLabel.w - labelOut * 256.0;
        mat4 boneMatrix = GetBoneFromIndex(boneIndex);
        mat4 jointMatrix = GetBoneFromIndex(aVertexSkinBones.x);

        // Note: There's a bug on Intel HD 4000 series (GT-3425) where not storing the result of calling
        // GetBoneFromIndex into a local variable would result in extreme geometry artefacts.
        //
        // Best guess is the shader compiler on the drivers for those iGPUs is so bad that it tries
        // to store each result of calling GetBoneFromIndex into separate registers. Then, since we're
        // dealing with matrices here, it quickly runs out of available registers instead of reusing them
        // and ends up doing weird stuff, resulting in artefacts.
        if (uSmoothSkinning > 0.0)
        {
            jointMatrix *= aVertexSkinWeights.x;

            mat4 curBone = GetBoneFromIndex(aVertexSkinBones.y);
            jointMatrix += curBone * aVertexSkinWeights.y;

            curBone = GetBoneFromIndex(aVertexSkinBones.z);
            jointMatrix += curBone * aVertexSkinWeights.z;

            curBone = GetBoneFromIndex(aVertexSkinBones.w);
            jointMatrix += curBone * aVertexSkinWeights.w;
        }

        return jointMatrix * boneMatrix;
    }
#elif defined(ANIMATION_VERTEX)
    mat4 GetBoneMatrix(out float labelOut)
    {
        labelOut = floor(aVertexPosition_BoneLabel.w / 256.0);
        mat4 boneMatrix = GetBoneFromIndex(aVertexPosition_BoneLabel.w - labelOut * 256.0);
        return boneMatrix;
    }
#endif

#if defined(ANIMATION_SKINNED_VERTEX)
    //Rigid bone matrix only uses first skin entry or the older boneLabel system for legacy animations - This is primarily for shadows as smoothskinning doesn't add any value there
    mat4 GetBoneMatrixRigid(out float labelOut)
    {
        labelOut = floor(aVertexPosition_BoneLabel.w / 256.0);
        float boneIndex = aVertexPosition_BoneLabel.w - labelOut * 256.0;
        mat4 boneMatrix = GetBoneFromIndex(boneIndex);
        mat4 jointMatrix = GetBoneFromIndex(aVertexSkinBones.x);
        return jointMatrix * boneMatrix;
    }
#elif defined(ANIMATION_VERTEX)
    //Rigid bone matrix only uses first skin entry or the older boneLabel system for legacy animations - This is primarily for shadows as smoothskinning doesn't add any value there
    mat4 GetBoneMatrixRigid(out float labelOut)
    {
        labelOut = floor(aVertexPosition_BoneLabel.w / 256.0);
        float boneIndex = aVertexPosition_BoneLabel.w - labelOut * 256.0;
        mat4 boneMatrix = GetBoneFromIndex(boneIndex);
        return boneMatrix;
    }
#endif

void AssignPositionNormalVaryings(out float labelOut)
{
    vec3 vertexPosition = aVertexPosition_BoneLabel.xyz;
    mat4 modelAndSkinMatrix;

#if defined(ANIMATION_VERTEX)
    // We must store the bone matrix in a temporary register first instead of multiplying by an in-place temporary,
    // otherwise certain drivers (i.e. ATI) have been known to crash
    mat4 boneMatrix = GetBoneMatrix(labelOut);
    mat4 modelMatrix = uModelMatrix;

    modelAndSkinMatrix = modelMatrix * boneMatrix;
#else
    labelOut = 0.0; // Keep Angle happy

    modelAndSkinMatrix = uModelMatrix;
#endif

    // Transform position
    vec3 scaledVertexPos = vertexPosition * uVertexScale;
    vec4 worldSpacePos = modelAndSkinMatrix * vec4(scaledVertexPos, 1.0);
    vec4 clipSpacePos = uViewProjMatrix * worldSpacePos;

    // Apply non-shadow rendering transforms
#if !defined(MODEL_GEOMETRY_SHADOW_VS)
    // Transform normal and output to varying
    // Also check for and fix 0.0 matrix scales (after transforming the position so that only normals are affected)
    mat3 modelAndSkinMat3 = ResolveNormalTransformMatrix(modelAndSkinMatrix);
    vNormal = modelAndSkinMat3 * aVertexNormal_BatchFlags.xyz;
    vNormal = normalize(vNormal);

    // Handle edge case where background objects can sometimes have a pixel-sized vert with NaN normal
    vNormal = isNaN(vNormal.x) ? vec3(0.0, 1.0, 0.0) : vNormal;

    // Transform tangent and output to varying
    #if defined(USE_NORMAL_MAP) && !defined(PER_FRAGMENT_TANGENTS)
        vTangent.xyz = modelAndSkinMat3 * aVertexTangent.xyz;
        vTangent.xyz = normalize(vTangent.xyz);
        vTangent.xyz = isNaN(vTangent.x) ? vec3(1.0, 0.0, 0.0) : vTangent.xyz;
        vTangent.w = aVertexTangent.w;
    #endif

    // Extrude the position along the normals in screenspace
    #if defined(MODEL_GEOMETRY_HIGHLIGHT_VS)
        clipSpacePos = OffsetPositionAlongNormal(worldSpacePos, vNormal, clipSpacePos, uViewportLookupScale.zw, uHighlightScale);
    #elif defined(MODEL_GEOMETRY_DEPTH_BIAS)
        #if defined(USE_NORMAL_MAP)
            #error Cannot combine USE_NORMAL_MAP and MODEL_GEOMETRY_DEPTH_BIAS as the use the tanget for different things
        #endif
        // aVertexTangent goes from 0-1 in 1/128 increments. dividing by 8 produces a depth bias of
        // 1/1024 increments in clip space, which ranges from 0-8192. This makes 1/1024 effectively
        // one float quantum.
        clipSpacePos.z -= aVertexTangent.x / 8.0;
    #endif

    vWorldPosition = worldSpacePos.xyz;
#endif


    gl_Position = clipSpacePos;

#if !defined(ANIMATION_VERTEX) && defined(PUSH_TO_FARPLANE)
    gl_Position.z = gl_Position.w * 0.9999;
#endif
}

#include "DefineInOut.inc"
