

void ComputeTangentBitangentFromDerivatives(
    inout vec3 tangentOut,
    inout vec3 bitangentOut,
    highp vec3 vertexNormal,
    highp vec3 positionDX,
    highp vec3 positionDY,
    highp vec2 uvDX,
    highp vec2 uvDY)
{
    // Method copied from: https://web.archive.org/web/20180802200253/http://www.thetenthplanet.de/archives/1180
    highp vec3 positionDXPerp = cross(vertexNormal, positionDX);
    highp vec3 positionDYPerp = cross(positionDY, vertexNormal);
    highp vec3 tangent = positionDYPerp * uvDX.x + positionDXPerp * uvDY.x;
    highp vec3 bitangent = positionDYPerp * uvDX.y + positionDXPerp * uvDY.y;

    highp float tangentDot = dot(tangent, tangent);
    highp float bitangentDot = dot(bitangent, bitangent);
    highp float tbMax = max(tangentDot, bitangentDot);
    highp float invMax = inversesqrt(tbMax);
    tangent *= invMax;
    bitangent *= invMax;

    if (isNaN(tangentDot + bitangentDot) || tbMax <= 0.0)
    {
        tangent = bitangent = vertexNormal;
    }

    tangentOut = tangent;
    bitangentOut = bitangent;
}

// Calculate bitangent using normal and tangent
vec3 ComputeBitangent(vec3 vertexNormal, vec4 tangent)
{
    highp vec3 bitangent = cross(vertexNormal, tangent.xyz);
    // Handedness of bitangent is stored in w component of tangent attribute
    bitangent *= tangent.w;
    return bitangent;
}

vec3 ApplyNormalMap(highp vec3 textureNormal, highp vec3 vertexNormal, highp vec3 positionDX, highp vec3 positionDY, highp vec2 uvDX, highp vec2 uvDY)
{
    highp vec3 tangent;
    highp vec3 bitangent;
    ComputeTangentBitangentFromDerivatives(tangent, bitangent, vertexNormal, positionDX, positionDY, uvDX, uvDY);

#if defined(DEBUG_TANGENTS)
    gl_FragColor.rgb = (normalize(tangent) * 0.5 + 0.5);
    gl_FragColor.a = 1.0;
#endif
#if defined(DEBUG_BITANGENTS)
    gl_FragColor.rgb = (normalize(bitangent) * 0.5 + 0.5);
    gl_FragColor.a = 1.0;
#endif

    // This is an optimisation - the full TBN matrix can be calculated with: mat3(tangent * invMax, bitangent * invMax, vertexNormal)
    highp vec3 result = (textureNormal.x * tangent) + (textureNormal.y * bitangent) + (textureNormal.z * vertexNormal);
    result = normalize(result);

    // We swap back to the original normal if the result has come out as all zeroes/NaNs.
    // This will happen in content where UV 0 has been used to show vertex colour only.
    // Note: this must be implemented as a conditional assignment, mix() will always
    // return zero if any parameter is NaN.
    // We are using two methods of detecting NaN here, the first validating that normalize()
    // has worked (for Angle/ES 2) and the second trying to detect NaN directly (for non-Angle).
    // We also validate tbMax is > 0 as that error condition appears to happen separately on some content.
    return abs(result.x) + abs(result.y) + abs(result.z) < 0.5 ? vertexNormal : result;
}

vec3 ApplyNormalMap(vec3 textureNormal, vec3 vertexNormal, vec3 tangent, vec3 bitangent)
{
#if defined(DEBUG_TANGENTS)
    gl_FragColor.rgb = (tangent * 0.5 + 0.5);
    gl_FragColor.a = 1.0;
#endif
#if defined(DEBUG_BITANGENTS)
    gl_FragColor.rgb = (bitangent * 0.5 + 0.5);
    gl_FragColor.a = 1.0;
#endif

    highp vec3 result = (textureNormal.x * tangent) + (textureNormal.y * bitangent) + (textureNormal.z * vertexNormal);
    result = normalize(result);
    return result;
}

vec3 ApplyNormalMap(vec3 textureNormal, vec3 vertexNormal, vec4 tangent)
{
    // Calculate bitangent using normal and tangent
    vec3 bitangent = ComputeBitangent(vertexNormal, tangent);
    // We intentionally skip normalizing the bitangent as it was introducing NaNs when doing so
    return ApplyNormalMap(textureNormal, vertexNormal, tangent.xyz, bitangent);
}

// Special optimised version for terrain where we can make assumptions about the input
// i.e. that UVs are directly related to world space position and are always valid
// This also uses highp because terrain positions derivatives can be huge for background pixels
vec3 ApplyNormalMapTerrain(vec3 textureNormal, highp vec3 vertexNormal, highp vec3 positionDX, highp vec3 positionDY)
{
    highp vec3 positionDXPerp = cross(vertexNormal, positionDX);
    highp vec3 positionDYPerp = cross(positionDY, vertexNormal);
    highp vec3 tangent = positionDYPerp * positionDX.x + positionDXPerp * positionDY.x;
    highp vec3 bitangent = positionDYPerp * positionDX.z + positionDXPerp * positionDY.z;

    highp float invMax = inversesqrt(max(dot(tangent, tangent), dot(bitangent, bitangent)));
    tangent *= invMax;
    bitangent *= invMax;

#if defined(DEBUG_TANGENTS)
    gl_FragColor.rgb = (normalize(tangent) * 0.5 + 0.5);
    gl_FragColor.a = 1.0;
#endif
#if defined(DEBUG_BITANGENTS)
    gl_FragColor.rgb = (normalize(bitangent) * 0.5 + 0.5);
    gl_FragColor.a = 1.0;
#endif

    highp vec3 result = (textureNormal.x * tangent) + (textureNormal.y * bitangent) + (textureNormal.z * vertexNormal);
    result = normalize(result);
    // The isNaN check here should not be necessary, but some old AMD cards randomly return bad derivatives resulting in black terrain otherwise
    return isNaN(result.x) ? vertexNormal : result;
}

// Optimised terrain normal map calculation, exploiting the fact that terrain can never be truly vertical,
// as terrain verts are mapped on a XZ grid (i.e. steep gradient is possible, >= 90 degree gradient is not).
// Therefore we can guarantee a vector pointing directly in any XZ direction will be orthogonal enough for normal mapping.
// Also terrain texture UVs map to positions and can't be rotated, so the tangent space is always oriented the same direction.
vec3 ApplyNormalMapTerrain(vec3 textureNormal, vec3 vertexNormal)
{
    // Logical 'right' direction, selected due to matching previous derivative-based algorithm output
    const vec3 right = vec3(0.0, 0.0, 1.0);
    vec3 tangent = cross(right, vertexNormal);
    vec3 bitangent = cross(tangent, vertexNormal);

    return ApplyNormalMap(textureNormal, vertexNormal, tangent, bitangent);
}
