#if defined(POINT_LIGHTING)

#if defined(POINT_LIGHTING_SPECULAR)
    //Used to calculate specular for pointlights
    #include "lighting/brdf/Fresnel.inc"
    #include "lighting/brdf/BRDF.inc"
#endif

    //Common utility methods
    highp float UnpackPointLightFloat(vec4 textureData)
    {
        return UnpackRGBAToIntegerFloat(textureData);
    }

    vec3 UnpackPointLightColour(vec4 textureData)
    {
        return textureData.xyz;
    }

    vec4 UnpackPointLightIndices(vec4 textureData)
    {
        return textureData * 255.0;
    }

    void EvaluatePointLights(
        inout vec3 pointLightsTerm,
        inout vec3 pointLightsSpecular,
        in vec3 specRefl,
        in float specPow,
        in float perceptualRoughness,
        in float fresnelExp,
        in highp vec3 viewVec,
        in highp vec3 worldPos,
        in vec3 worldNormal,
        in vec2 indexSamplePosition,
        in float level,
        in highp float pixelViewDepth)
    {
        //Common constant divisors
        const highp float inv5 = 1.0 / 5.0;
        const highp float inv8 = 1.0 / 8.0;
        const highp float inv128 = 1.0 / 128.0;
        const highp float inv256 = 1.0 / 256.0;
        const highp float inv255 = 1.0 / 255.0;
        const highp float inv510 = 1.0 / 510.0;
        const highp float inv512 = 1.0 / 512.0;
        const highp float inv65535 = 1.0 / 65535.0;
        const highp float lightDataStartUV_X = 128.5 / 256.0;
        const int UINT16_MINUS1 = 65535;

        //Some constants for where light data is sampled from
        const vec2 floatSampleOffsetY = vec2(inv256, 0.0);
        const vec2 floatSampleOffsetZ = vec2(2.0 * inv256, 0.0);
        const vec2 floatSampleOffsetRadius = vec2(3.0 * inv256, 0.0);
        const vec2 floatSampleOffsetColour = vec2(4.0 * inv256, 0.0);
        const vec2 floatSampleOffsetAttenuationAndBacklight = vec2(5.0 * inv256, 0.0);

        //Common variables used in the loops
        highp float pointLightProportion;
        highp float lightIntensity, lightRadius, attenuation, distance, sampleDataX, sampleDataY;
        highp vec2 samplePosition;
        highp vec3 lightVec, lightOffset, lightPosition, halfwayVec, specFresnel;
        vec3 lightColour, lightSpec;
        vec4 lightAttenuationAndBacklight;
    #if defined(POINT_LIGHT_INDICES_IS_OGLES_BACKEND)
        highp float outerIndex, innerIndex, innerIndexSub, lowIntensity, highIntensity;
    #endif

        //Offset the sample position to the correct level quadrant in the texture
        indexSamplePosition.x += step(0.01, fract(level * 0.5)) * 0.25;
        indexSamplePosition.y += step(1.01, level) * 0.5;

        //Same the light indices for this tile
        vec4 indices = UnpackPointLightIndices(texture2DLodCompat(uPointLightIndicesMap, indexSamplePosition, 0.0));

        //Loop for four lights, reading position, radius and colour from the texture and light intensity from the uniforms array
        //NOTE: This loop produces better code on GCN without being explicitly unrolled
        for (int iLight = 0; iLight < 4; ++iLight)
        {
            float index = indices[iLight];
            //Lights are packed sequentially, so we can break when we hit our first null (0.0) light
            if (index == 0.0)
            {
                break;
            }
            //Intensity
        #if defined(POINT_LIGHT_INDICES_IS_OGLES_BACKEND)
            //Adjust the index to access the correct vec4 component
            outerIndex = floor(index * inv8);
            innerIndex = index - (outerIndex * 8.0);
            innerIndexSub = floor(innerIndex * 0.5);
            lightIntensity = uPointLightData[int(outerIndex)][int(innerIndexSub)];
            //There are two intensities per float, get the one we need
            highIntensity = floor(lightIntensity * inv512); //Bit-shift by 9
            lowIntensity = lightIntensity - (highIntensity * 512.0);
            //0.0 (low) or 1.0 (high)
            innerIndexSub = innerIndex - (innerIndexSub * 2.0);
            //Select the correct intensity, and put back in the expected range
            lightIntensity = ((highIntensity * step(0.5, innerIndexSub)) + (lowIntensity * step(innerIndexSub, 0.5))) * inv5;
        #else
            lightIntensity = uPointLightData[int(index)];
        #endif
            //Light data sample position
            sampleDataX = lightDataStartUV_X + (step(128.0, index) * 0.25);
            sampleDataY = fract((index + 0.5) * inv128);
            samplePosition = vec2(sampleDataX, sampleDataY);

            //Position
            lightPosition.x = UnpackPointLightFloat(texture2DLodCompat(uPointLightIndicesMap, samplePosition, 0.0));

            //Branch based on the texture sample method available
            int shadowIndex = UINT16_MINUS1;
        #if defined(POINT_LIGHT_INDICES_USE_TEXEL_OFFSETS)
            lightPosition.y = UnpackPointLightFloat(textureLodOffset(uPointLightIndicesMap, samplePosition, 0.0, ivec2(1, 0)));
            lightPosition.z = UnpackPointLightFloat(textureLodOffset(uPointLightIndicesMap, samplePosition, 0.0, ivec2(2, 0)));
            vec4 lightRadiusShadowIndex = textureLodOffset(uPointLightIndicesMap, samplePosition, 0.0, ivec2(3, 0));
            shadowIndex = int(UnpackVec2ToFloat(lightRadiusShadowIndex.zw));
            lightRadius = UnpackVec2ToFloat(lightRadiusShadowIndex.xy);
            lightRadius = (lightRadius * inv65535) * POINT_LIGHT_INDICES_RADIUS_MAX;
            lightColour = clamp(UnpackPointLightColour(textureLodOffset(uPointLightIndicesMap, samplePosition, 0.0, ivec2(4, 0))), 0.0, 1.0);
            lightAttenuationAndBacklight = textureLodOffset(uPointLightIndicesMap, samplePosition, 0.0, ivec2(5, 0));
        #else
            lightPosition.y = UnpackPointLightFloat(texture2DLodCompat(uPointLightIndicesMap, samplePosition + floatSampleOffsetY, 0.0));
            lightPosition.z = UnpackPointLightFloat(texture2DLodCompat(uPointLightIndicesMap, samplePosition + floatSampleOffsetZ, 0.0));
            vec4 lightRadiusShadowIndex = texture2DLodCompat(uPointLightIndicesMap, samplePosition + floatSampleOffsetRadius, 0.0);
            shadowIndex = int(UnpackVec2ToFloat(lightRadiusShadowIndex.zw));
            lightRadius = UnpackVec2ToFloat(lightRadiusShadowIndex.xy);
            lightRadius = (lightRadius * inv65535) * POINT_LIGHT_INDICES_RADIUS_MAX;
            lightColour = clamp(UnpackPointLightColour(texture2DLodCompat(uPointLightIndicesMap, samplePosition + floatSampleOffsetColour, 0.0)), 0.0, 1.0);
            lightAttenuationAndBacklight = texture2DLodCompat(uPointLightIndicesMap, samplePosition + floatSampleOffsetAttenuationAndBacklight, 0.0);
        #endif

            lightOffset = lightPosition - worldPos;
            lightVec = normalize(lightOffset);
            //Put the attenuation into the correct range
            float attenuationFalloff = (UnpackVec2ToFloat(lightAttenuationAndBacklight.xy) * inv65535) * POINT_LIGHT_INDICES_ATTENUATION_FALLOFF_MAX;
            float backlight = lightAttenuationAndBacklight.z;

            vec3 lightVecRatio = lightOffset / lightRadius;
            float lightDistRatioSqr = dot(lightVecRatio, lightVecRatio);
            attenuation = pow(1.0 - min(1.0 - inv65535, lightDistRatioSqr), attenuationFalloff);

            //Lambert
            //We clamp nDotL by backlight instead of 0.0, to allow light to spill around objects
            float clampMin = -backlight; //Remember to negate the input value
            highp float nDotL = dot(worldNormal, lightVec);
            pointLightProportion = (clamp(nDotL, clampMin, 1.0) - clampMin) / (1.0 - clampMin);

            // Shadow attenuation
            float amountInLight = 1.0;
        #if defined(POINT_LIGHT_SHADOWS)
            if ((shadowIndex != UINT16_MINUS1))
            {
                PointLightSMConsts pointLightSMConsts = uPointLightSMConsts[shadowIndex]; //@todo No way to avoid this divergent uniform access :(, or is there?
                if (pointLightSMConsts.EnabledFadeViewParams.x != 0.0)
                {
                    // We only need to perform shadow attenuation test for pixels that are front facing to the light
                    if (pointLightProportion > 0.0)
                    {
                        // Bias light offset along the position of the surface normal for additional shadow bias, and keep increasing bias with attenuation.
                        // As we get further away from the origin of the light source, shadow map sampling accuracy decreases resulting in an increase in shadow acne.
                        // Have the normal bias increase as we get further away from the origin of the light source to further reduce shadow acne
                        lightOffset = lightOffset + (worldNormal * (pointLightSMConsts.FilterBiasParams.y + ((1.0 - attenuation) * pointLightSMConsts.FilterBiasParams.y)));
                        amountInLight = PointLightShadowAtten(
                            lightVec, pixelViewDepth, dot(lightOffset, lightOffset), 1.0 / (lightRadius * lightRadius), pointLightSMConsts, uPointLightShadowMap);
                    }
                }
            }
        #endif

            //Add to the total term
            pointLightsTerm += (lightColour * lightIntensity * attenuation * pointLightProportion) * amountInLight;

            // Calculate specular
        #if defined(POINT_LIGHTING_SPECULAR)
            nDotL = max(0.0, nDotL);
            // Calculate half vector and fresnel term
            halfwayVec = normalize(viewVec + lightVec);
            // We need to clamp the dot product to protect against NaNs forming causing black patches to form
            specFresnel = FresnelSchlick(specRefl, clamp(dot(lightVec, halfwayVec), 0.0, 1.0), fresnelExp); // More physically correct l dot h for punctual lights
            #if defined(RT7_MATERIAL)
                lightSpec = CookTorranceBRDF(specPow, perceptualRoughness, specFresnel, worldNormal, halfwayVec, viewVec, lightVec, nDotL);
            #else
                lightSpec = specFresnel * vec3(RunescapeRT5BRDF(halfwayVec, worldNormal, specPow));
            #endif
            // Add to the total specular
            pointLightsSpecular += (lightSpec * lightColour * attenuation * nDotL * lightIntensity) * amountInLight;
        #endif
        }

        ///////////////////////////////////////////////////////////////////////////
        //MISC DEBUG

        //vec3 levelColour;
        //if (level < 1.0) levelColour = vec3(1,0,0);
        //else if (level < 2.0) levelColour = vec3(0,0,1);
        //else if (level < 3.0) levelColour = vec3(1,0,1);
        //else levelColour = vec3(1,1,0);
        //pointLightsTerm.xyz = levelColour;

        //pointLightsTerm.xyz = vec3(localTilePosition.xy / 63.0, level / 3.0);
        //pointLightsTerm.xyz = vec3(level / 3.0, 0.0, 0.0);
        //pointLightsTerm.xyz = vec3(localTilePosition.xy / 63.0, 0);
    }

    #if defined(POINT_LIGHTING_VERTEX_DATA)
        void EvaluatePointLights(
            inout vec3 pointLightsTerm,
            inout vec3 pointLightsSpecular,
            in vec3 specRefl,
            in float specPow,
            in float perceptualRoughness,
            in float fresnelExp,
            in highp vec3 viewVec,
            in highp vec3 worldPos,
            in vec3 worldNormal,
            in highp float pixelViewDepth,
            in vec3 tilePosition)
        {
            //Common constant divisors
            const vec2 texelToUV = vec2(1.0 / 256.0, 1.0 / 128.0);

            //Tile position is provided by vertex data in the function parameter
            tilePosition = floor(tilePosition + vec3(0.1));
            vec2 localTilePosition = tilePosition.yz;
            vec2 samplePosition = (localTilePosition + vec2(0.5, 0.5)) * texelToUV;
            float level = tilePosition.x;

            EvaluatePointLights(pointLightsTerm, pointLightsSpecular, specRefl, specPow, perceptualRoughness, fresnelExp, viewVec, worldPos, worldNormal, samplePosition, level, pixelViewDepth);
        }
    #else
        void EvaluatePointLights(
            inout vec3 pointLightsTerm,
            inout vec3 pointLightsSpecular,
            in vec3 specRefl,
            in float specPow,
            in float perceptualRoughness,
            in float fresnelExp,
            in highp vec3 viewVec,
            in highp vec3 worldPos,
            in vec3 worldNormal,
            in highp float pixelViewDepth)
        {
            //Common constant divisors
            const vec2 texelToUV = vec2(1.0 / 256.0, 1.0 / 128.0);
            const float inv256 = 1.0 / 256.0;
            const float inv512 = 1.0 / 512.0;
            //The constants are provided via defines in the cpp source
            const highp float invTileSize = POINT_LIGHT_INDICES_INV_TILE_FINESIZE;
            const highp float mapSize = POINT_LIGHT_INDICES_MAPSQUARE_FINESIZE;
            const highp float invMapSize = POINT_LIGHT_INDICES_INV_MAPSQUARE_FINESIZE;

            //Get the tile we are in for this map square
            //If the tile position has not been provided then we need to calculate it from the world position of this pixel
            //This position is also clamped within bounds provided by a uniform
            highp vec2 clampedWorldPos = clamp(worldPos.xz, uPointLightIndicesClampLimits.xy, uPointLightIndicesClampLimits.zw);
            highp vec2 localTilePosition = floor((fract(clampedWorldPos * invMapSize) * mapSize) * invTileSize);
            highp vec2 samplePosition = (localTilePosition + vec2(0.5, 0.5)) * texelToUV;
            samplePosition = clamp(samplePosition, vec2(inv512, inv256), vec2(0.25 - inv512, 0.5 - inv256));
            float level = uPointLightIndicesLevel;

            EvaluatePointLights(pointLightsTerm, pointLightsSpecular, specRefl, specPow, perceptualRoughness, fresnelExp, viewVec, worldPos, worldNormal, samplePosition, level, pixelViewDepth);
        }
    #endif

#endif
