#ifndef REFRACTION_INC
#define REFRACTION_INC

#define REFRACTION_DEPTH_CHECK_ENABLED 1
#define USE_REFRACTED_WORLD_POS 0 // This is only really to going be useful for refractive objects like thick crystal, ice statues and deep pools of water

/**
    @param refrParams x: refraction normal scale, y: index of refraction, z: refraction vector world space depth bias, w: empty

    @return The refracted colour
 */
vec3 CalculateRefractionColour(
    highp vec3 worldPos,
    vec3 worldNormal,
    highp vec3 viewVec,
    float invRoughness,
    vec4 refrParams,
    vec2 refrTexCoordsDuDv)
{
    // Calculate refracted world pos
#if USE_REFRACTED_WORLD_POS
    vec3 refrVec = refract(viewVec, -worldNormal, refrParams.y);
    highp vec3 refrWorldPos = worldPos + (refrVec * refrParams.z);
#else
    highp vec3 refrWorldPos = worldPos;
#endif

    // Applying this delta to the world space position eliminates various artefacts when applying directly to the final texture coordinates
    highp vec2 dudv = refrTexCoordsDuDv * refrParams.w;

    // Convert refrWorldPos to clip space
    vec4 refrTexCoords = uViewProjMatrix * vec4(refrWorldPos.xy + dudv, refrWorldPos.z, 1.0);
    refrTexCoords.xy /= refrTexCoords.w;

    // Convert from clip space to normalised device coordinates
    refrTexCoords.xy = refrTexCoords.xy * 0.5 + 0.5;

#if REFRACTION_DEPTH_CHECK_ENABLED
    // If the depth in the refraction map is below our surface take an unperturbed refraction sample
    float refractionDepth = texture2DLodCompat(uRefractionDepth, refrTexCoords.xy, 0.0).r;
    if (refractionDepth < gl_FragCoord.z /*|| refractionDepth >= 1.0*/) // The foreground checks causes duplicate images to appear against sky
    {
        vec2 viewportOffset = uViewportOffsetScale.xy * uFullScreenLookupScale.zw;
        refrTexCoords.xy = (gl_FragCoord.xy - viewportOffset) * uViewportLookupScale.xy;
    }
#endif

    // Select mip level to use based on surface roughness
    float mipLevel = max(0.0, uRefractionViewportLookupScale.x - (invRoughness * uRefractionViewportLookupScale.x));
    return texture2DLodCompat(uRefractionMap, vec3(refrTexCoords.xy, mipLevel), 0.0).rgb;
}

/**
    Calculates the refraction colour and combines current fragment colour
 */
vec3 CalculateRefractionColour(
    highp vec3 worldPos,
    vec3 worldNormal,
    highp vec3 viewVec,
    float invRoughness,
    vec4 refrParams,
    vec2 refrTexCoordsDuDv,
    vec4 fragColour)
{
    vec3 refrColour = CalculateRefractionColour(worldPos, worldNormal, viewVec, invRoughness, refrParams, refrTexCoordsDuDv);
    return mix(refrColour, fragColour.rgb, fragColour.a);
}

/**
    Enhanced version of CalculateRefractionColour() that combines current fragment colour and indirect specular
    for a more realistic final rendering (i.e. glass will also show reflections at glancing angles, more physically correct)
 */
vec3 CalculateRefractionColour(
    highp vec3 worldPos,
    vec3 worldNormal,
    highp vec3 viewVec,
    float invRoughness,
    vec4 refrParams,
    vec2 refrTexCoordsDuDv,
    vec4 fragColour,
    vec3 indirectSpecular,
    vec3 envFresnel)
{
    vec3 refrColour = CalculateRefractionColour(worldPos, worldNormal, viewVec, invRoughness, refrParams, refrTexCoordsDuDv);
    refrColour = mix(refrColour, fragColour.rgb, fragColour.a);
    return mix(refrColour, indirectSpecular, envFresnel);
}



#endif
