#ifndef STIPPLE_CUTOUT_UTILS_INC
#define STIPPLE_CUTOUT_UTILS_INC

#if defined(STIPPLE_TRANSPARENCY_CUTOUT)

#include "Shared/StippleCommon.inc"
#include "Shared/NoiseUtils.inc"

// The stipple cutout effect can be applied to alpha-blended and alpha-tested geometry, which necessitates different
// approaches (adjusting the alpha for the former, using discard for the later). To have both visuals match slightly
// better, this multiplier is applied to the minimum alpha used when calculating the stipple alpha for alpha-tested
// geometry.
#define STIPPLE_CUTOUT_STIPPLED_ALPHA_MULTIPLIER 2.0

float GetStippleCutoutAdjustedAlpha(vec3 viewSpacePosition, vec2 pixelUV, float alphaMultiplier)
{
    float vsDepthCamera = uStippleCutoutPosDepthVisibility.z;
    float vsDepthModel = GetStippleViewSpaceDepthFromPos(viewSpacePosition);

    // Note: This is extremely basic and results in some clipping for some objects
    // TODO?: Base this on the min and max VS depth for the player's AABB. This would be more correct overall but may not have a huge impact.
    const float DEPTH_OFFSET = 500.0;
    const float DEPTH_SMOOTH_RANGE = 500.0;
    const float DEPTH_SMOOTH_RANGE_FACTOR = 1.0 / DEPTH_SMOOTH_RANGE;

    float result = 0.f;

    if (vsDepthModel < (vsDepthCamera + DEPTH_OFFSET))
    {
        // Get the depth factor so we can smoothly apply the effect on things without a harsh cutoff as they exceed a certain depth
        float depthFactor = abs(vsDepthModel - (vsDepthCamera + DEPTH_OFFSET));
        depthFactor = clamp(depthFactor * DEPTH_SMOOTH_RANGE_FACTOR, 0.0, 1.0);

        // Convert screen coords from a [0..1] range to [1..0..1]
        vec2 posRescaledUV = (pixelUV * 2.0) - vec2(1.0);
        vec2 posEffectSpace = posRescaledUV - uStippleCutoutPosDepthVisibility.xy;
        posEffectSpace = posEffectSpace;

        // Take into account the aspect ratio to preserve the spherical appearance of the effect
        float aspectRatio = uProjectionMatrix[1][1] / uProjectionMatrix[0][0];
        posEffectSpace.x *= aspectRatio;

        // Documenting constants
        float xrayFocusVisibility = uStippleCutoutPosDepthVisibility.w;
        float xrayMinAlphaRange = uStippleCutoutStartRangeAndMinAlpha.x;
        float xraySmoothOneOverFadeRange = uStippleCutoutStartRangeAndMinAlpha.y;
        float xrayMinAlpha = uStippleCutoutStartRangeAndMinAlpha.z;

        // Note: We may want to represent the player shape as a capsule on screen instead.
        float xrayAmount = clamp((length(posEffectSpace) - xrayMinAlphaRange) * xraySmoothOneOverFadeRange, 0.0, 1.0);

        // Fade the alpha from 0 to 1 based on distance from the centre of the screen
        result = mix(pow(depthFactor * (1.0 - xrayAmount), 2.0), 0.0, xrayFocusVisibility) - xrayMinAlpha;
    }

    return result;
}

// Checks if the specified pixel is stipple'd out, returning true if it should be discarded.
bool IsStippleCutoutVisible(vec3 viewSpacePosition, vec2 pixelUV, vec2 screenPos)
{
    return GetStippleCutoutAdjustedAlpha(viewSpacePosition, pixelUV, STIPPLE_CUTOUT_STIPPLED_ALPHA_MULTIPLIER) > GetInterleavedGradientNoise(screenPos);
}

#endif

#endif // STIPPLE_CUTOUT_UTILS_INC
