#include "Shared/NormalMap.inc"
#include "Shared/TextureCompression.inc"

#include "UndefineInOut.inc"

#if defined(TEXTURE_ATLAS)
    void SampleTerrainTextures(out vec4 texelOut, out vec2 specularLevel_ExponentOut, inout vec3 normal, highp vec3 worldPos, highp vec3 viewVector, vec4 vertexAlbedo, vec4 materialProperties)
    {
        TextureSettings textureSettings1; getTextureSettings1D(vMaterialSettingsSlots1D[0], textureSettings1);
        TextureSettings textureSettings2; getTextureSettings1D(vMaterialSettingsSlots1D[1], textureSettings2);
        TextureSettings textureSettings3; getTextureSettings1D(vMaterialSettingsSlots1D[2], textureSettings3);

        // Compute UVs from world space position
        highp vec3 textureScale = vTextureScale.xyz;
        highp vec3 invTextureScale = 8.0 / (textureScale * uGridSize);
        highp vec2 textureUV1 = worldPos.xz * invTextureScale[0];
        highp vec2 textureUV2 = worldPos.xz * invTextureScale[1];
        highp vec2 textureUV3 = worldPos.xz * invTextureScale[2];

        // Early out special code path for texel density debug
    #if defined(DEBUG_TEXEL_DENSITY)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);
        gl_FragColor.xyz += GetTexelDensityDebugColour(textureUV1, textureSettings1.textureMeta1.z, viewVector) * vTextureWeight.x;
        gl_FragColor.xyz += GetTexelDensityDebugColour(textureUV2, textureSettings2.textureMeta1.z, viewVector) * vTextureWeight.y;
        gl_FragColor.xyz += GetTexelDensityDebugColour(textureUV3, textureSettings3.textureMeta1.z, viewVector) * vTextureWeight.z;
    #endif

        // Mod with 1.0 early while in highp
        textureUV1 = mod(textureUV1, 1.0);
        textureUV2 = mod(textureUV2, 1.0);
        textureUV3 = mod(textureUV3, 1.0);

        // Take view-relative position derivatives and calculate UV derivatives for each texture from it
        highp vec3 positionDX = vec3(0.0);
        highp vec3 positionDY = vec3(0.0);
    #if defined(STANDARD_DERIVATIVES) && (defined(TEXTURE_LOD) || defined(NORMAL_MAP))
        positionDX = dFdx(viewVector);
        positionDY = dFdy(viewVector);
    #endif
        vec2 uvDX1 = positionDX.xz * invTextureScale[0];
        vec2 uvDY1 = positionDY.xz * invTextureScale[0];
        vec2 uvDX2 = positionDX.xz * invTextureScale[1];
        vec2 uvDY2 = positionDY.xz * invTextureScale[1];
        vec2 uvDX3 = positionDX.xz * invTextureScale[2];
        vec2 uvDY3 = positionDY.xz * invTextureScale[2];

        vec3 materialBrightness = mod(materialProperties.xyz, vec3(16.0)) - vec3(0.25); // Each material brightness is stored in 0->15 (lowest 4 bits).  Make sure to negate epsilon from vertex shader
        vec3 materialWhiteness = ((materialProperties.xyz - vec3(0.25)) - materialBrightness) / vec3(255.0); // Each material whiteness is stored in 16->255 (highest 4 bits).  Make sure to negate epsilon from vertex shader
        materialBrightness *= vec3(1.0 / 15.0); // Put into 0-1 range

        // Refer to ColourUtils::ApplyOffsetToFaceColour.  AmbientAdjust (which for terrain is always 74) is multipied by 2 to get the maximum 'white' value
        vec4 white = vec4(vec3((74.0 * 2.0) / 255.0), vertexAlbedo.a);
        vec4 vertexAlbedo1 = mix(vertexAlbedo, white, materialWhiteness.x) * (1.0 + materialBrightness.x);
        vec4 vertexAlbedo2 = mix(vertexAlbedo, white, materialWhiteness.y) * (1.0 + materialBrightness.y);
        vec4 vertexAlbedo3 = mix(vertexAlbedo, white, materialWhiteness.z) * (1.0 + materialBrightness.z);

    #if defined(GAMMA_CORRECT_INPUTS)
        vertexAlbedo1.rgb = SRGBToLinear(vertexAlbedo1.rgb);
        vertexAlbedo2.rgb = SRGBToLinear(vertexAlbedo2.rgb);
        vertexAlbedo3.rgb = SRGBToLinear(vertexAlbedo3.rgb);
    #endif

        vec4 texelAlbedo1, texelAlbedo2, texelAlbedo3;
        vec4 texelEmissiveNormal1, texelEmissiveNormal2, texelEmissiveNormal3;
        #if defined(COMPRESSED_NORMALS)
            // 0 Emissive, no-op normal (Y, Z (Unused), X)
            texelEmissiveNormal1 = vec4(0.0, 0.5, 0.0, 0.5);
            texelEmissiveNormal2 = vec4(0.0, 0.5, 0.0, 0.5);
            texelEmissiveNormal3 = vec4(0.0, 0.5, 0.0, 0.5);
        #else
            // 0 Emissive, no-op normal (Y, Z, X)
            texelEmissiveNormal1 = vec4(0.0, 0.5, 1.0, 0.5);
            texelEmissiveNormal2 = vec4(0.0, 0.5, 1.0, 0.5);
            texelEmissiveNormal3 = vec4(0.0, 0.5, 1.0, 0.5);
        #endif
        vec4 texelMetalnessRoughness1, texelMetalnessRoughness2, texelMetalnessRoughness3;
        getTexel(textureUV1, textureSettings1.textureMeta1, textureSettings1.textureMeta2, textureSettings1.textureMeta3, uAtlasMeta, uvDX1, uvDY1, 0.0, uTextureAtlas, texelAlbedo1, texelEmissiveNormal1, texelMetalnessRoughness1);
        getTexel(textureUV2, textureSettings2.textureMeta1, textureSettings2.textureMeta2, textureSettings2.textureMeta3, uAtlasMeta, uvDX2, uvDY2, 0.0, uTextureAtlas, texelAlbedo2, texelEmissiveNormal2, texelMetalnessRoughness2);
        getTexel(textureUV3, textureSettings3.textureMeta1, textureSettings3.textureMeta2, textureSettings3.textureMeta3, uAtlasMeta, uvDX3, uvDY3, 0.0, uTextureAtlas, texelAlbedo3, texelEmissiveNormal3, texelMetalnessRoughness3);
        texelAlbedo1 *= vertexAlbedo1;
        texelAlbedo2 *= vertexAlbedo2;
        texelAlbedo3 *= vertexAlbedo3;
    #if defined(ETC_CHANNEL_SWIZZLE)
        // Swizzle ETC channels back into the expected order
        //NX,NY,NZ,HD -> HD,NY,NZ,NX
        texelEmissiveNormal1 = texelEmissiveNormal1.agbr;
        texelEmissiveNormal2 = texelEmissiveNormal2.agbr;
        texelEmissiveNormal3 = texelEmissiveNormal3.agbr;
    #endif
        //TODO: Emissive calc
    #if defined(NORMAL_MAP)
        vec3 normalMap1 = UnpackNormal(texelEmissiveNormal1, textureSettings1.normalScale);
        vec3 normalMap2 = UnpackNormal(texelEmissiveNormal2, textureSettings2.normalScale);
        vec3 normalMap3 = UnpackNormal(texelEmissiveNormal3, textureSettings3.normalScale);
        vec3 normalMap = normalMap1 * vTextureWeight.x + normalMap2 * vTextureWeight.y + normalMap3 * vTextureWeight.z;
        #if defined(PER_FRAGMENT_TANGENTS)
            normal = ApplyNormalMapTerrain(normalMap, normal, positionDX, positionDY);
        #else
            normal = ApplyNormalMapTerrain(normalMap, normal);
        #endif
    #endif

        // Minimap terrain will use the vertex colour where the material has no whiteness.  Where it DOES have whiteness, it will bring in some texture (and sqrt to put in gamma space)
    #if defined(TEXTURE_FORCED_FLAG)
        texelAlbedo1 = mix(vertexAlbedo1, sqrt(texelAlbedo1), materialWhiteness.x);
        texelAlbedo2 = mix(vertexAlbedo2, sqrt(texelAlbedo2), materialWhiteness.y);
        texelAlbedo3 = mix(vertexAlbedo3, sqrt(texelAlbedo3), materialWhiteness.z);
    #endif
        texelOut = texelAlbedo1 * vTextureWeight.x + texelAlbedo2 * vTextureWeight.y + texelAlbedo3 * vTextureWeight.z;
        //Ensure that texel alpha is now opaque, as terrain is rendered as opaque
        //Any leftover transparency appears to cause issues in ANGLE builds
        texelOut.a = 1.0;

        vec3 materialSpecularFlags = step(vec3(0.5), fract(materialProperties.w * vec3(0.25, 0.125, 0.0625))); //(1-4 bits from LSB)

        vec3 weightedTexelAlphas = vec3(texelAlbedo1.a, texelAlbedo2.a, texelAlbedo3.a) * vTextureWeight.xyz;
        vec3 weightedSpecularLevels = weightedTexelAlphas * materialSpecularFlags;
        //Average the specular level
        specularLevel_ExponentOut.x = weightedSpecularLevels.x + weightedSpecularLevels.y + weightedSpecularLevels.z;
        //Select the maximum specular exponent
        specularLevel_ExponentOut.y = max(textureSettings1.specular, max(textureSettings2.specular, textureSettings3.specular));

        //Directly set the frag colour with debug values, if requested
    #if defined(DEBUG_RT7_EMISSIVE)
        gl_FragColor = vec4((texelEmissiveNormal1 * vTextureWeight.x + texelEmissiveNormal2 * vTextureWeight.y + texelEmissiveNormal3 * vTextureWeight.z).rrr, 1.0);
    #elif defined(DEBUG_RT7_METALNESS)
        gl_FragColor = vec4((texelMetalnessRoughness1 * vTextureWeight.x + texelMetalnessRoughness2 * vTextureWeight.y + texelMetalnessRoughness3 * vTextureWeight.z).rrr, 1.0);
    #elif defined(DEBUG_RT7_ROUGHNESS)
        gl_FragColor = vec4((texelMetalnessRoughness1 * vTextureWeight.x + texelMetalnessRoughness2 * vTextureWeight.y + texelMetalnessRoughness3 * vTextureWeight.z).ggg, 1.0);
    #endif
    }
#endif

#include "DefineInOut.inc"
