#include "ConversionUtils.inc"

// Gets the mipmap level we would be using if we sampled at this coordinate (scaled to pixels)
float getMipMapLevel(vec2 dx, vec2 dy)
{
    float level = 0.0;
#if defined(STANDARD_DERIVATIVES)
    float deltaMaxSqr = max(dot(dx, dx), dot(dy, dy));
    level = 0.5 * log2(deltaMaxSqr);
    level = max(0.0, level);
#endif
    return level;
}

#if defined(DEBUG_TEXEL_DENSITY)
    vec3 GetTexelDensityDebugColour(vec2 uv, float textureSize, vec3 viewRelativePosition)
    {
        // Get maximum absolute change in texels per surrounding screen pixel
        float pixelRatio = length(fwidth(uv) * textureSize);
        // Get maximum absolute change in world space position per surrounding screen pixel
        float worldRatio = length(fwidth(viewRelativePosition));
        // Get ratio of world space units to texels
        float density = pixelRatio / worldRatio;

        float densityMax = uDebugTexelDensity.x;
        float densityTarget = uDebugTexelDensity.y;
        float densityMin = uDebugTexelDensity.z;
        float densityRadius = uDebugTexelDensity.w;

        // Fully red at Max, through to green at Target with some overlap both sides (Radius), through to fully blue at Min
        vec3 result;
        result.r = smoothstep(densityTarget / (densityRadius + 1.0), densityMax, density);
        result.g = 1.0 - smoothstep(0.0, densityTarget * (densityRadius + 1.0), abs(density - densityTarget));
        result.b = smoothstep(1.0 - (densityTarget + densityMin * densityRadius), 1.0 - densityMin, 1.0 - density);

        // Make result more visible after postprocessing
        result *= result;

        return result;
    }
#endif

#if defined(LOOKUP_MODE_DYNAMIC) && !defined(NO_SAMPLER_WRAP)
    // Texture wrap mode.
    flat in mediump float vSamplerWrap;
#endif

// NOTE: Undefine inout after the sampler wrap declaration above
#include "UndefineInOut.inc"

// Core inner method that handles the actual sampling - do not call directly
// NOTE: textureUV must be a highp input here to ensure highp calculations on low-end devices
void getTexelBias_inner(float textureCount, highp vec2 textureUV, highp vec3 textureMeta1, highp vec3 textureMeta2, highp vec3 textureMeta3, vec4 atlasMeta, vec2 uvDX, vec2 uvDY, float samplerWrapPacked, sampler2D atlas, out vec4 texel1, out vec4 texel2, out vec4 texel3)
{
    // Grab some useful values from the provided parameters
    float atlasSlotSize = atlasMeta.x;
    highp float atlasSlotPixelSize = atlasMeta.y;
    highp float atlasInversePixelSize = atlasMeta.z;
    float atlasMaxMipLevel = atlasMeta.w;

    highp vec3 atlasXSlots = vec3(textureMeta1.x, textureMeta2.x, textureMeta3.x);
    highp vec3 atlasYSlots = vec3(textureMeta1.y, textureMeta2.y, textureMeta3.y);
    highp vec3 texturePixelSizes = vec3(textureMeta1.z, textureMeta2.z, textureMeta3.z);
    const vec2 one = vec2(1.0);

    // Handle the UV wrapping within the atlas
    vec3 halfPixelSizes = 0.5 / texturePixelSizes;
    highp vec2 uv1, uv2, uv3;

#if !defined(LOOKUP_MODE_CLAMP) && !defined(LOOKUP_MODE_REPEAT)
    const float t1 = 1.0 / 2.0;
    const float t2 = 1.0 / 4.0;
    const float t3 = 1.0 / 8.0;
    const float t4 = 1.0 / 16.0;
    vec4 samplerWrap = step(0.5, fract(samplerWrapPacked * vec4(t1, t2, t3, t4)));
#endif

#if defined(LOOKUP_MODE_CLAMP)
    uv1 = clamp(textureUV, vec2(halfPixelSizes.x), one - vec2(halfPixelSizes.x));
#elif defined(LOOKUP_MODE_REPEAT)
    // Wrap function
    uv1 = mod(textureUV, one);
#else //DYNAMIC
    const vec2 cHalf = vec2(0.5);
    const vec2 two = vec2(2.0);

    vec2 uvClamp1 = clamp(textureUV, vec2(halfPixelSizes.x), one - vec2(halfPixelSizes.x));
    vec2 uvWrap1 = mod(textureUV, one);
    vec2 tmp = textureUV - two * floor(cHalf * textureUV);
    vec2 uvMirror1 = one - abs(one - tmp);

    uv1 = samplerWrap.xy * uvClamp1 + samplerWrap.zw * uvWrap1 + (one - samplerWrap.xy - samplerWrap.zw) * uvMirror1;
#endif
    uv1 = uv1 * texturePixelSizes.x * atlasInversePixelSize;
    uv1 += vec2(atlasXSlots.x, atlasYSlots.x) * atlasSlotPixelSize * atlasInversePixelSize;

    if (textureCount > 1.0)
    {
    #if defined(LOOKUP_MODE_CLAMP)
        uv2 = clamp(textureUV, vec2(halfPixelSizes.y), one - vec2(halfPixelSizes.y));
    #elif defined(LOOKUP_MODE_REPEAT)
        // Wrap function
        uv2 = mod(textureUV, one);
    #else //DYNAMIC
        vec2 uvClamp2 = clamp(textureUV, vec2(halfPixelSizes.y), one - vec2(halfPixelSizes.y));
        vec2 uvWrap2 = uvWrap1;
        vec2 uvMirror2 = uvMirror1;

        uv2 = samplerWrap.xy * uvClamp2 + samplerWrap.zw * uvWrap2 + (one - samplerWrap.xy - samplerWrap.zw) * uvMirror2;
    #endif
        uv2 = uv2 * texturePixelSizes.y * atlasInversePixelSize;
        uv2 += vec2(atlasXSlots.y, atlasYSlots.y) * atlasSlotPixelSize * atlasInversePixelSize;

        if (textureCount > 2.0)
        {
        #if defined(LOOKUP_MODE_CLAMP)
            uv3 = clamp(textureUV, vec2(halfPixelSizes.z), one - vec2(halfPixelSizes.z));
        #elif defined(LOOKUP_MODE_REPEAT)
            //Wrap function
            uv3 = mod(textureUV, one);
        #else //DYNAMIC
            vec2 uvClamp3 = clamp(textureUV, vec2(halfPixelSizes.z), one - vec2(halfPixelSizes.z));
            vec2 uvWrap3 = uvWrap1;
            vec2 uvMirror3 = uvMirror1;

            uv3 = samplerWrap.xy * uvClamp3 + samplerWrap.zw * uvWrap3 + (one - samplerWrap.xy - samplerWrap.zw) * uvMirror3;
        #endif
            uv3 = uv3 * texturePixelSizes.z * atlasInversePixelSize;
            uv3 += vec2(atlasXSlots.z, atlasYSlots.z) * atlasSlotPixelSize * atlasInversePixelSize;
        }
    }


    // Sample texture(s)
    // Initialise secondary textures to 0 to avoid uninitialised errors in ANGLE
    texel2 = vec4(0.0);
    texel3 = vec4(0.0);
    // All platforms other than ES 2.0 have TEXTURE_MIP_LIMIT, which we use to enforce a partial mip chain
#if defined(TEXTURE_MIP_LIMIT)
    // If we have texture2DGrad, then we can use the hardware mipmap pipeline with custom derivatives to pick exactly the right mip level as if we were not using atlasing
    #if defined(TEXTURE_GRAD)
        // Scale original input UV derivatives from texture space into atlas space, so that they match the atlas UVs
        highp vec2 uvDXScaled = uvDX * atlasInversePixelSize;
        highp vec2 uvDYScaled = uvDY * atlasInversePixelSize;
        highp vec2 dx = uvDXScaled * texturePixelSizes.x;
        highp vec2 dy = uvDYScaled * texturePixelSizes.x;
        // Limit derivatives to avoid anisotropic filtering artifacts where the GPU tries to sample past GL_TEXTURE_MAX_LEVEL
        const vec2 maxD = vec2(0.025);
        dx = clamp(dx, -maxD, maxD);
        dy = clamp(dy, -maxD, maxD);
        texel1 = texture2DGrad(atlas, uv1, dx, dy);
        if (textureCount > 1.0)
        {
            dx = uvDXScaled * texturePixelSizes.y;
            dy = uvDYScaled * texturePixelSizes.y;
            dx = clamp(dx, -maxD, maxD);
            dy = clamp(dy, -maxD, maxD);
            texel2 = texture2DGrad(atlas, uv2, dx, dy);

            if (textureCount > 2.0)
            {
                dx = uvDXScaled * texturePixelSizes.z;
                dy = uvDYScaled * texturePixelSizes.z;
                dx = clamp(dx, -maxD, maxD);
                dy = clamp(dy, -maxD, maxD);
                texel3 = texture2DGrad(atlas, uv3, dx, dy);
            }
        }
    // Otherwise just sample normally. Less mips are uploaded in this case to prevent grid artifacts.
    // However we can only do this if TEXTURE_MIP_LIMIT is supported, which is not the case on ES 2.0 (just outputs black).
    #else
        texel1 = texture2D(atlas, uv1);
        if (textureCount > 1.0)
        {
            texel2 = texture2D(atlas, uv2);

            if (textureCount > 2.0)
            {
                texel3 = texture2D(atlas, uv3);
            }
        }
    #endif
// The remaining cases are for ES 2.0, where we try to emulate GL_TEXTURE_MAX_LEVEL.
#else
    // If we have texture2DLod, then we can calculate and force a suitable mip level. The vast majority of ES 2.0 platforms will follow this path.
    #if defined(TEXTURE_LOD)
        vec2 dx = uvDX * texturePixelSizes.x;
        vec2 dy = uvDY * texturePixelSizes.x;
        // Note if we don't have STANDARD_DERIVATIVES, this will force LOD 0
        float lodLevel = getMipMapLevel(dx, dy);
        // Clamp to smallest available mip level
        lodLevel = min(lodLevel, atlasMaxMipLevel);
        texel1 = texture2DLod(atlas, uv1, lodLevel);

        if (textureCount > 1.0)
        {
            dx = uvDX * texturePixelSizes.y;
            dy = uvDY * texturePixelSizes.y;
            // Note if we don't have STANDARD_DERIVATIVES, this will force LOD 0
            lodLevel = getMipMapLevel(dx, dy);
            // Clamp to smallest available mip level
            lodLevel = min(lodLevel, atlasMaxMipLevel);
            texel2 = texture2DLod(atlas, uv2, lodLevel);

            if (textureCount > 2.0)
            {
                dx = uvDX * texturePixelSizes.z;
                dy = uvDY * texturePixelSizes.z;
                // Note if we don't have STANDARD_DERIVATIVES, this will force LOD 0
                lodLevel = getMipMapLevel(dx, dy);
                // Clamp to smallest available mip level
                lodLevel = min(lodLevel, atlasMaxMipLevel);
                texel3 = texture2DLod(atlas, uv3, lodLevel);
            }
        }
    // And as a last resort, we just force the top mip level (only DX9 SM2 hardware will hit this path)
    #else
        // (Assumes C++ side has only uploaded 1 mip level)
        texel1 = texture2D(atlas, uv1);
        if (textureCount > 1.0)
        {
            texel2 = texture2D(atlas, uv2);

            if (textureCount > 2.0)
            {
                texel3 = texture2D(atlas, uv3);
            }
        }
    #endif
#endif
}

// Additional pass-through methods that do additional UV/texel handling

// Use these functions if you want to let the default bias scale handle everything
// You usually will use these for everything but normal map lookups
// This converts the sampled texel from sRGB space to linear space
// This also manages the non-diffuse texels needing to be in linear space
void getTexel_inner(float textureCount, vec2 textureUV, highp vec3 textureMeta1, highp vec3 textureMeta2, highp vec3 textureMeta3, vec4 atlasMeta, vec2 uvDX, vec2 uvDY, float samplerWrapPacked, sampler2D atlas, out vec4 texel1, out vec4 texel2, out vec4 texel3)
{
    getTexelBias_inner(textureCount, textureUV, textureMeta1, textureMeta2, textureMeta3, atlasMeta, uvDX, uvDY, samplerWrapPacked, atlas, texel1, texel2, texel3);
#if defined(SRGB_TEXTURES)
    if (textureCount > 1.0)
    {
        texel2 = vec4(LinearToSRGB(texel2.rgb), texel2.a);
    }
    if (textureCount > 2.0)
    {
        texel3 = vec4(LinearToSRGB(texel3.rgb), texel3.a);
    }
#else
    texel1 = vec4(SRGBToLinear(texel1.rgb), texel1.a);
#endif
}
void getTexel_inner(float textureCount, vec2 textureUV, highp vec3 textureMeta1, highp vec3 textureMeta2, highp vec3 textureMeta3, vec4 atlasMeta, float samplerWrapPacked, sampler2D atlas, out vec4 texel1, out vec4 texel2, out vec4 texel3)
{
    vec2 uvDX = vec2(0.0);
    vec2 uvDY = vec2(0.0);
#if defined(STANDARD_DERIVATIVES)
    uvDX = dFdx(textureUV);
    uvDY = dFdy(textureUV);
#endif
    getTexel_inner(textureCount, textureUV, textureMeta1, textureMeta2, textureMeta3, atlasMeta, uvDX, uvDY, samplerWrapPacked, atlas, texel1, texel2, texel3);
}

//Include the lookup variants that are called externally - 3 lookup variants (1 texture, 2 textures, 3 textures)
#include "Shared/TextureAtlasLookupMethods.inc"

#include "DefineInOut.inc"
