#ifndef TEXTURE_SETTINGS_INC
#define TEXTURE_SETTINGS_INC

//[
// RT5 material config settings
// These are baked into a texture at runtime and sampled for a particular surface using getTextureSettings()
struct TextureSettings
{
    // x: slot X , y: slot Y, z: pixel size
    highp vec3 textureMeta1;
    // x: slot X , y: slot Y, z: pixel size
    highp vec3 textureMeta2;
    // Texture UV animation deltas
    highp vec2 uvAnim;
    // Texture wrapping address mode
    float wrapping;
    // Specular power/roughness
    float specular;
    // Amount to scale normal mapped normals, 0-1
    float normalScale;
#if defined(REFRACTION)
    // x: Refraction normal scale, y: index of refraction, z: refraction vector world space depth bias
    vec4 refraction;
#endif
#if defined(VIEWPORTMAP)
    // Viewport Map UV scale (xy) and animation (zw)
    vec4 viewportMapUVScaleAndAnim;
#endif
#if defined(DEBUG_MATERIAL_HIGHLIGHT)
    highp float materialID;
#endif
};
//]

#include "UndefineInOut.inc"

void getTextureSettings(vec2 materialSettingsSlotXY, out TextureSettings outSettings)
{
    const highp float inv255 = 1.0 / 255.0;
    const highp float inv65535 = 1.0 / 65535.0;
    const highp float val32767 = 32767.0;
    const highp float inv32767 = 1.0 / 32767.0;
    const float invSettingsTotalPixelSize = 1.0 / MATERIAL_SETTINGS_TEXTURE_RESOLUTION;

    vec2 settingsUV = ((floor(materialSettingsSlotXY + 0.5) * vec2(MATERIAL_SETTINGS_SLOT_PIXEL_RESOLUTION_X, MATERIAL_SETTINGS_SLOT_PIXEL_RESOLUTION_Y)) + 0.5) * invSettingsTotalPixelSize;

    const float sampleOffset = invSettingsTotalPixelSize;

    // Read raw pixel values
    // Pixel layout is:
    // 012 (0 = slotAB, 1 = slotSizesAndSlotEtcAndWrapping, 2 = uvAnim)
    // 345 (3 = specularNormalScale, 4 = refraction, 5 = viewportMapUVScale)
    // 678 (6 = viewportMapUVAnimation)
    // 9AB

    // The first sample doesn't need an offset
    vec4 slotAB = texture2DLodCompat(uTextureAtlasSettings, settingsUV, 0.0);
    // The rest do
    vec4 slotSizesAndSlotEtcAndWrapping, uvAnim, specularNormalScale, refraction, viewportMapUVScale, viewportMapUVAnimation;
    vec4 debugFlags;
    // Branch based on the texture sample method available
#if defined(TEXTURE_SETTINGS_USE_TEXEL_OFFSETS)

    #define SAMPLE_OFFSET_SLOTSIZES_AND_SLOTETC_AND_WRAPPING    ivec2(1, 0)
    #define SAMPLE_OFFSET_UV_ANIM                                ivec2(2, 0)
    #define SAMPLE_OFFSET_SPECULAR_NORMAL_SCALE                    ivec2(0, 1)
    #define SAMPLE_OFFSET_REFRACTION                            ivec2(1, 1)
    #define SAMPLE_OFFSET_VIEWPORTMAP_UVSCALE                    ivec2(2, 1)
    #define SAMPLE_OFFSET_VIEWPORTMAP_UVANIMATION                ivec2(0, 2)
    #define SAMPLE_OFFSET_DEBUG                                    ivec2(2, 3)

    slotSizesAndSlotEtcAndWrapping = textureLodOffset(uTextureAtlasSettings, settingsUV, 0.0, SAMPLE_OFFSET_SLOTSIZES_AND_SLOTETC_AND_WRAPPING);
    uvAnim = textureLodOffset(uTextureAtlasSettings, settingsUV, 0.0, SAMPLE_OFFSET_UV_ANIM);
    specularNormalScale = textureLodOffset(uTextureAtlasSettings, settingsUV, 0.0, SAMPLE_OFFSET_SPECULAR_NORMAL_SCALE);
    #if defined(REFRACTION)
        refraction = textureLodOffset(uTextureAtlasSettings, settingsUV, 0.0, SAMPLE_OFFSET_REFRACTION);
    #endif
    #if defined(VIEWPORTMAP)
        viewportMapUVScale = textureLodOffset(uTextureAtlasSettings, settingsUV, 0.0, SAMPLE_OFFSET_VIEWPORTMAP_UVSCALE);
        viewportMapUVAnimation = textureLodOffset(uTextureAtlasSettings, settingsUV, 0.0, SAMPLE_OFFSET_VIEWPORTMAP_UVANIMATION);
    #endif
    #if defined(DEBUG_MATERIAL_HIGHLIGHT)
        debugFlags = textureLodOffset(uTextureAtlasSettings, settingsUV, 0.0, SAMPLE_OFFSET_DEBUG);
    #endif
#else

    vec2 SAMPLE_OFFSET_SLOTSIZES_AND_SLOTETC_AND_WRAPPING = vec2(sampleOffset, 0);
    vec2 SAMPLE_OFFSET_UV_ANIM = vec2(sampleOffset * 2.0, 0);
    vec2 SAMPLE_OFFSET_SPECULAR_NORMAL_SCALE = vec2(0, sampleOffset);
    vec2 SAMPLE_OFFSET_REFRACTION = vec2(sampleOffset, sampleOffset);
    vec2 SAMPLE_OFFSET_VIEWPORTMAP_UVSCALE = vec2(sampleOffset * 2.0, sampleOffset);
    vec2 SAMPLE_OFFSET_VIEWPORTMAP_UVANIMATION = vec2(0, sampleOffset * 2.0);
    vec2 SAMPLE_OFFSET_DEBUG = vec2(sampleOffset * 3.0, sampleOffset * 3.0);
    //TODO: Fix these defines to not use sampleOffset because the obfuscator doesn't also update this line and leads to compilation errors
    //#define SAMPLE_OFFSET_SLOTSIZES_AND_SLOTETC_AND_WRAPPING    vec2(sampleOffset, 0)
    //#define SAMPLE_OFFSET_UV_ANIM                                vec2(sampleOffset * 2.0, 0)
    //#define SAMPLE_OFFSET_SPECULAR_NORMAL_SCALE                vec2(0, sampleOffset)
    //#define SAMPLE_OFFSET_REFRACTION                            vec2(sampleOffset, sampleOffset)
    //#define SAMPLE_OFFSET_VIEWPORTMAP_UVSCALE                    vec2(sampleOffset * 2.0, sampleOffset)
    //#define SAMPLE_OFFSET_VIEWPORTMAP_UVANIMATION                vec2(0, sampleOffset * 2.0)
    //#define SAMPLE_OFFSET_DEBUG                                vec2(sampleOffset * 2.0, sampleOffset * 3.0)

    slotSizesAndSlotEtcAndWrapping = texture2DLodCompat(uTextureAtlasSettings, settingsUV + SAMPLE_OFFSET_SLOTSIZES_AND_SLOTETC_AND_WRAPPING, 0.0);
    uvAnim = texture2DLodCompat(uTextureAtlasSettings, settingsUV + SAMPLE_OFFSET_UV_ANIM, 0.0);
    specularNormalScale = texture2DLodCompat(uTextureAtlasSettings, settingsUV + SAMPLE_OFFSET_SPECULAR_NORMAL_SCALE, 0.0);
    #if defined(REFRACTION)
        refraction = texture2DLodCompat(uTextureAtlasSettings, settingsUV + SAMPLE_OFFSET_REFRACTION, 0.0);
    #endif
    #if defined(VIEWPORTMAP)
        viewportMapUVScale = texture2DLodCompat(uTextureAtlasSettings, settingsUV + SAMPLE_OFFSET_VIEWPORTMAP_UVSCALE, 0.0);
        viewportMapUVAnimation = texture2DLodCompat(uTextureAtlasSettings, settingsUV + SAMPLE_OFFSET_VIEWPORTMAP_UVANIMATION, 0.0);
    #endif
    #if defined(DEBUG_MATERIAL_HIGHLIGHT)
        debugFlags = texture2DLodCompat(uTextureAtlasSettings, settingsUV + SAMPLE_OFFSET_DEBUG, 0.0);
    #endif
#endif

    // Scale relevant pixel values up to integers
    slotAB = floor(slotAB * 255.0 + 0.5);
    slotSizesAndSlotEtcAndWrapping = floor(slotSizesAndSlotEtcAndWrapping * 255.0 + 0.5);

    // Process slot etc values first
    const float t1 = 1.0 / 2.0;
    const float t2 = 1.0 / 4.0;
    const float t3 = 1.0 / 8.0;
    const float t4 = 1.0 / 16.0;
    vec4 slotEtc = step(0.5, fract(slotSizesAndSlotEtcAndWrapping.z * vec4(t1, t2, t3, t4)));
    slotAB += vec4(256.0) * slotEtc;

    vec2 pixelSizes = slotSizesAndSlotEtcAndWrapping.xy * uAtlasMeta.y;
    outSettings.textureMeta1 = vec3(slotAB.xy, pixelSizes.x);
    outSettings.textureMeta2 = vec3(slotAB.zw, pixelSizes.y);
    outSettings.wrapping = slotSizesAndSlotEtcAndWrapping.w;
    outSettings.specular = UnpackVec2ToFloat(specularNormalScale.xy) * inv255;
    outSettings.normalScale = UnpackVec2ToFloat(specularNormalScale.zw) * inv255;

    // UV anim range is -2.0 to 2.0
    highp vec2 uvAnimUnpacked = vec2(UnpackVec2ToFloat(uvAnim.xy), UnpackVec2ToFloat(uvAnim.zw)) - val32767;
    // 0 can be either 127 or 128 in the texture (unpacks to 0 or 1) so we handle this with a deadzone
    uvAnimUnpacked *= step(1.5, abs(uvAnimUnpacked));
    outSettings.uvAnim = uvAnimUnpacked * inv32767 * 2.0;

    // Put normal scale back into the -8 to 8 range
    outSettings.normalScale = (outSettings.normalScale * 0.1) - 8.0;

#if defined(REFRACTION)
    outSettings.refraction = refraction;
    // Put refraction data back into the correct ranges
    outSettings.refraction.g = (outSettings.refraction.g * 2.0) + 1.0;
    outSettings.refraction.b = UnpackVec2ToFloat(outSettings.refraction.zw) * inv65535 * 10.0;
#endif

#if defined(VIEWPORTMAP)
    // Scale and animation ranges are -2.0 to 2.0
    highp vec2 viewportMapUVScaleUnpacked = vec2(UnpackVec2ToFloat(viewportMapUVScale.xy), UnpackVec2ToFloat(viewportMapUVScale.zw)) - val32767;
    highp vec2 viewportMapUVAnimationUnpacked = vec2(UnpackVec2ToFloat(viewportMapUVAnimation.xy), UnpackVec2ToFloat(viewportMapUVAnimation.zw)) - val32767;
    // 0 can be either 127 or 128 in the texture (unpacks to 0 or 1) so we handle this with a deadzone
    viewportMapUVScaleUnpacked *= step(1.5, abs(viewportMapUVScaleUnpacked));
    viewportMapUVAnimationUnpacked *= step(1.5, abs(viewportMapUVAnimationUnpacked));
    outSettings.viewportMapUVScaleAndAnim = vec4(viewportMapUVScaleUnpacked * inv32767 * 2.0, viewportMapUVAnimationUnpacked * inv32767 * 2.0);
#endif

#if defined(DEBUG_MATERIAL_HIGHLIGHT)
    outSettings.materialID = UnpackVec2ToFloat(debugFlags.xy);
#endif
}

void getTextureSettings1D(float materialSettingsSlot1D, out TextureSettings outSettings)
{
    const float invSettingsSlotCountX = 1.0 / MATERIAL_SETTINGS_SLOTS_DIMENSION_COUNT_X;

    float y = floor((materialSettingsSlot1D + 0.5) * invSettingsSlotCountX);
    float x = materialSettingsSlot1D - (y * MATERIAL_SETTINGS_SLOTS_DIMENSION_COUNT_X);

    getTextureSettings(vec2(x, y), outSettings);
}
#include "DefineInOut.inc"

#endif
