#ifndef VOLUMETRIC_FUNCTIONS_INC
#define VOLUMETRIC_FUNCTIONS_INC

#if defined(SUNLIGHT_SHADOWS) && defined(VOLUMETRIC_SCATTERING) && defined(SUNLIGHT_DIRECT_LIGHTING)

#define VOLUMETRIC_SCATTERING_SUPPORTED

// These are defined in code
//#define VOLUMETRIC_SCATTERING_NOISE 1
//#define VOLUMETRIC_GROUND_FOG 1

// x = 1 - g^2,  y = 1 + g^2,  z = 2 * g,  w = 1/4PI
uniform vec4 uMieG;

// x = Scattering Coefficient, y = Extinction Coefficient, z = Skybox Extinction, w = Exaggeration Factor
uniform vec4 uVolumetricScatteringParameters;

// x = Height where density is 1 below it, y = Falloff over height
#if defined(VOLUMETRIC_GROUND_FOG)
uniform vec4 uGroundFogHeight_Falloff;
#endif

#if defined(VOLUMETRIC_SCATTERING_NOISE)
    uniform sampler3D sNoiseTex;
    // xyz = Noise frequency, w = Noise strength
    uniform vec4 u3DNoiseFrequency_Strength;
    // xyz = Noise movement (Wind), w = Noise power (Variance)
    uniform vec4 u3DNoiseWind_Power;
#endif

uniform float uTime;
float ShadowSample(vec4 worldPos, vec4 lightViewPos, float pixelViewDepth)
{
    int cascadeCount = int(uMappingParams.w);
    vec4 shadowMapTexCoords;
    int iCascade;
#if defined(CASCADE_SPLIT_SELECTION)
    iCascade = ShadowMapSelectCascadeBySplit(pixelViewDepth, uCascadeFrustumViewDepths, uCascadeSplitSelectionFlags);
    #if defined(USE_LIGHT_VIEW_PROJ_TEX_MATRIX)
        shadowMapTexCoords = uSunlightViewProjTexMatrix[iCascade] * worldPos;
    #else
        shadowMapTexCoords = lightViewPos * uSunlightProjTexMatScale[iCascade] + uSunlightProjTexMatOffset[iCascade];
    #endif
#else
    #if defined(USE_LIGHT_VIEW_PROJ_TEX_MATRIX)
        iCascade = ShadowMapSelectCascadeByMap(shadowMapTexCoords, worldPos, uSunlightViewProjTexMatrix, uCascadeMinAtlasExtents);
    #else
        iCascade = ShadowMapSelectCascadeByMap(shadowMapTexCoords, lightViewPos, uSunlightProjTexMatScale, uSunlightProjTexMatOffset, uCascadeMinAtlasExtents);
    #endif
#endif
    // The map selection code can give cascade values outside of the range (e.g. for far away pixels), and we don't want to sample there
    return (iCascade >= cascadeCount) ? 1.0 : ShadowDepthMapFilter1x1(uSunlightShadowMap, shadowMapTexCoords);
}

float PhaseFunction(float cosAngle, vec4 g)
{
    return g.w * (g.x / pow(g.y - g.z * cosAngle, 1.5)); // This is the Henyey-Greenstein phase function
}

/**
    @param offset Initial ray offset into the scene from the ray origin, values of 0.0 are not valid a may result in unexpected behaviour
 */
vec4 GetScatteredInRay(int steps, vec3 viewDirection, float cameraDistance, float offset, vec4 mieG)
{
    float distanceInsideShadowMap = uSunlightFadeAttenParams.y * 1.4; // 1.4 is approximately sqrt of 2
    float sampledRayLength = min(distanceInsideShadowMap, cameraDistance);
    vec3 rayStart = uCameraPosition;
    vec3 rayEnd = rayStart + (viewDirection * sampledRayLength);
    vec4 sunlightViewPosStart = uSunlightViewMatrix * vec4(rayStart, 1.0); // This avoids matrix multiply per sample
    vec4 sunlightViewPosEnd = uSunlightViewMatrix * vec4(rayEnd, 1.0);
    vec4 sunlightViewAngleStart = uSunlightViewMatrix * vec4(rayStart, 0.0);
    vec4 sunlightViewAngleEnd = uSunlightViewMatrix * vec4(rayEnd, 0.0);
    vec3 camLook = vec3(uViewMatrix[0][2], uViewMatrix[1][2], uViewMatrix[2][2]);

    int cascadeCount = int(uMappingParams.w);
    float extinction = 0.0;

    float alongStep = 1.0 / float(steps);
    float along = offset * alongStep;
    float stepSize = sampledRayLength * alongStep;
    vec2 inScatterAccumulation = vec2(0.0, 0.0); // x = Lit fog amount, y = Unlit fog amount
#if defined(VOLUMETRIC_SCATTERING_NOISE)
    vec3 timeOffset = vec3(0.05 * uTime);
    vec3 windWork = u3DNoiseWind_Power.xyz * uTime;
    const float secondarySampleScale = 0.31;
    float scaledStrength = u3DNoiseFrequency_Strength.x * secondarySampleScale;
    vec3 windWorkStength = windWork * u3DNoiseFrequency_Strength.x;
    vec3 windWorkScaledStrength = windWork * scaledStrength;
#endif
    for (int i = 0; i < steps; ++i)
    {
        vec3 worldPos = mix(rayStart, rayEnd, along);
        vec4 sunlightViewPos = mix(sunlightViewPosStart, sunlightViewPosEnd, along);

        // Angular Noise
    #if defined(VOLUMETRIC_SCATTERING_NOISE)
        vec4 sunlightViewAngle = mix(sunlightViewAngleStart, sunlightViewAngleEnd, along);
        vec3 samplePos1 = (sunlightViewAngle.xyx * vec3(0.001)) + timeOffset;
        vec3 samplePos2 = (sunlightViewAngle.xyx * vec3(0.001)) - timeOffset;
        sunlightViewPos.xz += ((vec2(texture3D(sNoiseTex,samplePos2).r, texture3D(sNoiseTex,samplePos2).r)  * 128.0) - 64.0);
    #endif

        float viewDepth = 0.0;
    #if defined(USE_CASCADE_SPLIT_SELECTION)
        //float viewDepth = sampledRayLength * along; // This doesn't work as viewDepth does not correspond correctly with sampled world position
        // Camera plane distance more accurate and must be used for cascade interval select to work without artifacts
        vec3 rayViewDir = worldPos.xyz - uCameraPosition;
        viewDepth = abs(dot(rayViewDir, camLook));
    #endif
        float amountInSunlight = ShadowSample(vec4(worldPos, 1.0f), sunlightViewPos, viewDepth);

        float density = 1.0;
        float heightDensity = 1.0;
    #if defined(VOLUMETRIC_GROUND_FOG)
        if (uGroundFogHeight_Falloff.y != 0.0)
        {
            float height = max(0.0, (worldPos.y - uGroundFogHeight_Falloff.x) * uGroundFogHeight_Falloff.y);
            heightDensity = exp(-height) * 100.0;
        }
    #endif

        // 3D Noise
    #if defined(VOLUMETRIC_SCATTERING_NOISE)
        if (u3DNoiseFrequency_Strength.y != 0.0)
        {
            vec3 samplePos = worldPos * u3DNoiseFrequency_Strength.x + windWorkStength;
            float noiseDensity = float(texture3D(sNoiseTex, samplePos));

            vec3 samplePosB = worldPos * scaledStrength + windWorkScaledStrength;
            float noiseDensityB = float(texture3D(sNoiseTex, samplePosB));

            float mixedDensity = pow(mix(noiseDensity, noiseDensityB, 0.8) + 0.5, u3DNoiseWind_Power.w);
            heightDensity *= max(0, mix(1.0, mixedDensity, u3DNoiseFrequency_Strength.y));
        }
    #endif
        density += heightDensity;
        float atmosphere = density * stepSize;
        float scattering = uVolumetricScatteringParameters.x * atmosphere;
        extinction += uVolumetricScatteringParameters.y * atmosphere;

        inScatterAccumulation += (scattering * exp(-extinction)) * vec2(amountInSunlight, 1.0 - amountInSunlight);
        along += alongStep;
    }
    // Do an extra step at the end
    if (cameraDistance > distanceInsideShadowMap)
    {
        float atmosphere = cameraDistance - distanceInsideShadowMap;
        float scattering = uVolumetricScatteringParameters.x * atmosphere;
        extinction += uVolumetricScatteringParameters.y * atmosphere;
        inScatterAccumulation += vec2(scattering * exp(-extinction), 0.0);
    }

    float totalAccumulation = inScatterAccumulation.x + inScatterAccumulation.y;
    if (totalAccumulation > 0.0)
    {
        float totalProportionLit = inScatterAccumulation.x / totalAccumulation;
        float exaggerationFactor = uVolumetricScatteringParameters.w;

        totalProportionLit = pow(totalProportionLit, exaggerationFactor);
        inScatterAccumulation.xy = totalAccumulation * vec2(totalProportionLit, 1.0 - totalProportionLit);

        inScatterAccumulation.x = inScatterAccumulation.x * PhaseFunction(dot(viewDirection, uInvSunDirection), mieG); // Only do phase function for lit fog amount
    }
    return vec4(inScatterAccumulation.x, extinction, inScatterAccumulation.y, 1.0); // We put the unlit fog amount in the z channel here since it needs less precision (our output pixel format is the 11,11,10 HDR)
}
vec4 GetScatteredInRay2(int steps, vec3 viewDirection, float cameraDistance, float offset)
{
    return GetScatteredInRay(steps, viewDirection, cameraDistance, offset, uMieG);
}
#endif

#endif // VOLUMETRIC_FUNCTIONS_INC
