#include "Shared/DistanceFog/Uniforms.inc"
#include "Shared/DistanceFog/Functions.inc"
#if defined(TEXTURE_ATLAS)
    #include "Shared/TerrainTexture.inc"
#endif
#include "Shared/Caustics.inc"
#include "Shared/LightingUtils.inc"
#include "Shared/StippleTransparencyUtils.inc"


void main()
{
    highp vec3 worldPos = vWorldPos_ViewSpaceDepth.xyz;

#if defined(CLIP_PLANE)
    if (dot(vec4(worldPos, 1.0), uClipPlane) < 0.0)
    {
        //[
        discard;
        return;
        //]
    }
    #if defined(CLIP_PLANE_CLAMP)
        // For the verts previously clamped to the clip plane, also clamp them to the near plane (for CPU occlusion culling)
        gl_FragDepth = 0.0;
        // Currently valid to skip any shading, but this may not always be the case
        //[
        return;
        //]
    #endif
#endif

    vec3 worldNormal = normalize(vWorldNormal.xyz);
    highp vec3 viewVector = worldPos - uCameraPosition;

    float materialEmissiveFlag = step(0.5, fract(vMaterialProperties.w * 0.5));
    float materialHighlight;
    vec4 albedo = vec4(0.0, 0.0, 0.0, 1.0);
    // The specular level (.x) is the average of the specular values for all materials used
    // The specular exponent (.y) for terrain is the maximum value of any of the 3 materials used
    vec2 specularLevel_Exponent = vec2(0.0);

#if defined(TEXTURE_ATLAS)
    SampleTerrainTextures(albedo, specularLevel_Exponent, materialHighlight, worldNormal, worldPos, viewVector, vVertexAlbedo, vMaterialProperties, materialEmissiveFlag);
#elif defined(VERTEX_COLOUR)
    #if defined(DEBUG_TEXEL_DENSITY)
        gl_FragColor = vec4(1.0);
    #endif
    albedo = vVertexAlbedo;
#endif

#if defined(DEBUG_TERRAIN_SPECULAR_MAX)
    if (uDebugTerrainSpecular.x >= 0.0)
    {
        specularLevel_Exponent.y = max(specularLevel_Exponent.y, uDebugTerrainSpecular.x);
    }
    if (uDebugTerrainSpecular.y >= 0.0)
    {
        specularLevel_Exponent.y = uDebugTerrainSpecular.y;
    }
#endif

// Return if using these debug types - note gl_FragColor is set in SampleTerrainTextures
#if defined(DEBUG_TEXEL_DENSITY) || defined(DEBUG_TANGENTS) || defined(DEBUG_BITANGENTS)
    if (uDebugReturn != 0.0)
    {
        return;
    }
#endif

#if defined(STIPPLE_TRANSPARENCY_ENABLED)
    if (!IsStipplePixelVisible(viewVector, gl_FragCoord.xy, albedo.a))
    {
        discard;
        return;
    }
#endif

    //Lighting...

    //Calculate specular level using weighted alphas
    float specularGlossMask = 1.0;
#if defined(TEXTURE_ATLAS)
    // NOTE: Scale the specular strength up to compensate for multiplying the albedo lighting into the specular term below
    specularGlossMask += specularLevel_Exponent.x * 4.0;
#endif

#if defined(VIEWPORTLOOKUPSCALE)
    vec2 screenTexCoords = gl_FragCoord.xy * uFullScreenLookupScale.xy;
#endif

    // Ambient occlusion
    float ambientOcclusion = 1.0;
#if defined(SSAO)
    ambientOcclusion = texture2D(uSSAOMap, screenTexCoords).r;
#endif

    // Diffuse irradiance
    vec3 indirectLightingDiffuse;
#if defined(IRRADIANCE_LIGHTING)
    #if defined(NORMAL_MAP)
        // We have to do irradiance calculation in the fragment shader for worldNormal mapped terrain in order to see the normal map
        indirectLightingDiffuse = uAmbientColour * EvaluateSHLighting2ndOrder(worldNormal, uIrradianceSHCoefs);
    #else
        indirectLightingDiffuse = vAmbientColour;
    #endif
#else
    indirectLightingDiffuse = uAmbientColour;
#endif
    // Apply ambient occlusion to indirect diffuse lighting
#if defined(SSAO)
    indirectLightingDiffuse *= ambientOcclusion;
#endif

    highp float cameraDistance = length(viewVector);
    // Normalize view direction reusing length value
    highp vec3 viewDirection = viewVector / cameraDistance;

    LightingTerms directLighting;
    ClearLightingTerms(directLighting);

    // Direct sunlight lighting
#if defined(SUNLIGHT_DIRECT_LIGHTING)
    int iCascade = -1;
    // Sunlight shadow attenuation
    float amountInSunlight = 0.0;

    #if defined(DEFERRED_SHADOWS)
        directLighting = EvaluateSunlightRT5(amountInSunlight, worldNormal, worldNormal, -viewDirection, screenTexCoords, specularLevel_Exponent.y, materialEmissiveFlag);
    #else
        directLighting = EvaluateSunlightRT5(iCascade, amountInSunlight, vec4(worldPos, 1.0), worldNormal, worldNormal, vWorldPos_ViewSpaceDepth.w, -viewDirection, specularLevel_Exponent.y, materialEmissiveFlag);
    #endif
#else
    directLighting.Diffuse = vec3(1.0);
#endif

#if defined(TEXTURE_ATLAS) && defined(DEBUG_MATERIAL_HIGHLIGHT)
    if (uDebugMaterialHighlight != -1.0)
    {
        float lightStrength = mix(0.1, 1.0, length(directLighting.Diffuse));
        float base = 0.1;
        vec3 col = mix(vec3(base) * lightStrength, vec3(0,1,0), materialHighlight);
        gl_FragColor = vec4(col, 1);
        if (uDebugReturn != 0.0)
        {
            return;
        }
    }
#endif

    vec4 fragColour = vec4(0.0, 0.0, 0.0, albedo.a);

#if defined(POINT_LIGHTING)
    vec3 pointLightsTerm = vec3(0.0, 0.0, 0.0);
    vec3 pointLightsSpecular = vec3(0.0, 0.0, 0.0);
    const vec3 specReflDefault = vec3(0.65, 0.65, 0.65);
    const float perceptualRoughnessDefault = 1.0;
    const float fresnelExpDefault = 5.0;
    EvaluatePointLights(
        pointLightsTerm,
        pointLightsSpecular,
        specReflDefault,
        specularLevel_Exponent.y,
        perceptualRoughnessDefault,
        fresnelExpDefault,
        -viewDirection,
        worldPos,
        worldNormal,
        vWorldPos_ViewSpaceDepth.w);

    #if defined(DEBUG_POINTLIGHTS)
        gl_FragColor = vec4(pointLightsTerm,fragColour.a);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DEBUG_POINTLIGHTS_SPECULAR)
        gl_FragColor = vec4(pointLightsSpecular,fragColour.a);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DIFFUSE_LIGHTING)
        directLighting.Diffuse += pointLightsTerm;
    #else
        directLighting.Diffuse = pointLightsTerm;
    #endif

    #if defined(POINT_LIGHTING_SPECULAR)
        directLighting.Specular += pointLightsSpecular;
    #endif
#else
    #if defined(DEBUG_POINTLIGHTS)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DEBUG_POINTLIGHTS_SPECULAR)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if !defined(DIFFUSE_LIGHTING)
        directLighting.Diffuse = vec3(0.0);
    #endif
#endif

    float causticsTerm = 0.0;
#if defined(CAUSTICS)
    causticsTerm = CalculateCausticsTerm(worldPos, amountInSunlight, worldNormal);
#endif

    //Final compositing of lighting terms...

    // Ambient
#if defined(AMBIENT_LIGHTING)
    fragColour.rgb += indirectLightingDiffuse;
    //fragColour.rgb += indirectLighting.Specular * specularGlossMask;
#endif
    // Diffuse
    fragColour.rgb += directLighting.Diffuse;
    // Specular
#if defined(SPECULAR_LIGHTING)
    fragColour.rgb += directLighting.Specular * specularGlossMask;
#endif
// Caustics
#if defined(CAUSTICS)
    fragColour.rgb += uSunColour * causticsTerm;
#endif
    // Albedo
#if defined(ALBEDO_LIGHTING)
    fragColour.rgb *= albedo.rgb;
#endif
    //Emissive
#if defined(TEXTURE_ATLAS) && defined(SUNLIGHT_SHADOWS)
    #if defined(DEBUG_EMISSIVE_MAP)
        fragColour.rgb = vec3(materialEmissiveFlag);
    #endif
#elif defined(DEBUG_EMISSIVE_MAP)
    fragColour.rgb = vec3(0.0);
#endif
#if defined(TEXTURE_ATLAS) && !defined(DEBUG_EMISSIVE_MAP)
    fragColour.rgb = mix(fragColour.rgb, albedo.rgb, materialEmissiveFlag);
#endif

#if defined(DISCRETIZED_LUMINANCE)
    fragColour.rgb = DiscretizeLuminance(fragColour.rgb);
#endif

    //Light scattering (now includes fog!)
#if defined(FOG_DISTANCE) || (defined(SUNLIGHT_DIRECT_LIGHTING) && defined(LIGHT_SCATTERING))
    #if defined(OGLES2_BACKEND)
        vec3 outScattering, inScattering;

        #if defined(LIGHT_SCATTERING)
            ComputeInOutScattering(viewDirection, cameraDistance, uInvSunDirection, outScattering, inScattering);
        #else
            outScattering = vec3(1.0);
            inScattering = vec3(0.0);
        #endif

        #if defined(FOG_DISTANCE)
            float fogFactor = FogBasedOnDistance(cameraDistance) * FogBasedOnAngle(viewDirection);
            inScattering = mix(inScattering, uFogColour.rgb, fogFactor);
            outScattering *= 1.0 - fogFactor;
        #endif

        fragColour.rgb = mix(ApplyInOutScattering(fragColour.rgb, outScattering, inScattering), fragColour.rgb, materialEmissiveFlag);
    #else
        fragColour.rgb = mix(ApplyInOutScattering(fragColour.rgb, vOutScattering, vInScattering), fragColour.rgb, materialEmissiveFlag);
    #endif
#endif

//Debug texturing
//fragColour.rgb = vec3(1,0,1);
//#if defined(TEXTURE_ATLAS)
//    fragColour.rgb = vec3(vTextureWeight.xyz); //Texture weights
//#endif
//Debug lighting
#if defined(DEBUG_ALBEDO)
    fragColour = albedo;
#endif
#if defined(DEBUG_NORMALS)
    fragColour.rgb = worldNormal * 0.5 + 0.5;
    fragColour.a = 1.0;
#endif
#if defined(DEBUG_SPECULAR_MAP)
    #if defined(TEXTURE_ATLAS)
        //Remove the default specular amount (0.5) and divide by the specular scale we applied earlier (4.0)
        //This outputs the texture-supplied specular mask
        fragColour.rgb = vec3(max(0.0, (specularGlossMask - 0.5) / 4.0));
    #else
        fragColour.rgb = vec3(0.0);
    #endif
#endif
#if defined(SUNLIGHT_SHADOWS) && defined(DEBUG_SUNLIGHT_SHADOW_CASCADE) && !defined(DEFERRED_SHADOWS)
    fragColour.rgb *= ShadowMapCascadeColour(iCascade, int(uMappingParams.w)).rgb;
#endif
// RT7 debug, not in use so display magenta
#if defined(DEBUG_RT7_EMISSIVE) || defined(DEBUG_RT7_METALNESS) || defined(DEBUG_RT7_ROUGHNESS)
    fragColour = vec4(1.0, 0.0, 1.0, 1.0);
#endif

    gl_FragColor = fragColour;
    //}
}
