#include "Shared/DistanceFog/Uniforms.inc"
#include "Shared/DistanceFog/Functions.inc"
#include "Shared/TerrainTextureRT7.inc"
#include "Shared/Caustics.inc"
#include "lighting/brdf/Fresnel.inc"
#include "lighting/Lighting.inc"

void main()
{
    highp vec3 worldPos = vWorldPos_ViewSpaceDepth.xyz;

#if defined(CLIP_PLANE)
    if (dot(vec4(worldPos, 1.0), uClipPlane) < 0.0)
    {
        discard;
        return;
    }
    #if defined(CLIP_PLANE_CLAMP)
        // For the verts previously clamped to the clip plane, also clamp them to the near plane (for CPU occlusion culling)
        gl_FragDepth = 0.0;
        // Currently valid to skip any shading, but this may not always be the case
        return;
    #endif
#endif

    vec3 worldNormal = normalize(vWorldNormal.xyz);
    highp vec3 viewVector = worldPos - uCameraPosition;

    vec4 albedo = vec4(0.0, 0.0, 0.0, 1.0);
    vec2 specularLevel_Exponent = vec2(0.0);
#if defined(TEXTURE_ATLAS)
    SampleTerrainTextures(albedo, specularLevel_Exponent, worldNormal, worldPos, viewVector, vVertexAlbedo, vMaterialProperties);
#elif defined(VERTEX_COLOUR)
    #if defined(DEBUG_TEXEL_DENSITY)
        gl_FragColor = vec4(1.0);
    #elif defined(DEBUG_RT7_EMISSIVE) || defined(DEBUG_RT7_METALNESS) || defined(DEBUG_RT7_ROUGHNESS)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);
    #endif
    albedo = vVertexAlbedo;
#endif

    // Return if using these debug types - note gl_FragColor is set in SampleTerrainTextures
#if defined(DEBUG_TEXEL_DENSITY) || defined(DEBUG_RT7_EMISSIVE) || defined(DEBUG_RT7_METALNESS) || defined(DEBUG_RT7_ROUGHNESS) || defined(DEBUG_TANGENTS) || defined(DEBUG_BITANGENTS)
    if (uDebugReturn != 0.0)
    {
        return;
    }
#endif

    //Lighting...

    //Calculate specular level using weighted alphas
    float specularGlossMask = 0.5;
#if defined(TEXTURE_ATLAS)
    // NOTE: Scale the specular strength up to compensate for multiplying the albedo lighting into the specular term below
    specularGlossMask += specularLevel_Exponent.x * 4.0;
#endif

#if defined(VIEWPORTLOOKUPSCALE)
    vec2 fullscreenTexCoords = gl_FragCoord.xy * uViewportLookupScale.xy;
#endif

    LightingTerms indirectLighting;
    ClearLightingTerms(indirectLighting);

    // Diffuse irradiance
#if defined(IRRADIANCE_LIGHTING)
    #if defined(NORMAL_MAP)
        // We have to do irradiance calculation in the fragment shader for worldNormal mapped terrain in order to see the normal map
        indirectLighting.Diffuse = uAmbientColour * EvaluateSHLighting2ndOrder(worldNormal, uIrradianceSHCoefs);
    #else
        indirectLighting.Diffuse = vAmbientColour;
    #endif
#else
    indirectLighting.Diffuse = uAmbientColour;
#endif

    highp float cameraDistance = length(viewVector);
    // Normalize view direction reusing length value
    highp vec3 viewDirection = viewVector / cameraDistance;

    float materialEmissiveFlag = step(0.5, fract(vMaterialProperties.w * 0.5));

    LightingTerms directLighting;
    ClearLightingTerms(directLighting);

    // Direct sunlight lighting
#if defined(SUNLIGHT_DIRECT_LIGHTING)
    int iCascade = -1;
    // Sunlight shadow attenuation
    float amountInSunlight = 0.0;

    #if defined(DEFERRED_SHADOWS)
        directLighting = EvaluateSunlightRT5(amountInSunlight, worldNormal, -viewDirection, fullscreenTexCoords, specularLevel_Exponent.y, materialEmissiveFlag);
    #else
        directLighting = EvaluateSunlightRT5(iCascade, amountInSunlight, vec4(worldPos, 1.0), worldNormal, vWorldPos_ViewSpaceDepth.w, -viewDirection, specularLevel_Exponent.y, materialEmissiveFlag);
    #endif
#else
    directLighting.Diffuse = vec3(1.0);
#endif

    vec4 fragColour = vec4(0.0, 0.0, 0.0, albedo.a);

#if defined(POINT_LIGHTING)
    vec3 pointLightsTerm = vec3(0.0, 0.0, 0.0);
    vec3 pointLightsSpecular = vec3(0.0, 0.0, 0.0);
    EvaluatePointLights(pointLightsTerm, pointLightsSpecular, 0.65, specularLevel_Exponent.y, 1.0, 5.0, -viewDirection, worldPos, worldNormal, vWorldPos_ViewSpaceDepth.w);

    #if defined(DEBUG_POINTLIGHTS)
        gl_FragColor = vec4(pointLightsTerm,fragColour.a);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DEBUG_POINTLIGHTS_SPECULAR)
        gl_FragColor = vec4(pointLightsSpecular,fragColour.a);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DIFFUSE_LIGHTING)
        directLighting.Diffuse += pointLightsTerm;
    #else
        directLighting.Diffuse = pointLightsTerm;
    #endif

    #if defined(POINT_LIGHTING_SPECULAR)
        directLighting.Specular += pointLightsSpecular;
    #endif
#else
    #if defined(DEBUG_POINTLIGHTS)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DEBUG_POINTLIGHTS_SPECULAR)
        gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if !defined(DIFFUSE_LIGHTING)
        directLighting.Diffuse = vec3(0.0);
    #endif
#endif

    float causticsTerm = 0.0;
#if defined(CAUSTICS)
    causticsTerm = CalculateCausticsTerm(worldPos, amountInSunlight, worldNormal);
#endif

    // Combined lighting terms
    vec3 totalDiffuse = vec3(0.0);
    vec3 totalSpecular = vec3(0.0);

    // Ambient
#if defined(AMBIENT_LIGHTING)
    // Apply ambient occlusion to indirect lighting
    #if defined(SSAO) && !defined(REFRACTION)
        float ambientOcclusion = texture2D(uSSAOMap, fullscreenTexCoords).r;
        indirectLighting.Diffuse *= ambientOcclusion;
        indirectLighting.Specular *= ambientOcclusion;
    #endif

    totalDiffuse += indirectLighting.Diffuse;
    totalSpecular += indirectLighting.Specular;
#endif
    // Diffuse
    totalDiffuse += directLighting.Diffuse;
    // Specular
#if defined(SPECULAR_LIGHTING)
    totalSpecular += directLighting.Specular;
#endif
    // Albedo
#if defined(ALBEDO_LIGHTING)
    totalDiffuse *= albedo.rgb;
#endif

    // Although combining diffuse and specular here doesn't follow the laws of conservation of energy, for pure non-metals (i.e. metalness of 0.0),
    // diffuse albedo will actually be set to black, which has almost the same effect.
    fragColour.rgb = totalDiffuse + totalSpecular;

    // Caustics
#if defined(CAUSTICS)
    fragColour.rgb += uSunColour * causticsTerm;
#endif

    //Light scattering (now includes fog!)
#if defined(FOG_DISTANCE) || (defined(SUNLIGHT_DIRECT_LIGHTING) && defined(LIGHT_SCATTERING))
    #if defined(OGLES2_BACKEND)
        vec3 outScattering, inScattering;

        #if defined(LIGHT_SCATTERING)
            ComputeInOutScattering(viewDirection, cameraDistance, uInvSunDirection, outScattering, inScattering);
        #else
            outScattering = vec3(1.0);
            inScattering = vec3(0.0);
        #endif
        #if defined(FOG_DISTANCE)
            float fogFactor = FogBasedOnDistance(cameraDistance) * FogBasedOnAngle(viewDirection);
            inScattering = mix(inScattering, uFogColour.rgb, fogFactor);
            outScattering *= 1.0 - fogFactor;
        #endif

        fragColour.rgb = ApplyInOutScattering(fragColour.rgb, outScattering, inScattering);
    #else
        fragColour.rgb = ApplyInOutScattering(fragColour.rgb, vOutScattering, vInScattering);
    #endif
#endif

    // Debug RT7 maps
#if defined(DEBUG_RT7_EMISSIVE)
    fragColour = vec4(texelEmissiveNormal.rrr, 1.0);
#endif
#if defined(DEBUG_RT7_METALNESS)
    fragColour = vec4(texelMetalnessRoughness.rrr, 1.0);
#endif
#if defined(DEBUG_RT7_ROUGHNESS)
    fragColour = vec4(texelMetalnessRoughness.ggg, 1.0);
#endif

#if defined(FORCE_OPAQUE)
    fragColour.a = 1.0;
#endif

    //Emissive
#if defined(TEXTURE_ATLAS) && defined(SUNLIGHT_SHADOWS)
    #if defined(DEBUG_EMISSIVE_MAP)
        fragColour.rgb = vec3(materialEmissiveFlag);
    #endif
#elif defined(DEBUG_EMISSIVE_MAP)
    fragColour.rgb = vec3(0.0);
#endif
#if defined(TEXTURE_ATLAS) && !defined(DEBUG_EMISSIVE_MAP)
    fragColour.rgb = mix(fragColour.rgb, albedo.rgb, materialEmissiveFlag);
#endif

    //Debug texturing
    //fragColour.rgb = vec3(1,0,1);
    //#if defined(TEXTURE_ATLAS)
    //    fragColour.rgb = vec3(vTextureMeta[0].xyz); //Texture weights
    //#endif
    //Debug lighting
#if defined(DEBUG_ALBEDO)
    fragColour = albedo;
#endif
#if defined(DEBUG_NORMALS)
    fragColour.rgb = worldNormal * 0.5 + 0.5;
    fragColour.a = 1.0;
#endif
#if defined(DEBUG_SPECULAR_MAP)
    #if defined(TEXTURE_ATLAS)
        //Remove the default specular amount (0.5) and divide by the specular scale we applied earlier (4.0)
        //This outputs the texture-supplied specular mask
        fragColour.rgb = vec3(max(0.0, (specularGlossMask - 0.5) / 4.0));
    #else
        fragColour.rgb = vec3(0.0);
    #endif
#endif
#if defined(SUNLIGHT_SHADOWS) && defined(DEBUG_SUNLIGHT_SHADOW_CASCADE)
    fragColour.rgb *= ShadowMapCascadeColour(iCascade, int(uMappingParams.w)).rgb;
#endif

    gl_FragColor = fragColour;
}
