#ifndef TEXTURE_CUBE_UTILS_INC
#define TEXTURE_CUBE_UTILS_INC

#include "Common.inc"
#include "ConversionUtils.inc"

#define COMPUTE_FACE_INDEX 0


/**
    Converts an input 3D vector to 2D texture coordinates used for sampling a cube map face.
    Memory layout for each face of the cube map as an array of textures is assumed as follows

    ----------------------- U ------------------------
    | ************************************************
    | *      **      **      **      **      **      *
    V * 0 +x ** 1 -x ** 2 +y ** 3 -y ** 4 +z ** 5 -z *
    | *      **      **      **      **      **      *
    | ************************************************

    @return xy: Texture coordinates on the face in the range 0-1 UV. z: face index of cube map
 */
vec3 Vec3To2DCubeMapFaceTexCoords(inout vec4 uvExtentsOut, in vec3 location, in CubeMapFaceUVExtents atlasFaceUVExtents)
{
    vec3 locationAbs = abs(location);
    vec2 faceTexCoords = vec2(0.0, 0.0);
    float ma = 0.0;
    float faceIndex = 0.0;

    // The sign of the largest axis is used to determine the face index

    // Right or left face
    if (locationAbs.x >= locationAbs.y && locationAbs.x >= locationAbs.z)
    {
    #if COMPUTE_FACE_INDEX
        faceIndex = location.x < 0.0 ? 1.0 : 0.0;
    #endif
        uvExtentsOut = location.x < 0.0 ? atlasFaceUVExtents.FaceUVExtentsNegX : atlasFaceUVExtents.FaceUVExtentsPosX;

        faceTexCoords = vec2(location.x < 0.0 ? location.z : -location.z, -location.y);
        // Divide by x gives implicit normalisation when scaling texture coordinates to 0-1 range below
        // It also projects the texture coordinates onto the unit cube
        ma = 0.5 / locationAbs.x;
    }
    else if (locationAbs.y >= locationAbs.z) // Top or bottom face
    {
    #if COMPUTE_FACE_INDEX
        faceIndex = location.y < 0.0 ? 3.0 : 2.0;
    #endif
        uvExtentsOut = location.y < 0.0 ? atlasFaceUVExtents.FaceUVExtentsNegY : atlasFaceUVExtents.FaceUVExtentsPosY;

        faceTexCoords = vec2(location.x, location.y < 0.0 ? -location.z : location.z);
        ma = 0.5 / locationAbs.y;
    }
    else // Front or back face
    {
    #if COMPUTE_FACE_INDEX
        faceIndex = location.z < 0.0 ? 5.0 : 4.0;
    #endif
        uvExtentsOut = location.z < 0.0 ? atlasFaceUVExtents.FaceUVExtentsNegZ : atlasFaceUVExtents.FaceUVExtentsPosZ;

        faceTexCoords = vec2(location.z < 0.0 ? -location.x : location.x, -location.y);
        ma = 0.5 / locationAbs.z;
    }

    // Scale texture coordinates to 0-1 range from from -1.0 - 1.0
    faceTexCoords = (faceTexCoords * ma) + 0.5;
    return vec3(faceTexCoords.x, faceTexCoords.y, faceIndex);
}

/**
 */
vec3 Vec3To2DCubeMapTexCoords(inout vec4 uvExtentsOut, vec3 location, vec2 faceHalfTexelSize, vec2 faceUVScale, in CubeMapFaceUVExtents atlasFaceUVExtents)
{
    vec3 faceTexCoords = Vec3To2DCubeMapFaceTexCoords(uvExtentsOut, location, atlasFaceUVExtents);

#if defined(CLAMP_UV_EXTENTS2)
    // Clamp shadow map texture coordinates to prevent sampling outside of shadow map cube face
    faceTexCoords.x = clamp(faceTexCoords.x, 0.0 + faceHalfTexelSize.x, 1.0 - faceHalfTexelSize.x);
    faceTexCoords.y = clamp(faceTexCoords.y, 0.0 + faceHalfTexelSize.y, 1.0 - faceHalfTexelSize.y);
#endif

    faceTexCoords.xy = (faceTexCoords.xy * faceUVScale.xy) + uvExtentsOut.xy;
    return faceTexCoords;
}

/**
 */
vec3 Vec3To2DCubeMapTexCoords(inout vec4 uvExtentsOut, vec3 location, vec2 faceUVScale, in CubeMapFaceUVExtents atlasFaceUVExtents)
{
    vec3 faceTexCoords = Vec3To2DCubeMapFaceTexCoords(uvExtentsOut, location, atlasFaceUVExtents);

    faceTexCoords.xy = (faceTexCoords.xy * faceUVScale.xy) + uvExtentsOut.xy;
    return faceTexCoords;
}

/*
//[ //The obfuscator doesn't like functions with multiple returns
vec4 textureCubeSRGB(samplerCube sampler, vec3 reflDir)
{
    vec4 texel = textureCube(sampler, reflDir);
    #if !defined(SRGBTEXTURES)
        return vec4(SRGBToLinear(texel.rgb), texel.a);
    #else
        return texel;
    #endif
}
//]

//[ //The obfuscator doesn't like functions with multiple returns
vec4 textureCubeSRGB(samplerCube sampler, vec3 reflDir, float lod)
{
    vec4 texel = textureCube(sampler, reflDir, lod);
    #if !defined(SRGBTEXTURES)
        return vec4(SRGBToLinear(texel.rgb), texel.a);
    #else
        return texel;
    #endif
}
//]

//[ //The obfuscator doesn't like functions with multiple returns
vec4 textureCubeSRGB(sampler, reflDir, lod)
{
    vec4 texel = textureCubeLodCompat(sampler, reflDir, lod);
    #if !defined(SRGBTEXTURES)
        return vec4(SRGBToLinear(texel.rgb), texel.a);
    #else
        return texel;
    #endif
}
//]
*/


#endif
