#include "LuminanceUtils.inc"
#include "ConversionUtils.inc"

/**
    Reinhard's advanced tone mapping equation (See Eqn#4 from
    Photographic Tone Reproduction for Digital Images for more details) is:

         (      (   Lp    ))
    Lp * (1.0f +(---------))
         (      ((Lm * Lm)))
    -------------------------
            1.0f + Lp

    Lp is the luminance at the given point, this is computed using Eqn#2 from the above paper:

           exposure
      Lp = -------- * HDRPixelIntensity
             l.r

    The exposure (key in the above paper) can be used to adjust the overall balance of
    the image. l.r is the average luminance across the scene, computed via the luminance
    downsampling process. 'HDRPixelIntensity' is the measured brightness of the current pixel
    being processed.

    Allow high luminances (bright areas of the scene) to burn out in a controllable fashion.
    Tone mapping curve is applied separately to all three colour channels of the source colour.
 */

    vec3 ReinhardAdvRGBToneMap(vec3 colour, float avgLum, float minBlackLum, float maxWhiteLum, float exposureKey, float minExposure, float maxExposure)
    {
        float colourLum = CalculateLuminance(colour);
        colourLum = max(colourLum - minBlackLum, 0.0);

        // Calculate a scaled relative luminance of specified colour
        float lumScaled = clamp(colourLum * (exposureKey / avgLum), minExposure, maxExposure);

        // Calculate compressed luminance
        float maxWhiteLumSqr = maxWhiteLum * maxWhiteLum;
        float lumCompressed = (lumScaled * (1.0 + (lumScaled / maxWhiteLumSqr))) / (1.0 + lumScaled);

        // Apply compressed luminance to input colour
        return (colour * lumCompressed);
    }

/**
    Alternative advanced Reinhard tone mapper that tone maps the luminance of the source colour
    instead and then scales the source colour by the tone mapped luminance.
 */

    vec3 ReinhardAdvLumToneMap(vec3 colour, float avgLum, float minBlackLum, float maxWhiteLum, float exposureKey, float minExposure, float maxExposure)
    {
        float colourLum = CalculateLuminance(colour);

        // Calculate a scaled relative luminance of specified colour
        float lumScaled = clamp(colourLum * (exposureKey / avgLum), minExposure, maxExposure);
        lumScaled = max(lumScaled - minBlackLum, 0.0);

        // Calculate compressed luminance
        float maxLumSqr = maxWhiteLum * maxWhiteLum;
        float lumCompressed = (lumScaled * (1.0 + (lumScaled / maxLumSqr))) / (1.0 + lumScaled);

        // Apply tonemapped luminance to the colour luminance
        return (colour / colourLum) * (lumCompressed);
    }

/**
    Filmic tone mapping operator by Jim Hejl and Richard Burgess-Dawson, which is a rational function
    designed to mimic the response curve of a type of Kodak film commonly used in film making.
 */

    vec3 HejlDawsonFilmicToneMapOp(vec3 colour)
    {
       vec3 x = max(vec3(0.0, 0.0, 0.0), colour - 0.004);
       vec3 retColor = (x * (6.2 * x + 0.5)) / (x * (6.2 * x + 1.7) + 0.06);
       return SRGBToLinear(retColor);
    }

/**
    Filmic tone mapping operator used in Uncharted 2, see http://filmicgames.com/archives/75
 */

    vec3 Uncharted2FilmicToneMapOp(vec3 x)
    {
        // Numbers from the Uncharted2 presentation
        const float A = 0.22; // Shoulder Strength
        const float B = 0.30; // Linear Strength
        const float C = 0.10; // Linear Angle
        const float D = 0.20; // Toe Strength
        const float E = 0.01; // Toe Numerator
        const float F = 0.30; // Toe Denominator

        // Note: E/F = Toe Angle
        // LinearWhite = Linear White Point Value
        return ((x * (A * x + C * B) + D * E) / (x * (A * x + B) + D * F)) - E / F;
    }

/**
    Filmic tone mapping operator used in Uncharted 2, see http://filmicgames.com/archives/75
 */

    vec3 HableFilmicToneMapOp(vec3 x)
    {
        // From http://filmicgames.com/archives/75
        const float A = 0.15; // Shoulder Strength
        const float B = 0.50; // Linear Strength
        const float C = 0.10; // Linear Angle
        const float D = 0.20; // Toe Strength
        const float E = 0.02; // Toe Numerator
        const float F = 0.30; // Toe Denominator

        // Note: E/F = Toe Angle
        // LinearWhite = Linear White Point Value
        return ((x * (A * x + C * B) + D * E) / (x * (A * x + B) + D * F)) - E / F;
    }

/**
    A filmic tone mapping like that used in Uncharted 2

    @param colour The linear HDR colour from the current pixel
    @param linearWhite  A white point that is used to increase the contrast of the image
                        This is the value that is selected to be the brightest any pixel can be.
                        Pixels brighter than the white point will be clamped to 1.0.
 */


    vec3 HejlDawsonFilmicToneMap(vec3 colour, float minBlackLum, float maxWhiteLum)
    {
        return (HejlDawsonFilmicToneMapOp(colour) / HejlDawsonFilmicToneMapOp(vec3(maxWhiteLum, maxWhiteLum, maxWhiteLum))) - HejlDawsonFilmicToneMapOp(vec3(minBlackLum, minBlackLum, minBlackLum));
    }

    vec3 Uncharted2FilmicToneMap(vec3 colour, float minBlackLum, float maxWhiteLum)
    {
        return (Uncharted2FilmicToneMapOp(colour) / Uncharted2FilmicToneMapOp(vec3(maxWhiteLum, maxWhiteLum, maxWhiteLum))) - Uncharted2FilmicToneMapOp(vec3(minBlackLum, minBlackLum, minBlackLum));
    }

    vec3 HableFilmicToneMap(vec3 colour, float minBlackLum, float maxWhiteLum)
    {
        return (HableFilmicToneMapOp(colour) / HableFilmicToneMapOp(vec3(maxWhiteLum, maxWhiteLum, maxWhiteLum))) - HableFilmicToneMapOp(vec3(minBlackLum, minBlackLum, minBlackLum));
    }

/**
    Add blue shift for very low light conditions, the rods will dominate the perception
    of light, and therefore color will be desaturated and shifted towards blue.
 */

    vec3 BlueShift(vec3 colour, float avgLum, float exposureKey, float blueShiftCoef)
    {
        float colourLum = CalculateLuminance(colour);

        // Compute scotopic luminance
        vec3 rodColour = BLUE_SHIFT_SCALE * colourLum;

        // Calculate a scaled relative luminance of specified colour
        float lumScaled = exposureKey * (colourLum / avgLum);
        float blueShiftLerp = min(1.0, (1.0 - lumScaled) * blueShiftCoef);

        // Lerp between current color and blue, desaturated copy
        return mix(colour, rodColour, blueShiftLerp);
    }

/**
    Given a color, desaturate low-light color to simulate night-time
    and eyes not being able to distinguish color anymore
 */

    vec3 NightDesaturate(vec3 colour)
    {
        // Desaturate low-light pixels (per-pixel, so lit areas are unaffected)
        // Calculate black & white value, looking at brightness of red/green/blue individually
        float bw = 0.28 * colour.r + 0.59 * colour.b + 0.13 * colour.b;
        // Fully desaturated colour to interpolate to - add a bit of a blue shift
        // that seems to give a 'night' feeling
        vec3 bwColor = vec3(bw * 0.8, bw * 0.8, bw * 1.1);
        // Desaturation amount
        float desat = exp(-3.0 * bw);
        // Always leave a bit of colour
        desat = min(desat, 0.99);
        colour = mix(colour, bwColor, desat);

        return colour;
    }

/**
    Calculate the colour based on exposure
 */

    vec3 CalcExposedColor(vec3 colour, float avgLum, float exposureKey, float minExposure, float maxExposure)
    {
        // Calculate a scaled relative luminance of specified colour
        float linearExposure = (exposureKey / avgLum);
        linearExposure = clamp(linearExposure, minExposure, maxExposure);
        return colour * linearExposure;
        // Try this version if fullscreen flashes are still present
        //float autoExposure = log2(max(linearExposure, 0.0001));
        //return colour * exp2(autoExposure);
    }

/**
   Main tone mapping function

   @return colourHDR mapped to a low dynamic range
 */
    vec3 ToneMap(vec3 colourHDR, float avgLum, float minBlackLum, float maxWhiteLum, float exposureKey, float minExposure, float maxExposure)
    {
        vec3 colourLDR;
    #if defined(REINHARD_ADVANCED_RGB)
        colourLDR = ReinhardAdvRGBToneMap(colourHDR, avgLum, minBlackLum, maxWhiteLum, exposureKey, minExposure, maxExposure);
    #elif defined(REINHARD_ADVANCED_LUM)
        colourLDR = ReinhardAdvLumToneMap(colourHDR, avgLum, minBlackLum, maxWhiteLum, exposureKey, minExposure, maxExposure);
    #elif defined(FILMIC_HEJL_DAWSON)
        colourLDR = HejlDawsonFilmicToneMap(CalcExposedColor(colourHDR, avgLum, exposureKey, minExposure, maxExposure), minBlackLum, maxWhiteLum);
    #elif defined(FILMIC_UNCHARTED2)
        colourLDR = Uncharted2FilmicToneMap(CalcExposedColor(colourHDR, avgLum, exposureKey, minExposure, maxExposure), minBlackLum, maxWhiteLum);
    #elif defined(FILMIC_HABLE)
        colourLDR = HableFilmicToneMap(CalcExposedColor(colourHDR, avgLum, exposureKey, minExposure, maxExposure), minBlackLum, maxWhiteLum);
    #else
        #error No tone mapping operator specified
    #endif

        return colourLDR;
    }
