uniform SAMPLER_2D_AUTO_MULTISAMPLE sSourceTex;
uniform SAMPLER_2D_AUTO_MULTISAMPLE sDepthTex;
uniform sampler2D sDitherTex;

uniform mat4 uViewMatrix;
uniform mat4 uInvViewProjMatrix;
uniform vec3 uCameraPosition;
//x = near * far, y = near - far, z = far, w = 1 / far
uniform vec4 uZBufferParams;

uniform int uSampleCount;

// x = 1 - g^2,  y = 1 + g^2,  z = 2 * g,  w = 1/4PI
uniform vec4 uMieGSky;

in vec2 vTexCoord;

#define SUNLIGHT_SHADOWS 1
#define VOLUMETRIC_SCATTERING 1
#define SUNLIGHT_DIRECT_LIGHTING 1

#include "Shared/LightScattering/VSUniforms.inc"
#include "Shared/VolumetricScattering/VolumetricFunctions.inc"
#include "Shared/VolumetricScattering/ApplyVolumetrics.inc"
#include "Shared/DepthUtils.inc"

vec4 FetchWorldPos(vec2 uv)
{
    float depth;
#if defined(MSAA)
    depth = float(texelFetch(sDepthTex, ivec2(uv * uViewportLookupScale.zw), gl_SampleID)); //2DMultiSampled Sample
#else
    depth = float(texture2DLodCompat(sDepthTex, uv, 0.0)); // 2D LOD
#endif
    vec4 clipPos = vec4(2.0 * uv - 1.0, 2.0 * depth - 1.0, 1.0);
    vec4 worldPos = uInvViewProjMatrix * clipPos;
    worldPos.xyz = worldPos.xyz / worldPos.w;
    worldPos.w = GetViewSpaceDepth(depth, uZBufferParams.xyz) * uZBufferParams.w; // worldPos.w here is being used as a 0-1 value between the near and far planes
    return worldPos;
}

void main()
{
    vec4 worldPos = FetchWorldPos(vTexCoord);

    vec3 viewDirection = worldPos.xyz - uCameraPosition;
    float cameraDistance = length(viewDirection);
    viewDirection /= cameraDistance;

    vec2 screenPos = vTexCoord * uViewportLookupScale.zw;
    vec2 interleavedPos = (mod(floor(screenPos.xy), DITHER_SIZE) + 0.5) * INV_DITHER_SIZE;
    float offset = texture2D(sDitherTex, interleavedPos).r;

    vec4 mieG = mix(uMieG, uMieGSky, worldPos.w);

    vec4 scattering = GetScatteredInRay(uSampleCount, viewDirection, cameraDistance, offset, mieG);

    float skyboxExtintionCoef = mix(1.0, uVolumetricScatteringParameters.z, worldPos.w);
    scattering.rgb *= skyboxExtintionCoef;

#if defined(OUTPUT_SCATTERING_ONLY)
    gl_FragColor = scattering;
#else // Fullres scattering with compositing onto full resolution destination render target
    vec4 fullResDestinationSample;
    #if defined(MSAA)
        fullResDestinationSample = texelFetch(sSourceTex, ivec2(vTexCoord * vec2(textureSize(sSourceTex))), gl_SampleID);
    #else
        fullResDestinationSample = texture2DMultisample(sSourceTex, vTexCoord, vec2(textureSize(sSourceTex)));
    #endif
    gl_FragColor = ApplyVolumetricScattering(fullResDestinationSample, scattering);
#endif
}
