#include "lighting/brdf/Fresnel.inc"
#include "Shared/TextureCompression.inc"
#include "Shared/Caustics.inc"
#include "Shared/DepthUtils.inc"
#include "Shared/NoiseUtils.inc"
#include "Shared/LightingUtils.inc"
#include "Shared/DistanceFog/Uniforms.inc"
#include "Shared/DistanceFog/Functions.inc"
#include "UndefineInOut.inc"

/*********************************************************************************
 * Water surface - ie albedo, normals and alpha of then surface
 *********************************************************************************/

/**
 * Returns noise for sampling foam
 */
float GenerateNoise()
{
    const float tileSize = 512.0;
    const float noiseScale = 0.125;
    return clamp(cnoise(vPosition_WaterDepth.xz / tileSize) * noiseScale, 0.0, 1.0);
}

/**
 * Returns the flow vector at this fragment
 */
vec2 GetCombinedFlow()
{
    return vNoisyPatchFlow0_NoisyPatchFlow1.xy * vFlowControlMask_ViewSpaceDepth.x
         + vNoisyPatchFlow0_NoisyPatchFlow1.zw * vFlowControlMask_ViewSpaceDepth.y
         + vNoisyPatchFlow2_UVPack_NormalsFlow0Map0.xy * vFlowControlMask_ViewSpaceDepth.z;
}

#if defined(WATER_NORMAL_MAPS)
    /**
     * Samples the given normal map while applying the specified weighting and uv distortion in the NormalMapSamplingParams params argument.
     * Returns a vec4 with xy = xy weighted by params but unscaled by flow strength, zw = dxdy weighted by params and scaled by control mask and flow strength
     */
    vec4 WeightedNormalMap_XY_DXDY(highp float weight, sampler2D normalMap, vec2 uv, vec2 flow, float controlMask)
    {
        if (weight <= 0.0)
        {
            return vec4(0.0);
        }

        vec3 normal = UnpackNormal(texture2D(normalMap, uv));
        vec2 dxdy = normal.xy / normal.z;

        // Scale the normal based upon flow rate - still water should appear flatter.
        const float flowWeightScale = 6.0;
        float normalStrength = clamp(length(flow) * flowWeightScale + uWaterStillWaterNormalStrength_spareyzw.x, 0.0, 1.0);

        return vec4(normal.xy, dxdy * normalStrength) * weight * controlMask;
    }

    /**
     * Samples all normal maps and returns a weighted sum of the xy partial derivatives
     */
    vec2 GetNormalDXDYWeightedSum(
        highp vec4 sampleWeight_uvDistortion_sparezw0,
        highp vec4 sampleWeight_uvDistortion_sparezw1,
        highp vec4 sampleWeight_uvDistortion_sparezw2,
        vec2 uv0,
        vec2 uv1,
        vec2 uv2,
        vec2 flow,
        float uvDistortionScale)
    {
        vec4 weightedSum = vec4(0.0);

        weightedSum += WeightedNormalMap_XY_DXDY(sampleWeight_uvDistortion_sparezw0.x, uWaterNormalMapTexture0, uv0, flow, 1.0);

        uv1 += weightedSum.xy * sampleWeight_uvDistortion_sparezw1.y * uvDistortionScale;
        weightedSum += WeightedNormalMap_XY_DXDY(sampleWeight_uvDistortion_sparezw1.x, uWaterNormalMapTexture1, uv1, flow, 1.0);

        uv2 += weightedSum.xy * sampleWeight_uvDistortion_sparezw2.y * uvDistortionScale;
        weightedSum += WeightedNormalMap_XY_DXDY(sampleWeight_uvDistortion_sparezw2.x, uWaterNormalMapTexture2, uv2, flow, 1.0);

        return weightedSum.zw;
    }

    /**
     * Provides the weighted sum for the detailed normals of the water surface - ie the small waves.
     * The macro normal is supplied so it may be used to distorted texture coordinates.
     */
    vec3 WaterDetailNormalWeightedSum(vec2 uv0, vec2 uv1, vec2 uv2, vec2 flow, vec3 macroNormal)
    {
        //Apply uv distortion based on macro normal
        const float macroUVDistortionScale = 0.5;
        vec2 sampleUV0 = uv0 + macroNormal.xy * macroUVDistortionScale * uSampleWeight_uvDistortion_sparezw[0].y;
        vec2 sampleUV1 = uv1 + macroNormal.xy * macroUVDistortionScale * uSampleWeight_uvDistortion_sparezw[1].y;
        vec2 sampleUV2 = uv2 + macroNormal.xy * macroUVDistortionScale * uSampleWeight_uvDistortion_sparezw[2].y;

        const float uvDistortionScale = 0.8; //how much each sample's uv is distorted by the previous sample.
        vec2 dxdy = GetNormalDXDYWeightedSum(
            uSampleWeight_uvDistortion_sparezw[0],
            uSampleWeight_uvDistortion_sparezw[1],
            uSampleWeight_uvDistortion_sparezw[2],
            sampleUV0,
            sampleUV1,
            sampleUV2,
            flow,
            uvDistortionScale);

        return normalize(vec3(dxdy, 1.0));
    }

#if !defined(GLES2_COMPAT_MODE)
    /**
     * Provides the weighted sum of all large wave normals
     */
    vec3 WaterMacroNormalWeightedSum()
    {
        const float macroDefaultFlowSpeed = 0.25;
        const vec2 macroDefaultFlow = vec2(0.1, -0.13) * macroDefaultFlowSpeed;
        const float uvDistortion = 0.1;        // How much will each macro sample distort the uv coords of the following sample?

        vec2 dxdy = GetNormalDXDYWeightedSum(
            uMacroSampleWeight_uvDistortion_sparezw[0],
            uMacroSampleWeight_uvDistortion_sparezw[1],
            uMacroSampleWeight_uvDistortion_sparezw[2],
            vUVPack_NormalMapMacroUV_EmissiveUV[0].xy,
            vUVPack_NormalMapMacroUV_EmissiveUV[1].xy,
            vUVPack_NormalMapMacroUV_EmissiveUV[2].xy,
            macroDefaultFlow,
            uvDistortion);

        return normalize(vec3(dxdy, 1.0));
    }

#else // !defined(GLES2_COMPAT_MODE)

    /**
     * Provides the weighted sum of all large wave normals
     */
    vec3 WaterMacroNormalWeightedSum()
    {
        const float macroDefaultFlowSpeed = 0.25;
        const vec2 macroDefaultFlow = vec2(0.1, -0.13) * macroDefaultFlowSpeed;
        const float uvDistortion = 0.1; // How much will each macro sample distort the uv coords of the following sample?

        // We're in GLES2 compat mode, and in this limits the number of varyings to 8.
        // Because of this we can't pass out enough UVs from the vertex shader.
        // We'll build simplified macro UVs in the fragment shader.
        const float macroUVBaseScale = 0.1; // default scale applied to macro samples - in line with original implementation.

        vec2 staticUV0 = vPosition_WaterDepth.xz * uWaterNormalMapTextureScales_FlowNoiseScale.x * macroUVBaseScale;
        vec2 staticUV1 = vPosition_WaterDepth.xz * uWaterNormalMapTextureScales_FlowNoiseScale.y * macroUVBaseScale;
        vec2 staticUV2 = vPosition_WaterDepth.xz * uWaterNormalMapTextureScales_FlowNoiseScale.z * macroUVBaseScale;

        const float flowSpeed = 0.01;
        vec2 macroFlow0 = vec2(0.1, -0.13);
        vec2 macroFlow1 = vec2(-0.08, 0.1);
        vec2 macroFlow2 = vec2(0.11, -0.9);

        const float timeScale = 0.5;
        highp float time = uWaterTickFade.x * timeScale;

        vec2 normalMapMacroUV0 = staticUV0 + macroFlow0 * macroDefaultFlowSpeed * time;
        vec2 normalMapMacroUV1 = staticUV1 + macroFlow1 * macroDefaultFlowSpeed * time;
        vec2 normalMapMacroUV2 = staticUV2 + macroFlow2 * macroDefaultFlowSpeed * time;

        vec2 dxdy = GetNormalDXDYWeightedSum(
            uMacroSampleWeight_uvDistortion_sparezw[0],
            uMacroSampleWeight_uvDistortion_sparezw[1],
            uMacroSampleWeight_uvDistortion_sparezw[2],
            normalMapMacroUV0,
            normalMapMacroUV1,
            normalMapMacroUV2,
            macroDefaultFlow,
            uvDistortion);

        return normalize(vec3(dxdy, 1.0));
    }
#endif

    /**
     * Returns the normal for the water surface
     */
    vec3 WaterNormal()
    {
        vec3 macroNormal = WaterMacroNormalWeightedSum();

        vec3 detailNormalSum = vec3(0.0);
        vec2 flowDelta01 = vNoisyPatchFlow0_NoisyPatchFlow1.zw - vNoisyPatchFlow0_NoisyPatchFlow1.xy;
        vec2 flowDelta02 = vNoisyPatchFlow2_UVPack_NormalsFlow0Map0.xy - vNoisyPatchFlow0_NoisyPatchFlow1.xy;

        vec2 uvFlow0Map0 = vec2(0.0, 0.0);
        vec2 uvFlow0Map1 = vec2(0.0, 0.0);
        vec2 uvFlow0Map2 = vec2(0.0, 0.0);

        vec2 uvFlow1Map0 = vec2(0.0, 0.0);
        vec2 uvFlow1Map1 = vec2(0.0, 0.0);
        vec2 uvFlow1Map2 = vec2(0.0, 0.0);

        vec2 uvFlow2Map0 = vec2(0.0, 0.0);
        vec2 uvFlow2Map1 = vec2(0.0, 0.0);
        vec2 uvFlow2Map2 = vec2(0.0, 0.0);

        // Unpack tex coords, per flow per map
    #if !defined(GLES2_COMPAT_MODE)
        uvFlow0Map0 = vNoisyPatchFlow2_UVPack_NormalsFlow0Map0.zw;
        uvFlow0Map1 = vUVPack_NormalsFlow0Map1_NormalsFlow0Map2.xy;
        uvFlow0Map2 = vUVPack_NormalsFlow0Map1_NormalsFlow0Map2.zw;

        uvFlow1Map0 = vUVPack_NormalsFlow1Map0_NormalsFlow1Map1.xy;
        uvFlow1Map1 = vUVPack_NormalsFlow1Map0_NormalsFlow1Map1.zw;
        uvFlow1Map2 = vUVPack_NormalsFlow1Map2_NormalsFlow2Map0.xy;

        uvFlow2Map0 = vUVPack_NormalsFlow1Map2_NormalsFlow2Map0.zw;
        uvFlow2Map1 = vUVPack_NormalsFlow2Map1_NormalsFlow2Map2.xy;
        uvFlow2Map2 = vUVPack_NormalsFlow2Map1_NormalsFlow2Map2.zw;
    #else
        // We're in GLES2 compat mode here, and we can't pass out enough UV coords from the vertex shader.
        // We'll use the same UVs across all maps, but scale and flip for some varation.
        // This isn't very good as it basically ignores the UV and flow params for maps beyond 0 but it's
        // better than throwing out all other maps and is a quick turn around solution to get us past the post.
        // A fuller solution might move UV computation to the fragment shader, but this is more expense for a
        // platform which is likely to perform poorly already. Since we don't intend to continue ES2.0 support,
        // this is probably "good enough" to see us through.

        vec2 uvFlow0 = vNoisyPatchFlow2_UVPack_NormalsFlow0Map0.zw;
        vec2 uvFlow1 = vUVPack_NormalsFlow1Map0_NormalsFlow2Map0.xy;
        vec2 uvFlow2 = vUVPack_NormalsFlow1Map0_NormalsFlow2Map0.zw;

        uvFlow0Map0 = uvFlow0;
        uvFlow0Map1 = uvFlow0 * vec2(-0.96, 0.95);
        uvFlow0Map2 = uvFlow0 * vec2(0.97, -0.94);

        uvFlow1Map0 = uvFlow1;
        uvFlow1Map1 = uvFlow1 * vec2(-0.96, 0.95);
        uvFlow1Map2 = uvFlow1 * vec2(0.97, -0.94);

        uvFlow2Map0 = uvFlow2;
        uvFlow2Map1 = uvFlow2 * vec2(-0.96, 0.95);
        uvFlow2Map2 = uvFlow2 * vec2(0.97, -0.94);
    #endif

        const float threshold = 1.e-7;
        if(dot(flowDelta01, flowDelta01) <= threshold && dot(flowDelta02, flowDelta02) <= threshold)
        {
            // No variation in flow, we only need to consider one flow vector.
            detailNormalSum = WaterDetailNormalWeightedSum(uvFlow0Map0, uvFlow0Map1, uvFlow0Map2, vNoisyPatchFlow0_NoisyPatchFlow1.xy, macroNormal) + macroNormal;
        }
        else
        {
            // This fragment is affected by more than one flow vector, we'll need to sample normal maps for each and blend the result.
            detailNormalSum += (WaterDetailNormalWeightedSum(uvFlow0Map0, uvFlow0Map1, uvFlow0Map2, vNoisyPatchFlow0_NoisyPatchFlow1.xy, macroNormal) + macroNormal) * vFlowControlMask_ViewSpaceDepth.x;
            detailNormalSum += (WaterDetailNormalWeightedSum(uvFlow1Map0, uvFlow1Map1, uvFlow1Map2, vNoisyPatchFlow0_NoisyPatchFlow1.zw, macroNormal) + macroNormal) * vFlowControlMask_ViewSpaceDepth.y;
            detailNormalSum += (WaterDetailNormalWeightedSum(uvFlow2Map0, uvFlow2Map1, uvFlow2Map2, vNoisyPatchFlow2_UVPack_NormalsFlow0Map0.xy, macroNormal) + macroNormal) * vFlowControlMask_ViewSpaceDepth.z;
        }

        return normalize(detailNormalSum).xzy;
    }

#else

    vec3 WaterNormal()
    {
        return vec3(0.0, 1.0, 0.0);
    }

#endif

/**
 * Returns the albedo of the water surface taking foam into account
 */
vec3 WaterAlbedo(float waterDepth, float flowSpeed, float noise)
{
    vec3 albedo = vColour.rgb;

    #if defined(WATER_FOAM_MAP)
        float foamThickness = uWaterFoamScaleFoamDepth.y;
        float oldFoamTerm = clamp((1.0 - step(foamThickness, 0.0)) * min(1.0, 1.0 - min(waterDepth / foamThickness, 1.0)), 0.0, 1.0);
        float foamTime = uWaterTickFade.x + vPosition_WaterDepth.x * 0.0001 + vPosition_WaterDepth.z * 0.0001;
        float foamWaveTowardsShore = pow(max(0.0, cos(abs(waterDepth) * 0.005 + foamTime + noise * 4.0)), 8.0);
        float foamWaveLeavingShore = pow(max(0.0, cos(abs(waterDepth) * 0.005 - foamTime + noise * 8.0)), 4.0);
        float foamTerm = min(1.0, foamWaveTowardsShore + foamWaveLeavingShore * oldFoamTerm) * max(0.2 - flowSpeed, 0.0);

        #if !defined(GLES2_COMPAT_MODE)
            vec4 foamTexel = texture2D(uWaterTextureFoam, vUVPack_FoamUV.xy) * uWaterFoamScaleFoamDepth.x;
            vec3 foamColour = foamTexel.rgb * foamTexel.a;
            albedo = mix(vColour.rgb, foamColour, foamTerm * oldFoamTerm);
        #else
            albedo = vColour.rgb;
        #endif

    #endif

    return albedo;
}

/**
 * Returns the alpha of the water surface for soft edges
 */
float WaterSoftEdgeAlpha(float waterDepth)
{
    const float defaultScale = 0.004; //2m
    const float exponent = 1.0;
    float alpha;

#if defined(WATER_EXTINCTION)
    // Water extinction makes it possible to have opaque liquids. In this scenario, edge fade is visually disruptive.
    // We'll adjust the amount of edge fade based on how opaque the water is.

    const float unitsPerMetre = 0.002;
    float extinctionVisibilityGLUnits = uWaterExtinctionVisibilityMetres / unitsPerMetre;
    float scale = defaultScale + (1.0 / (extinctionVisibilityGLUnits + 0.001)); //The bias here is to prevent div by 0.
    float unclampedAlpha = pow(waterDepth * scale, exponent);
    alpha = clamp(unclampedAlpha, 0.0, 1.0);

#else
    float unclampedAlpha = pow(waterDepth * defaultScale, exponent);
    alpha = clamp(unclampedAlpha, 0.0, 1.0);
#endif

    return alpha;
}

/********************************************************************************
* Lighting, reflections and refraction + extinction.
*********************************************************************************/

float SunLightShadowAttenuation(vec3 albedo, vec3 normal, vec2 dudv, vec3 viewVector, float fresnel)
{
    float shadowAttenuation = 1.0;
    #if defined(SUNLIGHT_SHADOWS)
        int iCascade = -1;
        // 0.85 on the line below is half shadow strength (hack for water, so that we don't get black shadows)
        vec4 sunlightViewPos = uSunlightViewMatrix * vec4(vPosition_WaterDepth.xyz, 1.0);
        shadowAttenuation = DirLightShadowAtten(
            iCascade,
            vec4(vPosition_WaterDepth.xyz, 1.0) + vec4(dudv, 0.0, 0.0),
            sunlightViewPos + vec4(dudv, 0.0, 0.0),
            vFlowControlMask_ViewSpaceDepth.w,
            uSunlightShadowMap, uSunlightShadowTranslucencyMap,
            1.0);
    #endif

    return shadowAttenuation;
}

vec3 SunLightDiffuseContribution(vec3 albedo, vec3 normal, vec2 dudv, vec3 viewVector, float fresnel, float shadowAttenuation)
{
    vec3 diffuseContribution = vec3(0.0);
#if defined(SUNLIGHT_DIRECT_LIGHTING)
    diffuseContribution = albedo * (dot(normal, uInvSunDirection) * 0.5 + 0.5) * (1.0 - fresnel) * uSunColour;
#else
    // According to the shader source prior to refactoring, this path should only really be used for the minimap and the * 2 is because we don't have any lighting
    const float constantDiffuseAttentuation = 2.0;
    diffuseContribution = albedo * constantDiffuseAttentuation;
#endif

    return diffuseContribution * shadowAttenuation;
}

vec3 SunLightSpecularContribution(vec3 albedo, vec3 normal, vec2 dudv, vec3 viewVector, float fresnel, float shadowAttenuation)
{
    vec3 specularContribution = vec3(0.0);
#if defined(SPECULAR_LIGHTING)
    vec3 H = normalize(uInvSunDirection - viewVector);
    float dotSpec = clamp(dot(H, normal), 0.0, 1.0);

    // TODO: Sanitise this and make these magic numbers consts.
    // Why sun colour squared here but not for diffuse??
    specularContribution = (uSunColour * uSunColour) *
            fresnel * clamp(uInvSunDirection.y, 0.0, 1.0)
            * pow(dotSpec, uWaterNormalBRDFParams.z * 0.25) * (uWaterNormalBRDFParams.w * 1.8 + 0.2)
            * clamp(uWaterNormalBRDFParams.w - 0.05, 0.0, 1.0) * 25.0;
#endif

    return specularContribution * shadowAttenuation;
}


#if defined(GLOBAL_ENVIRONMENTMAPPING)
    vec3 GetEnvironmentMapReflection(vec3 viewVector, vec3 normal)
    {
        vec3 reflectedVector = reflect(viewVector, normal);
        reflectedVector.z = -reflectedVector.z;
        reflectedVector.y = abs(reflectedVector.y);
        vec3 reflection = textureCubeSRGB(uGlobalEnvironmentMap, reflectedVector).rgb;//uWaterReflectionStrength + vColour * max(0.0, 1.0 - uWaterReflectionStrength);
    #if defined(FOG_DISTANCE)
        float fogAmount = FogBasedOnAngle(normalize(reflect(viewVector, normal)));
        reflection = mix(reflection.rgb, uFogColour.rgb, fogAmount);
    #endif
        return reflection;
    }
#endif

/**
 * Gets then env map contribution
 */
vec3 EnvMapContribution(vec3 albedo, vec3 normal, vec3 viewVector, float fresnel)
{
    vec3 envMapContribution = vec3(0.0);
    float strength = 1.0;

#if defined(GLOBAL_ENVIRONMENTMAPPING)
#if defined(REFLECTION)
    strength = uWaterReflectionStrength;
    if (uWaterReflectionMapContribution < 1.0)
#endif
    {
        vec3 reflection = GetEnvironmentMapReflection(viewVector, normal);
    #if defined(REFLECTION)
        reflection *= 1.0 - uWaterReflectionMapContribution;
    #endif
        envMapContribution = reflection * mix(albedo, vec3(1.0, 1.0, 1.0), fresnel);
    }
#endif

    return envMapContribution * strength;
}

/**
 * Returns then planar reflection on the water's surface
 */
vec3 WaterReflection(vec3 albedo, vec3 normal, vec2 dudv, vec3 viewVector, float fresnel)
{
    vec3 reflection = vec3(0.0);

#if defined(REFLECTION)
    // TOFIX: Something in the planar reflections is introducing fringing.
    //        Suspect view/proj transforms during reflection pass not 100% complementary to the ones used during the forward pass.

    vec2 viewportOffsetPixels = uViewportOffsetScale.xy * uFullScreenLookupScale.zw;
    vec2 reflectionSamplePos = gl_FragCoord.xy + dudv;

    // Convert from fullscreen coords to viewport coords in a 0..1 range (needed to support multiviewport properly, otherwise we could just multiply by uViewportLookupScale.xy)
    reflectionSamplePos = (reflectionSamplePos - viewportOffsetPixels) * uViewportLookupScale.xy;

    vec4 planarReflection = texture2DLodCompat(uReflectionMap, vec2(1.0 - reflectionSamplePos.x, reflectionSamplePos.y), 0.0);
    // If we hit reduced alpha (ie any bilinear filtering towards alpha 0), take another reflection sample but without any perturbation
    if (planarReflection.a < 1.0)
    {
        reflectionSamplePos = gl_FragCoord.xy;

        // Convert from fullscreen coords to viewport coords in a 0..1 range (needed to support multiviewport properly, otherwise we could just multiply by uViewportLookupScale.xy)
        reflectionSamplePos = (reflectionSamplePos - viewportOffsetPixels) * uViewportLookupScale.xy;

        planarReflection = texture2DLodCompat(uReflectionMap, vec2(1.0 - reflectionSamplePos.x, reflectionSamplePos.y), 0.0);
        // If we hit zero alpha on the unperturbed sample, assume reflections are intentionally culled
        // (e.g. different height) and reflect from the environment map instead if available
    #if defined(GLOBAL_ENVIRONMENTMAPPING)
        if (planarReflection.a == 0.0)
        {
            planarReflection.rgb = GetEnvironmentMapReflection(viewVector, normal);
            planarReflection.a = 1.0;
        }
    #endif
    }
    reflection = planarReflection.rgb * mix(albedo, vec3(1.0), fresnel) * uWaterReflectionMapContribution * uWaterReflectionStrength;
#endif

    return reflection;
}

/**
 * Computes then water colour at then refracted point based upon extinction.
 */

vec3 WaterExtinction(vec3 albedo, vec3 refractionColour, float refractionDepth, vec2 refractionSamplePos, vec3 sunlightDiffuse, float edgeAlpha)
{
    vec3 extinctionColour = refractionColour;

#if defined(WATER_EXTINCTION) && defined(REFRACTION)

    // Tile size is 512 units, and each tile is apparently 1m^3 so one unit = 0.2cm.
    // To convert to metres scale by 0.002. Working in metres puts numbers into a more practical/sensible range for the artists.
    const float metresPerUnit = 0.002;

    if (uWaterFeatureFlags.y >= 1.0)
    {
        // Find the world space point in the scene we're looking at through the water's surface
        vec3 refractedSceneWorldSpacePos = GetWorldSpacePos(refractionSamplePos, refractionDepth, uSceneInvViewProjMatrix).xyz;

        // Find the amount of water the viewer is looking through, in metres.
        float accumulatedWater = length(refractedSceneWorldSpacePos - vPosition_WaterDepth.xyz) * metresPerUnit;

        // Find the depth of the water at the refracted scene position.
        highp float waterDepthAtRefractedPoint = abs(refractedSceneWorldSpacePos.y - vPosition_WaterDepth.y) * metresPerUnit;

        // A linear visibility falloff isn't a good one-size-fits-all approach.
        // At lower visibility lengths, the transition is too hard.
        // Similarly, a very soft falloff at longer visibility lengths doesn't appear hard enough.
        // Scale the falloff accordingly over 1m.
        highp float visibilityFalloffExponent = mix(0.04, 1.0, clamp(uWaterExtinctionVisibilityMetres, 0.0, 1.0));

        // The albedo comes from the overlay and foam.
        // This can interfere with extinction quite heavily.
        // Normalising the colour prevents excessive darkening of the water.
        vec3 rawSunlightDiffuseTerm = sunlightDiffuse / albedo;
        vec3 normalisedAlbedo = albedo / max(max(albedo.r, albedo.g), albedo.b);
        vec3 sunLight = rawSunlightDiffuseTerm * normalisedAlbedo;

        vec3 opaqueLitWaterColour = uWaterExtinctionOpaqueWaterColour * sunLight;
        float linearVisibility = clamp(accumulatedWater / uWaterExtinctionVisibilityMetres, 0.0, 1.0);
        float waterColourInterpolant = pow(linearVisibility, visibilityFalloffExponent);
        vec3 waterColour = mix(refractionColour, opaqueLitWaterColour, waterColourInterpolant);

        // Apply vertical extinction - Not using "bigDepthColour" because it can introduce undesirable colouring inconsistencies.
        const float extinctionFalloffExponent = 0.25;
        vec3 linearExtinctionInterpolant = clamp(waterDepthAtRefractedPoint / (uWaterExtinctionRGBDepthMetres * normalisedAlbedo), 0.0, 1.0);
        vec3 extinctionInterpolant = pow(linearExtinctionInterpolant, vec3(extinctionFalloffExponent));
        extinctionColour = waterColour * (1.0 - extinctionInterpolant);

        // Finally apply the edge fade to extinction for consistent behaviour.
        extinctionColour = mix(refractionColour, extinctionColour, edgeAlpha);
    }

#endif

    return extinctionColour;
}

// Returns refraction
vec3 WaterRefractionWithExtinction(vec3 albedo, vec3 normal, float waterDepth, float fresnel, vec3 sunlightDiffuse, float edgeAlpha)
{
    vec3 refraction = vec3(0.0);

#if defined(REFRACTION)

    vec2 viewportOffsetPixels = uViewportOffsetScale.xy * uFullScreenLookupScale.zw;

    vec3 refractionColour = vec3(0.0);
    float refractionDepth = 0.0;

    // Taking the sqrt of depth here produces a better looking curve, and using a higher max dudv than reflection makes the effect more pronounced
    highp vec2 dudvRefraction = normal.xz * min(sqrt(waterDepth) * edgeAlpha * 2.0, 128.0);
    vec2 refractionSamplePos = (gl_FragCoord.xy + dudvRefraction * fresnel * 2.0);

    // Convert from fullscreen coords to viewport coords in a 0..1 range (needed to support multiviewport properly, otherwise we could just multiply by uViewportLookupScale.xy)
    refractionSamplePos = (refractionSamplePos - viewportOffsetPixels) * uViewportLookupScale.xy;

    if (uWaterFeatureFlags.x >= 1.0)
    {
        // Check that the depth in the refraction map is below the water surface and there is something rendered at that point (depth != 1.0), otherwise take an unperturbed refraction sample
    #if !defined(NXT_MOBILE)
        refractionDepth = texture2DLodCompat(uRefractionDepth, refractionSamplePos, 0.0).r;
        if (refractionDepth < gl_FragCoord.z || refractionDepth >= 1.0)
        {
            refractionSamplePos = gl_FragCoord.xy;

            // Convert from fullscreen coords to viewport coords in a 0..1 range (needed to support multiviewport properly, otherwise we could just multiply by uViewportLookupScale.xy)
            refractionSamplePos = (refractionSamplePos - viewportOffsetPixels) * uViewportLookupScale.xy;
        }
    #endif

        refractionColour = texture2DLodCompat(uRefractionMap, vec3(refractionSamplePos, 0.0), 0.0).rgb;
    }

    refraction = WaterExtinction(albedo, refractionColour, refractionDepth, refractionSamplePos, sunlightDiffuse, edgeAlpha);
#endif

    return refraction;
}

#if defined(WATER_EMISSIVE) && !defined(GLES2_COMPAT_MODE) // Emissive only available outside of GLES2 compat mode.

    // The values of these #defines must match the corresponding value in the enum WaterParemters::EmissiveMapMode.
    #define EMISSIVE_MAP_MASK (0)
    #define EMISSIVE_MAP_RGBMAP_MASK (1)

    /**
    * Samples then emissive map. As with normal map sampling, the triangles can be subject to
    * up to three flow vectors. Where this is the case, we will sample the emissive map for
    * each and blend the result.
    */
    vec4 SampleEmissiveMap(sampler2D map, vec2 offset)
    {
       vec2 flowDelta01 = vUVPack_NormalMapMacroUV_EmissiveUV[1].zw - vUVPack_NormalMapMacroUV_EmissiveUV[0].zw;
       vec2 flowDelta02 = vUVPack_NormalMapMacroUV_EmissiveUV[2].zw - vUVPack_NormalMapMacroUV_EmissiveUV[0].zw;

       const float threshold = 1.e-7;
       if(dot(flowDelta01, flowDelta01) <= threshold && dot(flowDelta02, flowDelta02) <= threshold)
       {
           // No variation in flow, we only need to consider one flow vector.
           return texture2D(map, vUVPack_NormalMapMacroUV_EmissiveUV[0].zw + offset);
       }
       else
       {
           vec4 s = texture2D(map, vUVPack_NormalMapMacroUV_EmissiveUV[0].zw + offset) * vFlowControlMask_ViewSpaceDepth.x;
           s += texture2D(map, vUVPack_NormalMapMacroUV_EmissiveUV[1].zw + offset) * vFlowControlMask_ViewSpaceDepth.y;
           s += texture2D(map, vUVPack_NormalMapMacroUV_EmissiveUV[2].zw + offset) * vFlowControlMask_ViewSpaceDepth.z;
           return s;
       }
    }

    /**
     * An emissive implementation which operates as close to RT7 as possible utilising a single channel mask
     */
    vec4 EmissiveContribution_Mask(vec3 albedo, vec2 uvOffset)
    {
        float emissiveMask = SampleEmissiveMap(uWaterEmissiveMapTexture, uvOffset).r;

        // RT7 style emissive where source selected from albedo or uniform, and resulting emissiveColour = selectedColour * (1 + scale)
        vec3 emissiveColour = mix(albedo, uWaterEmissiveColourEmissiveSource.rgb, uWaterEmissiveColourEmissiveSource.a);

                float emissiveScale = uEmissiveScale_MapRefractionDepth_EmissiveMapMode_EmissiveMapExists.x;
        return vec4(emissiveColour * (1.0 + emissiveScale), emissiveMask);
    }

    /**
     * An emissive implementation which uses a colour emissive map with the mask in the alpha channel
     */
    vec4 EmissiveContribution_RGBMapMask(vec3 albedo, vec2 uvOffset)
    {
        vec4 emissiveMapSample = SampleEmissiveMap(uWaterEmissiveMapTexture, uvOffset);

        // RT7 style emissive where source selected from albedo or uniform, and resulting emissiveColour = selectedColour * (1 + scale)
        vec3 emissiveColour = mix(albedo, emissiveMapSample.rgb * uWaterEmissiveColourEmissiveSource.rgb, uWaterEmissiveColourEmissiveSource.a);

        float emissiveScale = uEmissiveScale_MapRefractionDepth_EmissiveMapMode_EmissiveMapExists.x;
        return vec4(emissiveColour * (1.0 + emissiveScale), emissiveMapSample.a);
    }

    /**
     * An emissive implementation which uses a solid colour without a map/mask.
     */
    vec4 EmissiveContribution_NoMap(vec3 albedo)
    {
        // RT7 style emissive where source selected from albedo or uniform, and resulting emissiveColour = selectedColour * (1 + scale)
        vec3 emissiveColour = mix(albedo, uWaterEmissiveColourEmissiveSource.rgb, uWaterEmissiveColourEmissiveSource.a);

        float emissiveScale = uEmissiveScale_MapRefractionDepth_EmissiveMapMode_EmissiveMapExists.x;
        return vec4(emissiveColour * (1.0 + emissiveScale), 1.0);
    }

    /**
     * Returns the emissive colour scaled for intensity in RGB and the emissive mask in A.
     * Water supports three emissive modes.
     * 1. EMISSIVE_MAP_MASK
     *    In this mode behaviour closely follows the RT7 implementation where a monochrome
     *    mask is used in conjection with either the albedo or a uniform to supply the emissive
     *    colour matching the RT7 emissive implementation.
     *
     * 2. EMISSIVE_MAP_RGBMAP_MASK
     *    In this mode a colour emissive map is supported and multiplied with the emissive colour uniform.
     *    The resulting colour can then be mixed with albedo, as per the RT7 implementation of emissive but
     *    with the added flexibility of an explicit emissive colour map.
     *    The mask exists in the alpha channel, and is used in exactly the same way as in the RT7 implementation.
     *
     * 3. When no map/mask texture is available a solid emissive colour is used.
     *
     * In addition to the two modes which make use of the map/mask texture, the emissive water also supports refraction of the emissive map.
     * This allows for a little more movement and variation where required.
     */
    vec4 EmissiveContribution(vec3 albedo, vec3 normal, float fresnel)
    {
        int emissiveMaskExists = int(uEmissiveScale_MapRefractionDepth_EmissiveMapMode_EmissiveMapExists.w);
        if(emissiveMaskExists == 0)
        {
            // No map/mask so use basic solid colour implementation.
            return EmissiveContribution_NoMap(albedo);
        }

        // Apply some basic refraction to the emissive mask - Using similar formula as in WaterRefraction.
        vec2 uvOffset = normal.xz * uEmissiveScale_MapRefractionDepth_EmissiveMapMode_EmissiveMapExists.y * fresnel * 2.0;

        int emissiveMapMode = int(uEmissiveScale_MapRefractionDepth_EmissiveMapMode_EmissiveMapExists.z);
        if(emissiveMapMode == EMISSIVE_MAP_MASK)
        {
            return EmissiveContribution_Mask(albedo, uvOffset);
        }
        else //emissiveMapMode == EMISSIVE_MAP_RGBMAP_MASK
        {
            return EmissiveContribution_RGBMapMask(albedo, uvOffset);
        }
    }

#endif

/********************************************************************************
* Water Debugging
*********************************************************************************/

#if (defined(SUNLIGHT_SHADOWS) && defined(DEBUG_SUNLIGHT_SHADOW_CASCADE) && !defined(DEFERRED_SHADOWS)) || defined(DEBUG_ALBEDO) || defined(DEBUG_NORMALS)
    #define WATER_DEBUG_OUTPUT
#endif

#if defined(WATER_DEBUG_OUTPUT)
    void WaterWriteDebugColour(vec3 albedo, vec3 normal)
    {
    #if defined(DEBUG_ALBEDO)
        gl_FragColor = vec4(albedo, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    #if defined(DEBUG_NORMALS)
        gl_FragColor = vec4((normal * 0.5 + 0.5), 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif

    // Debug shadows
    #if defined(SUNLIGHT_SHADOWS) && defined(DEBUG_SUNLIGHT_SHADOW_CASCADE) && !defined(DEFERRED_SHADOWS)
        gl_FragColor = vec4(ShadowMapCascadeColour(iCascade, int(uMappingParams.w)).rgb, 1.0);

        if (uDebugReturn != 0.0)
        {
            return;
        }
    #endif
    }
#endif



/********************************************************************************
* Water implementation
*********************************************************************************/

void WaterFragment()
{
    // View info
    vec3 viewVectorUnnormalised = vPosition_WaterDepth.xyz - uCameraPosition;
    float viewDistance = length(viewVectorUnnormalised);
    vec3 viewVector = viewVectorUnnormalised / viewDistance;

    // A triangle can be influenced by more than one flow vector, find the flow vector at this fragment.
    vec2 flow = GetCombinedFlow();
    float flowSpeed = length(flow);

    // Get noise for foam. Also used by legacy normals, if we switched to those.
    float noise = GenerateNoise();

    // Get water albedo - this varies depending on if foam is in use.
    vec3 albedo = WaterAlbedo(vPosition_WaterDepth.w, flowSpeed, noise);

    // Get water normal.
    vec3 normal = WaterNormal();
    vec2 dudv = normal.xz * min(vPosition_WaterDepth.w, 32.0);

    // Water has soft edges where it meets land.
    // This does not, and should not, apply to items in (on) the water like boats.
    float edgeAlpha = WaterSoftEdgeAlpha(vPosition_WaterDepth.w);

    // This acts as accumulator for the final water colour.
    vec4 fragColour = vec4(0.0);
    fragColour.a = edgeAlpha;

    // Fresnel
    const float R0 = 0.28;
    const float minFresnel = 0.0;
    const float maxFresnel = 0.6;
    float fresnel = clamp(FresnelSchlick(normal, -viewVector, R0), minFresnel, maxFresnel);
    float biasedFresnel = clamp(fresnel + uWaterNormalBRDFParams.x, minFresnel, maxFresnel);
    fragColour.a *= fresnel;

    // Reflection
    fragColour.rgb += WaterReflection(albedo, normal, dudv, viewVector, biasedFresnel);

    // Env map
    fragColour.rgb += EnvMapContribution(albedo, normal, viewVector, biasedFresnel);

    // Lighting
    float sunlightShadowAttenuation = SunLightShadowAttenuation(albedo, normal, dudv, viewVector, fresnel);
    vec3 sunlightDiffuse = SunLightDiffuseContribution(albedo, normal, dudv, viewVector, fresnel, sunlightShadowAttenuation);
    vec3 sunlightSpecular = SunLightSpecularContribution(albedo, normal, dudv, viewVector, biasedFresnel, sunlightShadowAttenuation);

    fragColour.rgb += sunlightSpecular;

    //Light scattering (now includes fog!)
#if defined(SUNLIGHT_DIRECT_LIGHTING) && (defined(LIGHT_SCATTERING) || defined(FOG_DISTANCE))
    vec3 outScattering = vOutScattering;
    vec3 inScattering = vInScattering;

    fragColour.rgb = ApplyInOutScattering(fragColour.rgb, outScattering, inScattering);
    fragColour.a = mix(fragColour.a, 1.0, vColour.a);
#endif

    // Refraction and lighting are intertwined because modelling polutants/opaque fluids requires that we consider
    // the diffuse lighting contribution to the particulates poluting the water.
#if !defined(REFRACTION)
    // If we have refraction enabled, the diffuse lighting contribution is applied as part of the water
    // extinction modelling opaque/poluted fluids. In this code path we're not using refraction and extinction
    // so apply diffuse directly.
    fragColour.rgb += sunlightDiffuse;

    #if defined(WATER_EMISSIVE) && !defined(GLES2_COMPAT_MODE)
        vec4 emissiveContribution = EmissiveContribution(albedo, normal, fresnel);

        // RT7 style blend - lerp between existing frag colour and emissive colour.
        // The uEmissiveBlend uniform provides control over this blending.
        fragColour.rgb = mix(fragColour.rgb, emissiveContribution.rgb, emissiveContribution.a * edgeAlpha * uEmissiveBlend);
    #endif

#else
    vec3 refractionColour = WaterRefractionWithExtinction(albedo, normal, vPosition_WaterDepth.w, fresnel, sunlightDiffuse, edgeAlpha);

    // If we have emissive enabled we need to blend this in prior to blending in the refraction colour.
    // We also need to serparate the edge alpha and fresnel factors so fresnel doesn't impact the emissive blend.
    // We could avoid these extra steps if emissive was additive, but this would mean using a model which differs from RT7.
    #if defined(WATER_EMISSIVE) && !defined(GLES2_COMPAT_MODE)
        // The colour visible at this point on the waters surface without considering edge fade.
        fragColour.rgb = mix(refractionColour, fragColour.rgb, fresnel);

        // RT7 style blend - lerp between existing frag colour and emissive colour.
        // The uEmissiveBlend uniform provides control over this blending.
        vec4 emissiveContribution = EmissiveContribution(albedo, normal, fresnel);
        fragColour.rgb = mix(fragColour.rgb, emissiveContribution.rgb, emissiveContribution.a * uEmissiveBlend);

        // Finally apply edge fade
        fragColour.rgb = mix(refractionColour, fragColour.rgb, edgeAlpha);
    #else
        // Not using emissive so can apply edge fade and fresnel in one pass.
        fragColour.rgb = mix(refractionColour, fragColour.rgb, fragColour.a);
    #endif

#endif

// If not using direct sun lighting, have a fixed alpha
// According to the shader prior to refactoring, this path should only really be used for the minimap
#if !defined(SUNLIGHT_DIRECT_LIGHTING)
    fragColour.a = 1.0;
#endif


    gl_FragColor = fragColour;

#if defined(WATER_DEBUG_OUTPUT)
    WaterWriteDebugColour(albedo, normal);
#endif
}

/*********************************************************************************
* Water caustics
*********************************************************************************/

#if defined(CAUSTICS_STENCIL)
    void CausticsStencil()
    {
        gl_FragColor.r = float(texture2D(uCausticsMap, gl_FragCoord.xy / uCausticsMapSize).r) * uCausticsScale / CAUSTICS_FIXED_POINT_SCALE;
    }
#endif

#if defined(CAUSTICS_COMPUTE)
    void CausticsCompute()
    {
        // Sample normal maps and produce final normal for the water
        vec3 normal = WaterNormal();

        WriteCausticsRay(normal, vPosition_WaterDepth.w);
        //[
        discard;
        return;
        //]
    }
#endif

/*********************************************************************************
* Culling
*********************************************************************************/

#if defined(CLIP_PLANE_CLAMP)
    void ClipPlaneClamp()
    {
        // Calibrated against 0,39,54 waterfalls
        const float waterPatchReflectionPlaneEpsilon = 200.0;
        // Cull water patches that are on a different height level so they don't end up with broken reflections
        if (abs(vPosition_WaterDepth.y + uClipPlane.w) < waterPatchReflectionPlaneEpsilon)
        {
            gl_FragDepth = 1.0;
            gl_FragColor = vec4(uFogColour.rgb, 1.0);
        }
        else
        {
            discard;
        }
    }
#endif

/*********************************************************************************
* shader entry point
*********************************************************************************/

void main()
{
#if defined(CLIP_PLANE_CLAMP)
    ClipPlaneClamp();
#elif defined(CAUSTICS_STENCIL)
    CausticsStencil();
#elif defined(CAUSTICS_COMPUTE)
    CausticsCompute();
#else
    WaterFragment();
#endif
}
