
attribute vec4 aWaterPosition_Depth;
attribute vec4 aVertexColour;
attribute vec2 aWaterFlowDataPatchFlow0;
attribute vec2 aWaterFlowDataPatchFlow1;
attribute vec2 aWaterFlowDataPatchFlow2;
attribute vec4 aWaterFlowDataFlowNoise0_FlowNoise1;
attribute vec4 aWaterFlowDataFlowNoise2_FlowIndex_Spare;

#include "Shared/DistanceFog/Uniforms.inc"
#include "Shared/DistanceFog/Functions.inc"
#include "Shared/VolumetricScattering/VolumetricFunctions.inc"
#include "Shared/VolumetricScattering/ApplyVolumetrics.inc"
#include "UndefineInOut.inc"

/**
 * Produces texture coordinates that are vertically aligned to the given flow vector and applies flow.
 */
vec2 GetFlowOrientedUV(vec2 staticUV, vec2 flowVector, float flowSpeed, float time)
{
    // We're going to project into the local coordinate frame of the flow
    // find flow foward and flow right vectors
    // If we get a zero magnitude flow vector we'll default back to axis aligned.
    float flowMagnitudeSqr = dot(flowVector, flowVector);

    const float epsilon = 1.e-6;
    vec2 flowForward = flowMagnitudeSqr < epsilon ? vec2(0.0, 1.0) : normalize(flowVector);
    vec2 flowRight = vec2(flowForward.y, -flowForward.x);

    // Find the flow delta to advance the texture coords
    highp float flowDelta = length(flowVector * flowSpeed * time);

    // Produce final texture coord
    return vec2(dot(flowRight, staticUV), dot(flowForward, staticUV) + flowDelta);
}

/**
 * Produces texture coordinates that are vertically aligned to the given flow vector and applies flow.
 */
vec2 GetFlowOrientedUV(vec4 worldPosition, vec2 flowVector, float flowSpeed, vec2 uvScale, float time)
{
    vec2 staticUV = vec2(worldPosition.xz);
    float flowSign = sign(uvScale.y);
    return GetFlowOrientedUV(staticUV, flowVector * flowSign, flowSpeed, time) * uvScale;
}

#if defined(WATER_NORMAL_MAPS)

    /**
     * Transforms a static water patch texture coordinate according to the given normal map params and applies flow.
     * The resulting texture coordinates are aligned to the flow of the water prior to UV rotation
     */
    vec2 TransformNormalMapTexCoordFlowAligned(in NormalMapTexCoordParams params, vec2 staticUV, vec2 flow, vec2 flowNoise)
    {
        const float timeScale = 0.5;
        highp float time = uWaterTickFade.x * timeScale;

        float flowSpeed = params.flowSpeed_spareyzw.x;
        vec2 orientedUV = GetFlowOrientedUV(staticUV, params.flowRotation * flow, flowSpeed, time);
        vec2 uv = (params.uvRotation * orientedUV + flowNoise * flowSpeed * time) * params.uvScale_uvOffset.xy + params.uvScale_uvOffset.zw;

        return uv;
    }

    /**
     * Transforms a static water patch texture coordinate according to the given normal map params and applies flow.
     * The resulting texture coordinates are aligned axis aligned prior to UV rotation
     */
    vec2 TransformNormalMapTexCoordAxisAligned(in NormalMapTexCoordParams params, vec2 staticUV, vec2 flow, vec2 flowNoise)
    {
        const float timeScale = 0.5;
        highp float time = uWaterTickFade.x * timeScale;

        vec2 flowVelocity = (flow + flowNoise) * params.flowSpeed_spareyzw.x;
        vec2 uvFlowDelta = params.flowRotation * flowVelocity * time;
        vec2 uv = (params.uvRotation * (staticUV + uvFlowDelta)) * params.uvScale_uvOffset.xy + params.uvScale_uvOffset.zw;

        return uv;
    }

    #if !defined(GLES2_COMPAT_MODE)
        /**
         * Computes the normal map detail texture coordinates for each normal map using each of the flow vectors.
         */
        void ComputeNormalMapDetailTexCoords(
            vec2 staticUV0,
            vec2 staticUV1,
            vec2 staticUV2,
            vec2 flow0,
            vec2 flow1,
            vec2 flow2,
            vec2 flowNoise0,
            vec2 flowNoise1,
            vec2 flowNoise2)
        {
            // Apply flow noise differently to each normal map layer.
            // This adds variation and prevents the appearance of chunks sliding over each other.
            vec2 map0FlowNoiseWeight = vec2(1.0);
            vec2 map1FlowNoiseWeight = vec2(-1.0);
            vec2 map2FlowNoiseWeight = vec2(1.0, -1.0);

            // Compute each normal map's flowing texture coords from the map's static tex coord while applying the sampling transforms.
            vec2 uvFlow0Map0 = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[0], staticUV0, flow0, flowNoise0 * map0FlowNoiseWeight);
            vec2 uvFlow0Map1 = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[1], staticUV1, flow0, flowNoise0 * map1FlowNoiseWeight);
            vec2 uvFlow0Map2 = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[2], staticUV2, flow0, flowNoise0 * map2FlowNoiseWeight);

            vec2 uvFlow1Map0 = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[0], staticUV0, flow1, flowNoise1 * map0FlowNoiseWeight);
            vec2 uvFlow1Map1 = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[1], staticUV1, flow1, flowNoise1 * map1FlowNoiseWeight);
            vec2 uvFlow1Map2 = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[2], staticUV2, flow1, flowNoise1 * map2FlowNoiseWeight);

            vec2 uvFlow2Map0 = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[0], staticUV0, flow2, flowNoise2 * map0FlowNoiseWeight);
            vec2 uvFlow2Map1 = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[1], staticUV1, flow2, flowNoise2 * map1FlowNoiseWeight);
            vec2 uvFlow2Map2 = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[2], staticUV2, flow2, flowNoise2 * map2FlowNoiseWeight);

            // Tightly pack the results (because intel)
            vNoisyPatchFlow2_UVPack_NormalsFlow0Map0.zw = uvFlow0Map0;
            vUVPack_NormalsFlow0Map1_NormalsFlow0Map2 = vec4(uvFlow0Map1, uvFlow0Map2);
            vUVPack_NormalsFlow1Map0_NormalsFlow1Map1 = vec4(uvFlow1Map0, uvFlow1Map1);
            vUVPack_NormalsFlow1Map2_NormalsFlow2Map0 = vec4(uvFlow1Map2, uvFlow2Map0);
            vUVPack_NormalsFlow2Map1_NormalsFlow2Map2 = vec4(uvFlow2Map1, uvFlow2Map2);
        }

        /**
         * Computes texture coordinates for the "macro" normal map sampling.
         */
        void ComputeNormalMapMacroTexCoords(vec2 staticUV0, vec2 staticUV1, vec2 staticUV2)
        {
            const float macroDefaultFlowSpeed = 0.25;
            const vec2 macroDefaultFlow = vec2(0.1, -0.13) * macroDefaultFlowSpeed;
            const float macroUVBaseScale = 0.1; // default scale applied to macro samples - in line with original implementation.

            vUVPack_NormalMapMacroUV_EmissiveUV[0].xy = TransformNormalMapTexCoordAxisAligned(uWaterNormalMapMacroTexCoordParams[0], staticUV0 * macroUVBaseScale, macroDefaultFlow, vec2(0.0));
            vUVPack_NormalMapMacroUV_EmissiveUV[1].xy = TransformNormalMapTexCoordAxisAligned(uWaterNormalMapMacroTexCoordParams[1], staticUV1 * macroUVBaseScale, macroDefaultFlow, vec2(0.0));
            vUVPack_NormalMapMacroUV_EmissiveUV[2].xy = TransformNormalMapTexCoordAxisAligned(uWaterNormalMapMacroTexCoordParams[2], staticUV2 * macroUVBaseScale, macroDefaultFlow, vec2(0.0));
        }

    #else

        /**
         * Computes the normal map detail texture coordinates for each normal map using each of the flow vectors.
         */
        void ComputeNormalMapDetailTexCoords(
            vec2 staticUV0,
            vec2 flow0,
            vec2 flow1,
            vec2 flow2,
            vec2 flowNoise0,
            vec2 flowNoise1,
            vec2 flowNoise2)
        {
            // Under GLES compatibility mode we only support one set of UVs.
            vec2 normalMapDetailFlow0UV = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[0], staticUV0, flow0, flowNoise0);
            vec2 normalMapDetailFlow1UV = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[0], staticUV0, flow1, flowNoise1);
            vec2 normalMapDetailFlow2UV = TransformNormalMapTexCoordFlowAligned(uWaterNormalMapTexCoordParams[0], staticUV0, flow2, flowNoise2);

            // Tightly pack the results
            vNoisyPatchFlow2_UVPack_NormalsFlow0Map0.zw = normalMapDetailFlow0UV;
            vUVPack_NormalsFlow1Map0_NormalsFlow2Map0 = vec4(normalMapDetailFlow1UV, normalMapDetailFlow2UV);
        }

    #endif


#endif

#if defined(WATER_EMISSIVE) && !defined(GLES2_COMPAT_MODE)
    // The values of these #defines must match the corresponding value in the enum WaterParemters::EmissiveUVMode.
    #define EMISSIVE_UV_MODE_AXIS_ALIGNED (0)
    #define EMISSIVE_UV_MODE_FLOW_ALIGNED (1)

    void GetEmissiveFlow(out vec2 flowOut0, out vec2 flowOut1, out vec2 flowOut2, vec2 patchFlow0, vec2 patchFlow1, vec2 patchFlow2, float emissiveFlowRotationTheta)
    {
        float c = cos(emissiveFlowRotationTheta);
        float s = sin(emissiveFlowRotationTheta);
        mat2 emissiveFlowRotationMatrix = mat2(vec2(c, -s), vec2(s, c));

        vec2 flow0 = emissiveFlowRotationMatrix * patchFlow0;
        vec2 flow1 = emissiveFlowRotationMatrix * patchFlow1;
        vec2 flow2 = emissiveFlowRotationMatrix * patchFlow2;

        // When a flow vector is zero length or near zero length, try matching an existing flow vector if any have some magnitude.
        // This is a work around for tile edges having zero flow due to there being no adjacency data at the point of terrian generation.
        // This isn't an ideal solution as it can introduce hard edges, but in tests this looked more acceptable than static emissive textures at tile edges.
        // A more correct solution to this problem would require terrain tiles to be able to provide adjacent water patch data when building the terrain.
        // We could then properly assess water patch flow at tile boundaries.
        const float epsilon = 1.e-6;

        // Pack flow vectors with their square magnitude for use in selection.
        vec3 flowData0 = vec3(flow0, dot(flow0, flow0));
        vec3 flowData1 = vec3(flow1, dot(flow1, flow1));
        vec3 flowData2 = vec3(flow2, dot(flow2, flow2));

        // If any are near zero magnitude, try finding a better flow vector
        if (flowData0.z < epsilon)
        {
            flowData0 = flowData1.z > flowData2.z ? flowData1 : flowData2;
        }

        if (flowData1.z < epsilon)
        {
            flowData1 = flowData2.z > flowData0.z ? flowData2 : flowData0;
        }

        if (flowData2.z < epsilon)
        {
            flowData2 = flowData0.z > flowData1.z ? flowData0 : flowData1;
        }

        // unpack the selected flow vectors
        flowOut0 = flowData0.xy;
        flowOut1 = flowData1.xy;
        flowOut2 = flowData2.xy;
    }

    void ComputeEmissiveTexCoords(vec4 worldPosition, vec2 patchFlow0, vec2 patchFlow1, vec2 patchFlow2)
    {
        // Unpack emissive flow control params
        float emissiveFlowSpeed = uEmissiveFlowSpeed_EmissiveFlowRotation_EmissiveUVScale.x;
        float emissiveFlowRotationTheta = uEmissiveFlowSpeed_EmissiveFlowRotation_EmissiveUVScale.y;
        float emissiveMapTextureScale = uEmissiveMapScale_EmissiveUVMode.x;
        vec2 emissiveMapUVSCale = uEmissiveFlowSpeed_EmissiveFlowRotation_EmissiveUVScale.zw;

        // GetEmissiveFlow will set these to something useful.
        vec2 flow0 = vec2(0.0, 0.0);
        vec2 flow1 = vec2(0.0, 0.0);
        vec2 flow2 = vec2(0.0, 0.0);
        GetEmissiveFlow(flow0, flow1, flow2, patchFlow0, patchFlow1, patchFlow2, emissiveFlowRotationTheta);

        const float timeScale = 0.5; //consistent with fragment shader
        highp float time = uWaterTickFade.x * timeScale;

        // To not break existing water, normal maps continue to support a texture scale despite the addition of independent UV scaling.
        // For consistent behaviour, emissive map sampling will use the same method.
        vec2 staticUV = worldPosition.xz * emissiveMapTextureScale;

        if(int(uEmissiveMapScale_EmissiveUVMode.y) == EMISSIVE_UV_MODE_FLOW_ALIGNED)
        {
            // compute the flow aligned UVs.
            vUVPack_NormalMapMacroUV_EmissiveUV[0].zw = GetFlowOrientedUV(staticUV, flow0, emissiveFlowSpeed, time) * emissiveMapUVSCale;
            vUVPack_NormalMapMacroUV_EmissiveUV[1].zw = GetFlowOrientedUV(staticUV, flow1, emissiveFlowSpeed, time) * emissiveMapUVSCale;
            vUVPack_NormalMapMacroUV_EmissiveUV[2].zw = GetFlowOrientedUV(staticUV, flow2, emissiveFlowSpeed, time) * emissiveMapUVSCale;
        }
        else
        {
            // compute axis aligned UVs
            vec2 uvScaleSign = sign(emissiveMapUVSCale);
            vUVPack_NormalMapMacroUV_EmissiveUV[0].zw = staticUV * emissiveMapUVSCale + flow0 * uvScaleSign * emissiveFlowSpeed * time;
            vUVPack_NormalMapMacroUV_EmissiveUV[1].zw = staticUV * emissiveMapUVSCale + flow1 * uvScaleSign * emissiveFlowSpeed * time;
            vUVPack_NormalMapMacroUV_EmissiveUV[2].zw = staticUV * emissiveMapUVSCale + flow2 * uvScaleSign * emissiveFlowSpeed * time;
        }
    }
#endif

void main()
{
    vec3 position = aWaterPosition_Depth.xyz;
    vec4 worldPosition = uModelMatrix * vec4(position, 1.0);

    vPosition_WaterDepth.xyz = worldPosition.xyz;
    vPosition_WaterDepth.w = aWaterPosition_Depth.w;
    #if defined(CLIP_PLANE_CLAMP) && defined(CLIP_PLANE)
        // For the reflection mask, ensure water y position is exactly the same as the clip plane (often not the case)
        worldPosition.y = -uClipPlane.w;
    #endif
    gl_Position = uViewProjMatrix * worldPosition;
    // Can skip everything else, reflection mask fragment shader only needs position
    //[
    #if defined(CLIP_PLANE_CLAMP) && defined(CLIP_PLANE)
        return;
    #endif
    //]

    #if defined(CAUSTICS_COMPUTE) || defined(CAUSTICS_STENCIL)
        // 32 is 1 MapEd terrain height unit, and multiply for some leeway
        const float waterPatchCausticsPlaneEpsilon = 32.0 * 8.0;
        // Cull water patches that are on a different height level so they don't end up with broken caustics
        if (abs(vPosition_WaterDepth.y - uCausticsPlaneHeight) > waterPatchCausticsPlaneEpsilon)
        {
            // Cull via z-clip and early out
            gl_Position.w = -1.0;
            //[
            return;
            //]
        }
    #endif

    vColour = aVertexColour;

    #if defined(LIGHT_SCATTERING) || defined(FOG_DISTANCE) || defined(SUNLIGHT_SHADOWS) || defined(VOLUMETRIC_SCATTERING_SUPPORTED)
        vec3 viewVector = worldPosition.xyz - uCameraPosition;
        float cameraDistance = length(viewVector);
    #endif

    //Shadow mapping
    #if defined(SUNLIGHT_SHADOWS)
        // Compute depth in camera space
        // Camera plane distance more accurate and must be used for cascade interval select to work without artifacts
        vec3 camLook = vec3(uViewMatrix[0][2], uViewMatrix[1][2], uViewMatrix[2][2]);
        vFlowControlMask_ViewSpaceDepth.w = abs(dot(viewVector.xyz, camLook));
    #endif

    #if defined(LIGHT_SCATTERING) || defined(FOG_DISTANCE)
        viewVector /= cameraDistance;

        #if defined(LIGHT_SCATTERING) && defined(SUNLIGHT_DIRECT_LIGHTING)
            ComputeInOutScattering(viewVector, cameraDistance, uInvSunDirection, vOutScattering, vInScattering);
        #else
            vOutScattering = vec3(1.0);
            vInScattering = vec3(0.0);
        #endif

        #if defined(FOG_DISTANCE)
            float fogFactor = FogBasedOnDistance(cameraDistance);
            fogFactor = fogFactor + (fogFactor * uWaterTickFade.y) - uWaterTickFade.y;
            vInScattering = mix(vInScattering, uFogColour.rgb, fogFactor);
            vOutScattering *= 1.0 - fogFactor;
            vColour.a = fogFactor; //Forces the water more opaque when it's fogging out
        #else
            vColour.a = 0.0;
        #endif
    #endif

    #if defined(VOLUMETRIC_SCATTERING_SUPPORTED)
        vec4 scattering = GetScatteredInRay2(8, viewVector, cameraDistance, 100.0); // Using a fixed value for the offset here gives more stability
        vec3 extraExtinction = vec3(1.0);
        vec3 extraInscatter = vec3(0.0);

        GetInAndOutScattering(scattering, extraExtinction, extraInscatter);

        vInScattering += extraInscatter;
        vOutScattering *= extraExtinction;
    #endif

    float flowNoiseLevel = uWaterNormalMapTextureScales_FlowNoiseScale.w;
    vec2 flowNoise0 = (aWaterFlowDataFlowNoise0_FlowNoise1.xy / 127.0) * flowNoiseLevel;
    vec2 flowNoise1 = (aWaterFlowDataFlowNoise0_FlowNoise1.zw / 127.0) * flowNoiseLevel;
    vec2 flowNoise2 = (aWaterFlowDataFlowNoise2_FlowIndex_Spare.xy / 127.0) * flowNoiseLevel;

    vec2 patchFlow0 = -aWaterFlowDataPatchFlow0 * (1.0 / 4095.0); //Refer to BuildTerrainWorker.cpp fixedPointValueForOne
    vec2 patchFlow1 = -aWaterFlowDataPatchFlow1 * (1.0 / 4095.0); //Refer to BuildTerrainWorker.cpp fixedPointValueForOne
    vec2 patchFlow2 = -aWaterFlowDataPatchFlow2 * (1.0 / 4095.0); //Refer to BuildTerrainWorker.cpp fixedPointValueForOne

    int flowIndex = int(aWaterFlowDataFlowNoise2_FlowIndex_Spare.z);
    vFlowControlMask_ViewSpaceDepth.xyz = vec3(flowIndex == 0 ? 1.0 : 0.0, flowIndex == 1 ? 1.0 : 0.0, flowIndex == 2 ? 1.0 : 0.0);

    vNoisyPatchFlow0_NoisyPatchFlow1.xy = patchFlow0 + flowNoise0;
    vNoisyPatchFlow0_NoisyPatchFlow1.zw = patchFlow1 + flowNoise1;
    vNoisyPatchFlow2_UVPack_NormalsFlow0Map0.xy = patchFlow2 + flowNoise2;

    #if defined(WATER_NORMAL_MAPS) || defined(WATER_FOAM_MAP)
        #if !defined(GLES2_COMPAT_MODE)
            vUVPack_FoamUV = worldPosition.xz * uWaterNormalMapTextureScales_FlowNoiseScale.x; // TOFIX: This was propagated from the old code but is wrong as it should use the foam materal scale.
        #endif

        vec2 staticNormalUV0 = worldPosition.xz * uWaterNormalMapTextureScales_FlowNoiseScale.x;
        vec2 staticNormalUV1 = worldPosition.xz * uWaterNormalMapTextureScales_FlowNoiseScale.y;
        vec2 staticNormalUV2 = worldPosition.xz * uWaterNormalMapTextureScales_FlowNoiseScale.z;

        #if defined(WATER_NORMAL_MAPS)
            #if !defined(GLES2_COMPAT_MODE)
                ComputeNormalMapDetailTexCoords(
                    staticNormalUV0,
                    staticNormalUV1,
                    staticNormalUV2,
                    patchFlow0,
                    patchFlow1,
                    patchFlow2,
                    flowNoise0,
                    flowNoise1,
                    flowNoise2);
                ComputeNormalMapMacroTexCoords(staticNormalUV0, staticNormalUV1, staticNormalUV2);

            #else

                ComputeNormalMapDetailTexCoords(
                    staticNormalUV0,
                    patchFlow0,
                    patchFlow1,
                    patchFlow2,
                    flowNoise0,
                    flowNoise1,
                    flowNoise2);
            #endif
        #endif
    #endif

    #if defined(WATER_EMISSIVE) && !defined(GLES2_COMPAT_MODE)
        ComputeEmissiveTexCoords(worldPosition, patchFlow0, patchFlow1, patchFlow2);
    #endif
}
