#ifndef LIGHTING_INC
#define LIGHTING_INC

#include "UndefineInOut.inc"

#include "lighting/brdf/Fresnel.inc"
#include "lighting/brdf/BRDF.inc"


struct LightingTerms
{
    vec3 Diffuse;
    vec3 Specular;
};

void ClearLightingTerms(inout LightingTerms lightingInOut)
{
    lightingInOut.Diffuse = vec3(0.0, 0.0, 0.0);
    lightingInOut.Specular = vec3(0.0, 0.0, 0.0);
}

void AddLightingTerms(inout LightingTerms lightingInOut, LightingTerms lightingIn)
{
    lightingInOut.Diffuse += lightingIn.Diffuse;
    lightingInOut.Specular += lightingIn.Specular;
}

/**
    @param diffuseOut The direct diffuse lighting contribution
    @param specularOut The direct specular lighting contribution
    @param normal Surface normal for pixel
    @param viewVec The vector from surface position to eye
    @param lightDir The vector from surface position to light
    @param specRefl The surface specular reflectance value
    @param specPow How shiny the surface is (i.e. specular power)
    @param lightAtten Typically n dot l combined with other attenuation factors (i.e. shadow attenuation). This values is not part of the BRDF!
 */
void EvaluateDirLightRT5(
    inout LightingTerms lightingTermsOut,
    vec3 normal,
    vec3 tangent,
    vec3 viewVec,
    vec3 lightDir,
    float specRefl,
    float specPow,
    float fresnelExp,
    float lightAtten,
    float nDotL,
    vec3 lightColour)
{
    // Direct lambert diffuse
    lightingTermsOut.Diffuse += (lightColour * lightAtten);

    // Direct Specular
#if defined(SPECULAR_LIGHTING)
    // Calculate half vector and fresnel term
    vec3 halfwayVec = normalize(viewVec + lightDir);
    // We need to clamp the dot product to protect against NaNs forming causing black patches to form
    float specFresnel = FresnelSchlick(specRefl, clamp(dot(lightDir, halfwayVec), 0.0, 1.0), fresnelExp); // More physically correct l dot h for punctual lights

#if defined(ANISOTROPY_BRDF)
    vec3 specLighting = AnisotropicBRDF(halfwayVec, normal, tangent, lightDir, viewVec, nDotL, specPow, 0.5, 0.5);
#else
    vec3 specLighting = vec3(specFresnel) * vec3(RunescapeRT5BRDF(halfwayVec, normal, specPow));
#endif
    specLighting *= (lightColour * lightAtten);

    lightingTermsOut.Specular += specLighting;
#endif
}

/**
    Use in conjunction with RT7 materials.
    Note. lightingTerms will still need to be modulated with light shadow attenuation after calling this function

    @param specPow Blinn-Phong specular power/shininess, range 0-8192
    @param perceptualRoughness Roughness from 0-1, needed for NDFs
 */
void EvaluateDirLightRT7(
    inout LightingTerms lightingTermsOut,
    vec3 normal,
    vec3 tangent,
    vec3 viewVec,
    vec3 lightDir,
    vec3 specRefl,
    float specPow,
    float perceptualRoughness,
    float fresnelExp,
    float lightAtten,
    float nDotL,
    vec3 lightColour)
{
    // Direct lambert diffuse
    lightingTermsOut.Diffuse += (lightColour * lightAtten);

    // Direct Specular
#if defined(SPECULAR_LIGHTING)
    // Calculate half vector and fresnel term
    vec3 halfwayVec = normalize(viewVec + lightDir);
    // We need to clamp the dot product to protect against NaNs forming causing black patches to form
    vec3 specFresnel = FresnelSchlick(specRefl, clamp(dot(lightDir, halfwayVec), 0.0, 1.0), fresnelExp); // More physically correct l dot h for punctual lights

#if defined(ANISOTROPY_BRDF)
    vec3 specLighting = AnisotropicBRDF(halfwayVec, normal, tangent, lightDir, viewVec, nDotL, specPow, 0.5, 0.5);
#else
    vec3 specLighting = CookTorranceBRDF(specPow, perceptualRoughness, specFresnel, normal, halfwayVec, viewVec, lightDir, nDotL);
#endif
    specLighting *= (lightColour * lightAtten);

    lightingTermsOut.Specular += specLighting;
#endif
}

/**
    A helper that attempts to mask specular lighting that bleeds through to parts of a surface
    whose reflection vectors around the normal actually ends up underneath the surface of the object
    and should ideally not exhibit specular lighting. The returned value can be applied to either
    or both direct and indirect specular lighting.

    See, https://marmosetco.tumblr.com/post/81245981087

    @param horizonScaleFactor A value between 0-2 to control the amount of specular horizon occlusion to apply
    @param viewDir The incident view vector in the direction from eye to surface
    @param normal The world space surface normal at the given pixel
 */
float SpecularHorizonOcclusion(float horizonScaleFactor, vec3 viewDir, vec3 normal)
{
    // When calculating the reflection vector, we need the incoming view vector directed towards the surface
    vec3 reflVec = reflect(viewDir, normal);
    float horzOcclusion = clamp(1.0 + horizonScaleFactor * dot(reflVec, normal), 0.0, 1.0);
    horzOcclusion *= horzOcclusion;
    return horzOcclusion;
}


#include "DefineInOut.inc"

#endif

