#ifndef BRDF_INC
#define BRDF_INC

#include "lighting/brdf/NDF.inc"
#include "lighting/brdf/VisFunc.inc"

// Normal distribution functions
//#define BLINN_PHONG_NDF
#define GGX_NDF
//#define BECKMANN_NDF

// Shadow and masking terms/functions
#define SCHLICK_SMITH_VIS
//#define KELEMEN_SZIRMAY_KALOS_VIS


/**
    A Cook-Torrance based BRDF based on the following equation. Various aspects of this equation (i.e. PI / 4.0 and (n dot lc)(n dot v))
    are accounted for by some of the NDF and visibility functions, which is why you don't see them being accounted for explicitly
    in all cases.

    Lo(v) =     D(h) G(lc, v, h) F(l, h)
             PI ------------------------
                    4(n dot l)(n dot v)

    @param specPow Blinn-Phong specular power/shininess, range 0-8192
    @param perceptualRoughness Roughness from 0-1, needed for NDFs
    @param specFresnel The fresnel specular reflectance term (fresnel already applied)

    @todo Can be optimized for mobile, but will require custom mobile version, see http://filmicworlds.com/blog/optimizing-ggx-shaders-with-dotlh/
 */
vec3 CookTorranceBRDF(
    float specPow,
    float perceptualRoughness,
    vec3 specFresnel,
    vec3 normal,
    vec3 halwayVec,
    vec3 viewVec,
    vec3 lightDir,
    float nDotL)
{
    float nDotH = max(0.0, dot(normal, halwayVec));

    // Calculate normal distribution term
    float D = 1.0;
#if defined(BLINN_PHONG_NDF)
    D = BlinnPhongNDF(specPow, nDotH); // * (PI / 4.0) is already accounted for by BlinnPhongNDF
#elif defined(GGX_NDF)
    D = GGXTrowbridgeReitzNDF(PerceptualRoughnessToRoughness(perceptualRoughness), nDotH)/* * PI_OVER_4*/;
#elif defined(BECKMANN_NDF)
    D = max(0.0f, BeckmannNDF(SpecPowToBeckmannRoughness(specPow), nDotH))/* * PI_OVER_4*/; // Beckmann needs clamping, otherwise black pathces will be seen
#else
    #error CookTorranceBRDF normal distribution function not specified
#endif

    // Calculate visibility/geometry term
    float G = 1.0; // G(lc, v, h) / (n dot l)(n dot v) == 1
#if defined(SCHLICK_SMITH_VIS)
    G = SchlickSmithVis(specPow, nDotL, max(0.0, dot(normal, viewVec))); // (n dot l)(n dot v) done in SchlickSmithVisFunc
#elif defined(KELEMEN_SZIRMAY_KALOS_VIS)
    G = KelemenSzirmayKalosVis(lightDir, viewVec);
#endif

    // Combine terms
    return specFresnel * (D * G);
}

/**
    Deprecated
 */
float RunescapeLegacyBRDF(vec3 invSunDirection, vec3 invViewDirection, vec3 worldNormal, float specPower, float attenuation)
{
    vec3 reflectionDirection = reflect(-invSunDirection, worldNormal);
    float specularBRDF = pow(max(0.0, dot(reflectionDirection, invViewDirection)), specPower);
    return (specularBRDF * attenuation);
}

/**
    Specular BRDF for Runtek5 materials, assumes a visibility/geometry term of 1
    based on http://renderwonk.com/publications/s2010-shading-course/hoffman/s2010_physically_based_shading_hoffman_b.pdf
 */
float RunescapeRT5BRDF(vec3 halwayVec, vec3 worldNormal, float specPower)
{
    return BlinnPhongNDF(specPower, max(0.0, dot(halwayVec, worldNormal)));
}

vec3 ShiftTangent(vec3 T, vec3 N, float shift)
{
    return normalize(T + shift * N);
}

/**
    Anisoptric BRDF adapted from Hair Rendering Shading, ATI, Scheuermann
 */
vec3 AnisotropicBRDF(vec3 halfwayVec, vec3 normal, vec3 biTangent, vec3 lightDir, vec3 viewDir, float nDotL, float specPow, float specMask, float anisoMask)
{
    const float _SpecularMultiplierT = 7.5;         // Aniso spec power
    const float _GlossInterp = 1.0;                 // Blend between Anios and Blinn (range 0-1)
    const float _Shift = 0.5;                       // Spec highlight tangent shift amount (range 0-1)
    const float _AnisoMaskAmount = 1.0;             // Strength of the Aniso shift mask (range 0-1)

    float shiftTex = specMask - 0.5;
    biTangent = ShiftTangent(biTangent, normal, (_Shift + ((anisoMask * 2.0 - 1.0) * _AnisoMaskAmount)) + shiftTex);

    float biDotL = abs(dot(biTangent, lightDir));
    float sq1 = (1.0 - biDotL);
    float sq2 = (1.0 - abs(dot(biTangent, viewDir)));
    float aniso = biDotL * dot(normal, viewDir);
    aniso += sq1 * sq2;
    aniso = pow(aniso, _SpecularMultiplierT) * nDotL;
    aniso = mix(aniso, aniso * anisoMask, _AnisoMaskAmount);

    float blinn = pow(dot(normal, halfwayVec), specPow);
    float spec = mix(blinn, aniso, _GlossInterp);

    return vec3(spec, spec, spec);
}


#endif
