#ifndef FRESNEL_INC
#define FRESNEL_INC


/**
    Fresnel approximation function for use only specular and diffuse lighting contribution.

    @param specRefl Normal-Incidence Fresnel Reflectance of material (a.k.a base reflectance (specular colour))
    @param product Should be l dot h or v dot h for specular highlights and v dot n for environment mapping
    @param exp Allows for customisation of the exponent value, default is 5.0
    @return The fresnel specular reflectance term

    @todo Implement optimised spherical gaussian version
 */
vec3 FresnelSchlick(vec3 specRefl, float product, highp float exp)
{
    vec3 fresnel = specRefl + (1.0 - specRefl) * pow(1.0 - product, exp);
    return fresnel;
}

/**
    Same as FresnelSchlick with the additiona of taking surface roughness into account.
    Use this version for indirect image based specular environment mapped lighting.

    @param invRoughness 1.0 - Roughness
 */
vec3 FresnelSchlickRoughness(vec3 specRefl, float product, highp float exp, float invRoughness)
{
    vec3 fresnel = specRefl + (max(vec3(invRoughness), specRefl) - specRefl) * pow(1.0 - product, exp);
    return fresnel;
}

/**
    Same as FresnelSchlick except only a single value for the specular reflectance parameter is used
 */
float FresnelSchlick(float specRefl, float product, highp float exp)
{
    float fresnel = specRefl + (1.0 - specRefl) * pow(1.0 - product, exp);
    return fresnel;
}

/**
    Same as FresnelSchlick with the additiona of taking surface roughness into account.
    Use this version for indirect image based specular environment mapped lighting.

    @param invRoughness 1.0 - Roughness
 */
float FresnelSchlickRoughness(float specRefl, float product, highp float exp, float invRoughness)
{
    float fresnel = specRefl + (max(invRoughness, specRefl) - specRefl) * pow(1.0 - product, exp);
    return fresnel;
}

/**
    Alternative Schlick Fresnel that computes the dot product
 */
float FresnelSchlick(vec3 normal, vec3 eyeVec, float specRefl)
{
    float cosAngle = max(0.0, dot(normal, eyeVec));
    return specRefl + (1.0 - specRefl) * pow(1.0 - cosAngle, 5.0);
}

float Fresnel(vec3 normal, vec3 eyeVec, float refractionStrength, float specRefl)
{
    float angle = 1.0 - max(0.0, dot(normal, eyeVec));
    float fresnel = angle * angle;
    fresnel = fresnel * fresnel;
    fresnel = fresnel * angle;
    return clamp(fresnel * (1.0 - clamp(specRefl, 0.0, 1.0)) + specRefl - refractionStrength, 0.0, 1.0);
}

#endif
