#ifndef NDF_INC
#define NDF_INC

#include "MathUtils.inc"


/**
    BRDF specular normal distribution functions (a.k.a N term or specular distribution)
 */

/**
    Blinn-Phong normal distribution function for a punctual light source.
    As used in Call of Duty Black Ops, http://blog.selfshadow.com/publications/s2013-shading-course/lazarov/s2013_pbs_black_ops_2_notes.pdf

    @param specPow Must be between 0-8192
 */
float BlinnPhongNDF(float specPow, float nDotH)
{
    // (a + 2.0) / 8.0 normalization factor, which results from multiplying the (ap + 2) / 2PI normal distribution
    // function normalization factor with the PI / 4 constant from Equation 5, Crafting Physically Motivated Shading Models for Game Development.
    // This explains why when using BlinnPhongNDF, there is no need to multiply the BRDF by PI / 4, as seen in CookTorrence, equation 5.
    return ((specPow + 2.0) * INV_EIGHT) * pow(nDotH, specPow);
}

/**
    As used in UE4, http://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf

    @param roughness is roughness squared or academic roughness 'm', use PerceptualRoughnessToRoughness()
 */
float GGXTrowbridgeReitzNDF(float roughness, float nDotH)
{
    float roughnessSqrd = roughness * roughness;
    float nDotHSqrd = nDotH * nDotH;
    float distribution = nDotHSqrd * (roughnessSqrd - 1.0) + 1.0;
    return roughnessSqrd / (PI * (distribution * distribution + 1e-4));
    //float distribution = (nDotH * roughnessSqrd - nDotH) * nDotH + 1.0; // 2 mad
    //return INV_PI * roughnessSqrd / (distribution * distribution + 1e-4f);

}

/**
    @param Beckmann roughness, use SpecPowToBeckmannRoughness(specPow), other sources say roughness can also be 0-1 range (i.e. Lazarov 2011), try and see.
 */
float BeckmannNDF(float roughness, float nDotH)
{
    float roughnessSqrd = roughness * roughness;
    float nDotHSqrd = nDotH * nDotH;
    return exp((nDotHSqrd - 1.0) / (roughnessSqrd * nDotHSqrd)) / (PI * roughnessSqrd * (nDotHSqrd * nDotHSqrd));
}


#endif
