#ifndef VISIBILITY_FUNC_INC
#define VISIBILITY_FUNC_INC

#include "MathUtils.inc"


/**
    BRDF specular geometric visibility attenuation terms (a.k.a G term).
    Also described as shadowing-masking grouped with foreshortening terms.
 */

/**
    Shadowing-masking grouped with foreshortening terms
    As used in Call of Duty Black Ops, http://blog.selfshadow.com/publications/s2013-shading-course/lazarov/s2013_pbs_black_ops_2_notes.pdf

    @param specPow Blinn-Phong specular power
 */
float SchlickSmithVis(float specPow, float nDotL, float nDotV)
{
    // Calculate k from Blinn-Phong specular power. Note, k is not Beckmann roughness as some sources seem to indicate.
    // If BeckMann roughness is available, use k = m * sqrt(2.0 / PI); with m being Beckmann roughness.
    // These two equations produce the same result, but the latter has been simplified (however the former may be quicker due to MAD efficiency.)
    float k = 1.0 / sqrt((PI_OVER_4 * specPow) + PI_OVER_2);
    //float k = 2.0 / sqrt(PI * (specPow + 2.0));
    //float k = specPow * SQRT_2_PI; // Use if specPow == PerceptualRoughnessToRoughness(perceptualRoughness)
    float one_minus_k = 1.0 - k;
    float visibility_term = (nDotL * one_minus_k + k) * (nDotV * one_minus_k + k); // Both dot products should be saturated
    return 1.0 / (visibility_term + 1e-4);
}

/**
    Kelemen-Szirmay-Kalos approximation to Cook-Torrance visibility function.
    Downside of KSK is that roughness is not taken into account.
 */
float KelemenSzirmayKalosVis(vec3 viewVec, vec3 lightVec)
{
    vec3 VL = viewVec + lightVec;
    return 4.0 / max(0.0, dot(VL, VL));
}


#endif
